#include "Config_YM7101.h"
#include "YM7101.h"
#include "DisplayAdapter.h"
#include <cs_core/Logging/Logging.h>
#include <cs_core/Utils/Maths.h>
#include <boost/format.hpp>
#include <fstream>
#include <cs_core/Utils/Debugging.h>

using codeslinger::logging::LogMessage ;
using codeslinger::maths::BitConditionalSet ;
using codeslinger::maths::BitSet ;
using codeslinger::maths::BitReset ;
using codeslinger::maths::TestBit ;
using codeslinger::maths::BitCount ;
using codeslinger::maths::BitGetVal ;

// todo list:
	// PAL -  PAL in status register
	// light gun
	// extended vram
	// dma timings, it seems galaxy force 2 done something weird where it checks the status register bit to make sure it is in dma transferring and sets up other stuff 

//////////////////////////////////////////////////////////////////////

YM7101::PTR YM7101::Create()
{
	codeslinger::logging::LoggingAddNewGroup("YM7101");
	return PTR( new YM7101() ) ;
}

//////////////////////////////////////////////////////////////////////

YM7101::YM7101(void)
	: m_VCounter(0)
	, m_HCounter(0)
	, m_StatusRegister(0)
	, m_ControlPortPending(false)
	, m_ControlPortData(0)
	, m_RequestInterrupt(false)
	, m_Refresh(false)
	, m_LineInterrupt(0xFF)
	, m_MaxSprites(0)
	, m_SpritePixelCount(0)
	, c_maxInt(std::numeric_limits<unsigned char>().max())
	, m_SpriteMask2(0xff)
{
	Reset( ) ;
}

//////////////////////////////////////////////////////////////////////

YM7101::~YM7101(void)
{
}

//////////////////////////////////////////////////////////////////////

bool YM7101::ApplySTE( bool priorityHigh, int xpos, bool isSprite, int spriteColour, int spritePalette )
{
	bool res = false;
	int newSte = DisplayAdapter::STE_NOTSET;
	int currentSte = m_LineHighlightData[xpos];
	if (currentSte == DisplayAdapter::STE_NOTSET)
	{
		if (!priorityHigh)
		{
			newSte = DisplayAdapter::STE_SHADOW;
		}
		else
		{
			newSte = DisplayAdapter::STE_NORMAL;
		}
	}
	else if (currentSte == DisplayAdapter::STE_SHADOW)
	{
		if (priorityHigh)
		{
			newSte = DisplayAdapter::STE_NORMAL;
		}
	}

	if (isSprite && (spritePalette == 3) )
	{
		if (spriteColour == 14 && (currentSte < DisplayAdapter::STE_HIGHLIGHT))
		{
			res = true;
			newSte = currentSte + 1;
		}
		else if (spriteColour == 15 && (currentSte > DisplayAdapter::STE_SHADOW))
		{
			res = true;
			newSte  = currentSte - 1;
		}
	}

	if (newSte != DisplayAdapter::STE_NOTSET)
	{
		m_LineHighlightData[xpos] = static_cast<DisplayAdapter::STE>(newSte);
	}
	return res ;
}

//////////////////////////////////////////////////////////////////////

bool YM7101::IsTimeToRender() const
{
	return m_Refresh ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::SetFuncReadMemory16( const t_FuncReadMemory16& func )
{
	m_FuncReadMemory16 = func ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::ResizeScreen(const std::pair<int,int>& size)
{
	m_DisplayAdapter.ResizeScreen(size);
	m_DisplayAdapter.ResetScreen();
}

//////////////////////////////////////////////////////////////////////

void YM7101::Reset()
{
	ResizeScreen(std::make_pair<int,int>(256,224));

	m_SpriteMaskData = std::make_pair(0xff,std::make_pair(0xff,0xff));
	m_RequestInterrupt = false ;
	m_StatusRegister = 0x3600 ;
	m_StatusRegister = BitSet(m_StatusRegister, 9) ;
	memset(m_VRAM.vram8, 0x0, 0x10000) ;

	for (int i = 0; i < sc_numregisters; ++i)
	{
		m_Registers[i] = 0 ;
	}

	for (int i = 0 ; i < sc_numcram; ++i)
	{
		m_CRAM[i] = 0 ;
	}

	for (int i = 0 ; i < sc_numvsram; ++i)
	{
		m_VSRAM[i] = 0 ;
	}

	m_HCounterFirst = true ;
	m_VCounterFirst = true ;
}

//////////////////////////////////////////////////////////////////////

bool YM7101::IsRequestingInterrupt() const
{
	return m_RequestInterrupt ;
}

//////////////////////////////////////////////////////////////////////

std::pair<int,int> YM7101::GetScreenResolution( ) const
{
	return m_DisplayAdapter.GetScreenRes() ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::IncrementAddress()
{
 	const UBYTE c_amount = m_Registers[0xF] ;
 	UWORD address = GetAddressRegister() ;
 	address += c_amount ;

 	// must be a faster way to do this
 	m_ControlPortData = BitConditionalSet( TestBit(address, 15), m_ControlPortData, 1 ) ;
 	m_ControlPortData = BitConditionalSet( TestBit(address, 14), m_ControlPortData, 0 ) ;
 	for (int i = 0 ; i < 14; ++i)
 	{
 		m_ControlPortData = BitConditionalSet( TestBit(address, 13-i), m_ControlPortData, 29-i ) ;
 	}
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetHVCounter() const
{
	UBYTE hcounter = (UBYTE)m_HCounter >> 1 ;
	UBYTE vcounter = (UBYTE)GetVCounter();
	return (vcounter << 8) | hcounter ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteControlPort16( UWORD data )
{
	if (!m_ControlPortPending)
	{
		UBYTE code = static_cast<UBYTE>( (data >> 14) & 0x3 );

		if (code == 2)
		{
			UpdateRegister( data ) ;
		}
		else
		{
			m_ControlPortData &= 0xFFFF ;
			m_ControlPortData |= ((UDWORD)data << 16);
			m_ControlPortPending = true ;
		}
	}
	else
	{
		m_ControlPortPending = false ;
		m_ControlPortData &= 0xFFFF0000 ;
		m_ControlPortData |= data ;

		if (TestBit( m_ControlPortData, 7) && IsRegBitSet(1,4) )
		{
			DoDMATransfer( data ) ;
			return ;
		}
	}
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteControlPort8( UBYTE data )
{
	UWORD doubledata = (data << 8) | data ;
	WriteControlPort16(doubledata) ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::DoDMA68kToVRAM()
{
 	UWORD length = (m_Registers[20] << 8) | m_Registers[19] ;

	if (length == 0x0)
		length = 0xffff;

	const UDWORD c_sourceAddress = ((m_Registers[23] & 0x7F) << 17) | (m_Registers[22] << 9) | (m_Registers[21] << 1) ;
	UDWORD sourceAddress = c_sourceAddress;
	const UWORD c_addressRegister = GetAddressRegister();

	int code = (m_ControlPortData >> 30) & 0x3;
	code = BitConditionalSet(TestBit(m_ControlPortData, 4),code,2);
 	do
 	{
 		if (sourceAddress > 0xFFFFFF )
 		{
 			sourceAddress = (sourceAddress - 0xFFFFFF) + 0xFF0000 ;
 		}
 		const UWORD c_data = m_FuncReadMemory16( sourceAddress ) ;
 		sourceAddress += 2;
		switch(code)
		{
		case 1: WriteVram( c_data ) ;break;
			case 3: WriteCram( c_data ) ;break;
				case 5: WriteVsram( c_data ) ;break;
		}

		IncrementAddress( ) ;
 		--length ;
 	} while ( length > 0);
	m_Registers[19] = length & 0xFF;
	m_Registers[20] = (length >> 8) & 0xFF;
	m_Registers[21] = (sourceAddress >> 1) & 0xFF;
	m_Registers[22] = (sourceAddress >> 9) & 0xFF;
	m_Registers[23] = ( m_Registers[23] & 0x80) | ((sourceAddress >> 17) & 0x7F);

}

//////////////////////////////////////////////////////////////////////

void YM7101::DoDMAVRAMCopy( )
{
	UWORD length = (m_Registers[20] << 8) | m_Registers[19] ;
	UWORD source = (m_Registers[22] << 8) | m_Registers[21];
	do 
	{
		UBYTE data = m_VRAM.vram8[source];
		UWORD address = GetAddressRegister();
		IncrementAddress();
		m_VRAM.vram8[address] = data;
		++source;
		--length;
	} while (length);
}

//////////////////////////////////////////////////////////////////////

void YM7101::DoDMAVRAMFill( UWORD data )
{
	LogMessage("YM7101", "info", 0, "Doing DMA VRAM Fill");
//	m_StatusRegister = BitSet(m_StatusRegister, 1) ;
	UWORD length = (m_Registers[20] << 8) | m_Registers[19] ;
	const UWORD c_address = GetAddressRegister() ;
	m_VRAM.vram8[c_address] = (UBYTE)( (data >> 8) & 0xFF ) ;

	IncrementAddress() ;
	do
	{
		const UWORD c_address = GetAddressRegister() ;
		m_VRAM.vram8[c_address] = (UBYTE)( (data >> 8) & 0xFF ) ;
		IncrementAddress() ;
	} while (--length);
}

//////////////////////////////////////////////////////////////////////

void YM7101::DoDMATransfer( UWORD data )
{
	const int c_dmatype = (m_Registers[23] >> 6) & 0x3 ;
  	switch(c_dmatype)
  	{
		case 0: DoDMA68kToVRAM(); break ; 
		case 1: DoDMA68kToVRAM(); break ;
		case 2: DoDMAVRAMFill(data) ; break ;
		case 3: DoDMAVRAMCopy(); break;
	default: LogMessage("YM7101", "err", "Invalid vdp dma type") ; break ;
  	}
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteDataPort16( UWORD data )
{
	const ControlCode c_code = GetControlCode() ;

	// ignore reads in write
	if (c_code == CC_READ_VRAM || c_code == CC_READ_CRAM || c_code == CC_READ_VSRAM)
	{
		return ;
	}

	m_ControlPortPending = false ;

	switch(c_code)
	{
		case CC_WRITE_VRAM: WriteVram(data) ; break ;
		case CC_WRITE_CRAM: WriteCram(data) ; break ;
		case CC_WRITE_VSRAM: WriteVsram(data) ;break ;
		default: LogMessage("YM7101", "wrn", "Unknown code in WriteDataPort16") ; break ;
	}
	IncrementAddress( ) ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteDataPort8( UBYTE data )
{
	UWORD doubledata = (data << 8) | data ;
	WriteDataPort16(doubledata) ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::UpdateRegister( UWORD data )
{
	UBYTE reg = ( data >> 8 ) & 0x7F ;
	if (reg < sc_numregisters)
	{
		m_Registers[reg] = data & 0xFF ;
	}
}

//////////////////////////////////////////////////////////////////////

YM7101::ControlCode YM7101::GetControlCode() const
{
	const UBYTE c_low = static_cast<UBYTE>( m_ControlPortData >> 30 ) & 0x3 ;
	const UBYTE c_high = static_cast<UBYTE>( (m_ControlPortData >> 4 ) & 0x3 ) ;
	return static_cast<ControlCode>( (c_high << 2) | c_low ) ;
}


//////////////////////////////////////////////////////////////////////

UWORD YM7101::ReadDataPort16()
{
	const ControlCode c_code = GetControlCode() ;

	// ignore writes in read
	if (c_code == CC_WRITE_VRAM || c_code == CC_WRITE_REG
		|| c_code == CC_WRITE_CRAM || c_code == CC_WRITE_VSRAM)
	{
		return 0x0;
	}

	m_ControlPortPending = false ;

	UWORD res = 0 ;
	switch(c_code)
	{
	case CC_READ_VRAM: res = ReadVram16( ) ; break ;
	case CC_READ_CRAM: res = ReadCram16( ) ; break ;
	case CC_READ_VSRAM: res = ReadVsram16( ) ;break ;
	default: LogMessage("YM7101", "wrn", "Unknown code in ReadDataPort16") ; break ;
	}
	IncrementAddress( ) ;
	return res ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::ReadControlPort16()
{
	UWORD res = m_StatusRegister ;
	m_ControlPortPending = false ;
	m_StatusRegister = BitReset(m_StatusRegister, 1) ;
	return res ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::GetHCounter() const
{
	return (UBYTE)m_HCounter ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::GetVCounter() const
{
	UBYTE res = (UBYTE)m_VCounter;
	if ( IsRegBitSet(0xC, 1)) // interlaced mode
	{
		res = BitConditionalSet(TestBit(m_VCounter, 8), res, 0) ;
	}
	return res ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteVram( UWORD data )
{
	UWORD address = GetAddressRegister( ) ;
	if ( TestBit(address,0) )
	{
		codeslinger::maths::SwapEndian( data ) ;
	}
	int count = 0;
	if (address == 0xc129)
		++count;
	m_VRAM.vram8[address] = data >> 8 ;
	m_VRAM.vram8[address+1] = data & 0xff ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteCram( UWORD data )
{
	UWORD address = GetAddressRegister( ) % 0x80 ;
	//address = BitReset(address,0);
	address = address >> 1 ;
	m_CRAM[address] = data ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::WriteVsram( UWORD data )
{
	UWORD address = GetAddressRegister( ) % 0x80 ;
//	address = BitReset(address,0);
	address = address >> 1 ;
	//codeslinger::maths::SwapEndian(data);
// 	if (data == 0x300)
// 		_asm int 3;

	m_VSRAM[address] = data ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetAddressRegister() const
{
	UWORD res = m_ControlPortData >> 16 ;
	res &= 0x3FFF ;

	UWORD bits = m_ControlPortData & 0x3 ;
	res |= bits << 14 ;

	return res ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::ReadVram8() const
{
	const UWORD c_address = GetAddressRegister() ;
	return m_VRAM.vram8[c_address] ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::ReadCram8() const
{
	LogMessage("YM7101", "wrn"," dont know what to do ReadCram8 as cram is 16 bit not 8") ;
	return 0 ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::ReadVsram8() const
{
	LogMessage("YM7101", "wrn"," dont know what to do ReadVsram8 as vsram is 16 bit not 8") ;
	return 0 ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::ReadVram16() const
{
	UWORD address = GetAddressRegister( );// % 0x80 ;
	//address = address >> 1 ;
	//address &= 0x3F ;
	return m_VRAM.vram8[address] << 8 | m_VRAM.vram8[address+1] ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::ReadCram16() const
{
	ASSERT(false); // untested
	UWORD address = GetAddressRegister( ) % 0x80 ;
	address = address >> 1 ;
	//address &= 0x3F ;
	return m_CRAM[address] ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::ReadVsram16() const
{
	ASSERT(false); // untested
	const UWORD c_address = (GetAddressRegister() % 0x80) >> 1 ;

	return m_VSRAM[c_address] ;
}

//////////////////////////////////////////////////////////////////////

void YM7101::Update( int numclicks )
{
	std::pair<int, int> currentRes = GetScreenResolution();
	const bool c_is40 = GetIs40CellMode();
	if (m_DisplayAdapter.GetScreenRes().first == 256 && c_is40)
	{
		ResizeScreen(std::make_pair<int,int>(320,224));
		currentRes = GetScreenResolution();
	}
	else if (m_DisplayAdapter.GetScreenRes().first == 320 && !c_is40)
	{
		ResizeScreen(std::make_pair<int,int>(256,224));
		currentRes = GetScreenResolution();
	}
	const int c_height = currentRes.second ;

	UWORD hcount = m_HCounter ;
	bool nextline = false ;
//	m_IsVBlank = false ;
	m_Refresh = false ;


	// are we moving off this scanline onto the next?
	if ((hcount + numclicks) > 0xFF && !m_HCounterFirst)
		nextline = true ;

	// if we are starting a new scanline reset the hcounter
	m_HCounter = (m_HCounter + numclicks) % 0xFF;

	const int c_jumpBackCoord = c_is40 ? 0xe9 : 0xe9;
	const int c_jumpBackTo = c_is40 ? 0x53 : 0x93;

	if (m_HCounterFirst && m_HCounter > c_jumpBackCoord)
	{
		m_HCounterFirst = false ;
		const int c_delta = m_HCounter - c_jumpBackCoord ;
		m_HCounter = c_jumpBackTo + c_delta ;
	}

	m_StatusRegister = BitConditionalSet( m_HCounter >= 0xE4 || m_HCounter < 0x8,
		m_StatusRegister, 2 ) ;

	// we are moving onto the next scanline
	if (nextline)
	{
		m_HCounterFirst = true ;

		// store current scanline
		UWORD vcount = m_VCounter ;
		m_VCounter++ ; // move onto next scanline

		// are we coming to the end of the vertical refresh?
		// if so we are starting a new frame from scanline 0
		if (vcount == 0xFF)
		{
			m_VCounter = 0 ;
			m_VCounterFirst = true ;
			m_StatusRegister = BitReset(m_StatusRegister, 7) ;
			m_StatusRegister = BitReset(m_StatusRegister, 5) ;
		//	m_DisplayAdapter.ResetScreen();
			m_DisplayAdapter.Begin();
			m_LineInterrupt = m_Registers[0xA] ;
			if (IsRegBitSet(0xc,1))
			{
				// odd frame
				m_StatusRegister = BitConditionalSet( !TestBit(m_StatusRegister, 4), m_StatusRegister, 4);
			}
			m_SpriteMaskData = std::make_pair(0xff,std::make_pair(0xff,0xff));
			m_SpriteMask2 = 0xff;
			Render( ) ;
		}
		// is it time to jump the vcounter backwards?
		else if ((vcount == 0xEB) && m_VCounterFirst)
		{
			m_VCounterFirst = false ;
			const int c_delta = m_VCounter - 0xEB ;
			m_VCounter = 0xE5 + c_delta ;
		}

		// are we just about to enter vertical refresh?
		else if (m_VCounter == c_height)
		{
		//	m_IsVBlank = true ;
			m_StatusRegister = BitSet(m_StatusRegister, 7) ; // irq pending
			m_Refresh = true ;
		}

		m_StatusRegister = BitConditionalSet(m_VCounter >= 0xe0, m_StatusRegister, 3 ) ;

		if (m_VCounter >= c_height)
		{
			//// do not reload the line interupt until we are past the
			//// FIRST line of the none active display period
			if (m_VCounter != c_height)
				m_LineInterrupt = m_Registers[0xA] ;
		}

		// if we are in active display then draw next scanline
		if (m_VCounter < c_height)
		{
		//	m_ScreenDisabled = !IsRegBitSet(1,6) ;
			Render( ) ;
		}

		// decrement the line interupt counter during the active period
		// including the first line of the none active display period
 		if (m_VCounter <= c_height)
 		{
 			bool underflow = false ;
 			if (m_LineInterrupt == 0)
 			{
 				underflow = true ;
 			}
 			m_LineInterrupt-- ;

 			// it is going to underflow
 			if (underflow)
 			{
 				m_LineInterrupt = m_Registers[0xA] ;
 				if (IsRegBitSet(0,4))
 					m_RequestInterrupt = true ;
 			}
 		}
	}
	// do we want to signal an interrupt
 	if (TestBit(m_StatusRegister,7) && IsRegBitSet(1,5))
 		m_RequestInterrupt = true ;
}

//////////////////////////////////////////////////////////////////////

int YM7101::GetInterruptType()
{
	int res = -1 ;
	if (m_RequestInterrupt)
	{
		if (TestBit(m_StatusRegister,7) && IsRegBitSet(1,5))
		{
			m_StatusRegister = BitReset(m_StatusRegister, 7) ;
			res = 6  ;
		}
		else
		{
			res = 4 ;
		}
	}
	m_RequestInterrupt = false;
	return res ;
}

//////////////////////////////////////////////////////////////////////

bool YM7101::IsRegBitSet(int reg, int bit) const
{
	return TestBit(m_Registers[reg], bit) ;
}

//////////////////////////////////////////////////////////////////////

bool YM7101::GetIs40CellMode() const
{
	int val = 0 ;
	if (TestBit(m_Registers[0xC], 7))
	{
		val = BitSet( val, 0 ) ;
	}
	if (TestBit(m_Registers[0xC], 0))
	{
		val = BitSet( val, 1 ) ;
	}
	return val > 0 ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetPlaneABase() const
{
	UWORD res = m_Registers[0x2] & 0x38 ;
	res <<= 10 ;
	return res ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetWindowBase() const
{
	UWORD res = m_Registers[0x3] & 0x3E ;
	res <<= 10 ;
	return res ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetPlaneBBase() const
{
	UWORD res = m_Registers[0x4] & 0x7 ;
	res <<= 13 ;
	return res ;
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetSATBase() const
{
	UWORD res = m_Registers[0x5] & 0x7F ;
	res <<= 9 ;
	if (GetIs40CellMode())
	{
		res = BitReset( res, 9) ;
	}
	return res ;
}

//////////////////////////////////////////////////////////////////////

UBYTE YM7101::GetColourShade(UBYTE val) const
{
	// operating in 8 colour mode
	if (!TestBit(m_Registers[0x0], 2))
		val &= 0x3 ;

	val = BitReset(val, 0) ;
	UBYTE index = (val << 4) ;
	return index ;
}

//////////////////////////////////////////////////////////////////////

static int GetPatternRes( int data )
{
	int res = 32;
	switch( data )
	{
		case 0: res = 32;break;
		case 1: res = 64;break;
		case 3: res = 128; break;
		default:
			LogMessage("YM7101", "err", (boost::format("Invalid pattern size %1%") % data ).str());
			break;
	}
	return res;
}


std::pair<int,int> YM7101::GetPlaneSize() const
{
	UBYTE data = m_Registers[0x10];
	int h = data & 0x3;
	int v = (data >> 4) & 0x3;
	return std::make_pair<int,int>(GetPatternRes(h), GetPatternRes(v));
}

//////////////////////////////////////////////////////////////////////

void* YM7101::GetScreen() const
{
	return m_DisplayAdapter.GetScreen();
}

//////////////////////////////////////////////////////////////////////

UWORD YM7101::GetVScrollAmount( PlaneToRender plane, int doubleColumnNum ) const
{
	const bool c_fullPlane = !TestBit(m_Registers[0xB],2);
	// hmmm why are there only 2 scroll types? Full screen and line by line when HSCroll has 3
	switch (plane)
	{
		case PLANE_A:
			if (c_fullPlane) return m_VSRAM[0x0];
			return m_VSRAM[doubleColumnNum * 2];
		case PLANE_B:
			if (c_fullPlane) return m_VSRAM[0x1];
			return m_VSRAM[(doubleColumnNum * 2)+1];
		default:
			return 0;
	}
}

void YM7101::GetHScrollAmount(int line, UWORD& aScroll, UWORD& bScroll ) const
{
	const UWORD c_address = (m_Registers[0xD] & 0x3F) << 10;
	switch(m_Registers[0xb] & 0x3)
	{
	case 0: /* Full screen */
		aScroll = (m_VRAM.vram8[c_address ] << 8) | m_VRAM.vram8[c_address + 1];
		bScroll = (m_VRAM.vram8[c_address + 2] << 8) | m_VRAM.vram8[c_address + 3];
		break;

	case 1: /* First eight lines */
		{
			const int c_val = ((line & 7) * 4);
			aScroll = (m_VRAM.vram8[c_address + c_val ] << 8) | m_VRAM.vram8[c_address + c_val + 1];
			bScroll = (m_VRAM.vram8[c_address + c_val + 2] << 8) | m_VRAM.vram8[c_address + c_val + 3];
		}
		break;

	case 2: /* Every row */
		{
			const int c_val = ((line & ~7) * 4);
			aScroll = (m_VRAM.vram8[c_address + c_val] << 8) | m_VRAM.vram8[c_address + c_val + 1];
			bScroll = (m_VRAM.vram8[c_address + c_val + 2] << 8) | m_VRAM.vram8[c_address + c_val+3];
		}
		break;

	case 3: /* Every line */
		{
			const int c_val = (line * 4);
			aScroll = (m_VRAM.vram8[c_address + c_val] << 8) | m_VRAM.vram8[c_address + c_val + 1];
			bScroll = (m_VRAM.vram8[c_address + c_val + 2] << 8) | m_VRAM.vram8[c_address + c_val + 3];
		}
		break;
	}
	aScroll &= 0x3FF;
	bScroll &= 0x3FF;
}

//////////////////////////////////////////////////////////////////////

void YM7101::Render( )
{
	m_SpritePixelCount = 0;
	if (IsRegBitSet(0, 0 ) )
	{
		// screen disabled
		return ;
	}

	memset( &m_LineHighlightData, DisplayAdapter::STE_NOTSET, sizeof(DisplayAdapter::STE) * 320);

	RenderBackdrop();

	if (!IsRegBitSet(1,6))
	{
		// only render backdrop
		return;
	}


	const UWORD c_planeABase = GetPlaneABase() ;
	const UWORD c_planeBBase = GetPlaneBBase() ;
	const UWORD c_windowBase = GetWindowBase() ;
	const std::pair<int, int> c_planseSize = GetPlaneSize();

	UWORD aHScroll = 0;
	UWORD bHScroll = 0;
	GetHScrollAmount(m_VCounter, aHScroll, bHScroll);

	if (m_VCounter == 104)
	{
		int j = 0 ;
		++j;
	}

	const bool c_40cellmode = GetIs40CellMode( ) ;
	memset(&m_SpriteLineData,c_maxInt ,320 * sizeof(UBYTE));
	for (int i = 0 ; i < 2; ++i)
	{
		m_MaxSprites = c_40cellmode ? 80 : 60 ;
		const bool c_highPriority = i == 1;
		RenderTiles(c_planeBBase, c_highPriority, c_planseSize.first, c_planseSize.second, bHScroll, PLANE_B) ;
		RenderTiles(c_planeABase, c_highPriority, c_planseSize.first, c_planseSize.second, aHScroll, PLANE_A) ;
		RenderSprites(c_highPriority) ;
		RenderWindow( c_windowBase, c_highPriority ) ;
	}

	if (IsRegBitSet(0xc,3))
	{
		const std::pair<int, int> c_screenres = m_DisplayAdapter.GetScreenRes();
		for (int i = 0 ; i < c_screenres.first; ++i)
		{
			m_DisplayAdapter.ApplySTE(i, m_VCounter, m_LineHighlightData[i]) ;
		}
	}
}	

//////////////////////////////////////////////////////////////////////

void YM7101::RenderBackdrop()
{
	const std::pair<int,int>& c_screenres = GetScreenResolution() ;
	if (m_VCounter >= c_screenres.second)
		return ;
	const UWORD c_colour = m_CRAM[m_Registers[0x7] & 0x3f]; ;

	const UBYTE c_blue = GetColourShade( (c_colour >> 8) & 0xF ) ;
	const UBYTE c_red =  GetColourShade( c_colour & 0xF ) ;
	const UBYTE c_green =  GetColourShade( ( c_colour >> 4 ) & 0xF ) ;

	for (int x = 0; x < c_screenres.first; ++x)
	{
		m_DisplayAdapter.WriteToScreen(x, m_VCounter, c_red, c_green, c_blue);
	}
}

//////////////////////////////////////////////////////////////////////

void YM7101::RenderTiles(const UWORD baseAddress, bool priorityhigh, int hSize, int vSize, UWORD hScrollAmount, PlaneToRender plane )
{
	const int c_vCounter = m_VCounter ;

	const std::pair<int,int> c_screenres = GetScreenResolution() ;

  	UWORD hStartingCol = (hScrollAmount >> 3) ;
  	hStartingCol = (hSize - (hStartingCol % hSize)) % hSize;

  	UWORD hFineScroll = (hScrollAmount & 0x7) ;
  	hFineScroll = (8 - (hFineScroll % 8)) % 8;

	int numHorizonalTiles =  (c_screenres.first / 8); // 32 or 40 tiles
	const int c_numVerticleTiles = c_screenres.second / 8;

	if (hFineScroll != 0) // fine scrolling will cause one tile to only appear partial and therefore introduce another partial tile
	{
		++numHorizonalTiles;
		if (hStartingCol == 0)
			hStartingCol = hSize - 1;
		else
			--hStartingCol;
	}

	const int c_tileRow = (c_vCounter / 8);
	int pixelCount = 0 ;
	// draw all 32 columns for this scanline (we only actually draw a pixel of the current column which is the scanline)
	for (int tilecolumn = 0; tilecolumn < numHorizonalTiles; ++tilecolumn)
	{
		// draw all 8 pixels in the column
		for (int xAdvance = hFineScroll ; xAdvance < 8; ++xAdvance)
		{
			const int c_absx = pixelCount++;

			UWORD c_vScrollAmount = GetVScrollAmount(plane, c_absx / 16) ;
			if (IsRegBitSet(0xc,1))
			{
				c_vScrollAmount >>=1;
			}
			const UWORD c_vStartingRow = c_vScrollAmount >> 3;
			const UWORD c_vFineScroll = c_vScrollAmount & 0x7;

			// a row is 8 pixels large
			int scrollRow =  c_tileRow + c_vStartingRow ;
			if ( ((c_vCounter % 8) + c_vFineScroll) > 7)
 				++scrollRow;
//
			scrollRow %= vSize;


			// get correct position in memory for this tile
			UWORD nameBaseOffset = baseAddress ;
			nameBaseOffset += scrollRow * (hSize * 2) ; //each plane row has hSize tiles but 1 tile is 2 bytes in memory
			nameBaseOffset += (codeslinger::maths::modulus_neg<int>(hStartingCol + tilecolumn, hSize) * 2) ; // each tile is two bytes in memory

			const bool c_hFlip = TestBit(m_VRAM.vram8[nameBaseOffset], 3) ;
			const bool c_vFlip = TestBit(m_VRAM.vram8[nameBaseOffset], 4);

			const int c_tilexpixel = (( c_hFlip ? (7 - xAdvance) : xAdvance) ) % 8 ;
			const int c_scrollypixel  = (c_vCounter + c_vFineScroll) % 8;
			const int c_tileypixel = (c_vFlip ? (7 - c_scrollypixel) : c_scrollypixel);


			// check if correct priority
 			if (TestBit(m_VRAM.vram8[nameBaseOffset], 7) != priorityhigh)
 			{
				continue ;
 			}

			// get tile information data from memory
			UWORD tileData = (m_VRAM.vram8[nameBaseOffset] & 0x7) << 8 ;
			tileData |= m_VRAM.vram8[nameBaseOffset+1] ;
 			if ( IsRegBitSet(0xc, 1) )
 			{
 				tileData *= 64;
 
 				// each pattern line if 4 bytes. Offset contains correct pattern line
 				tileData += 8 * c_tileypixel;
 			}
 			else
 			{
				tileData *= 32 ;

				// each pattern line if 4 bytes. Offset contains correct pattern line
				tileData += 4 * c_tileypixel;
			}


			// get pattern line data
			UBYTE data1 = m_VRAM.vram8[tileData] ;
			UBYTE data2 = m_VRAM.vram8[tileData+1] ;
			UBYTE data3 = m_VRAM.vram8[tileData+2] ;
			UBYTE data4 = m_VRAM.vram8[tileData+3] ;

			int colourindex = 0 ;
			// quick hack

			switch(c_tilexpixel)
			{
				case 0 : colourindex = data1 >> 4 ; break ;
				case 1 : colourindex = data1 & 0xF ; break ;
				case 2 : colourindex = data2 >> 4 ; break ;
				case 3 : colourindex = data2 & 0xF ; break ;
				case 4 : colourindex = data3 >> 4 ; break ;
				case 5 : colourindex = data3 & 0xF ; break ;
				case 6 : colourindex = data4 >> 4 ; break ;
				case 7 : colourindex = data4 & 0xF ; break ;
			}


			const UBYTE c_colourpalette = ( m_VRAM.vram8[nameBaseOffset] >> 5 ) & 0x3 ;

			const int c_colourIndex = (16 * c_colourpalette) + colourindex;
			const UWORD c_colour = m_CRAM[c_colourIndex] ;

			UBYTE c_red =  GetColourShade( c_colour & 0xF ) ;
			UBYTE c_green =  GetColourShade( ( c_colour >> 4 ) & 0xF ) ;
			UBYTE c_blue = GetColourShade( (c_colour >> 8) & 0xF ) ;

 			if (IsRegBitSet(0xc,3))
 			{
 				ApplySTE( priorityhigh, c_absx, false, colourindex, c_colourpalette ) ;
 			}
			
			if (colourindex == 0) // really unsure about this
				continue;

			// blank left most column
			if (IsRegBitSet(0,5) && c_absx < 8)
				continue;

// 			if (priorityhigh)
// 				m_SpriteLineData[c_absx] = c_maxInt;

			// we now have all the data needed to colour this pixel. So write it to the screen			
			m_DisplayAdapter.WriteToScreen(c_absx, c_vCounter, c_red, c_green, c_blue);
		}
		hFineScroll  = 0;
	}
}

//////////////////////////////////////////////////////////////////////

void YM7101::RenderSprites( bool highPriority )
{
	const std::pair<int,int> c_screenres = GetScreenResolution() ;

	int vCounter = m_VCounter ;
	int spriteCount = 0;
	UWORD satbase = GetSATBase();


	const bool c_40cellmode = GetIs40CellMode( ) ;
	const int c_maxspritesonscanline = c_40cellmode ? 20 : 16 ;
	const int c_maxpixels = c_40cellmode ? 320 : 256 ;

	int nextsprite = 0 ;
	int spriteIndex = -1 ;
	// loop through all 80 sprites and see which fall within range of the current scanline
	do
	{
		++spriteIndex;
		const int c_sprite = nextsprite;
		const int c_index = c_sprite * 8 ;
		nextsprite = m_VRAM.vram8[satbase + c_index + 3] & 0x7f;

		--m_MaxSprites; 
		if (TestBit( m_VRAM.vram8[satbase + c_index + 4], 7) != highPriority)
			continue;

		int ySpritePos = m_VRAM.vram8[satbase+c_index] ;
		ySpritePos <<= 8;
		ySpritePos |= m_VRAM.vram8[satbase + c_index + 1] ;

		if ( IsRegBitSet(0xC, 1))
		{
			ySpritePos = (ySpritePos & 0x3fe) >> 1;
		}
		else
		{
			ySpritePos &= 0x3FF;
		}
		
		ySpritePos -= 128;
		const int c_vSize = (m_VRAM.vram8[satbase + c_index + 2] & 0x3);

		const int c_ySpriteHeight = ( 8 * (c_vSize + 1) ) ;

		// does the sprite fall within range of current scanline taking into account its size?
		if ((vCounter>=ySpritePos) && (vCounter < (ySpritePos+c_ySpriteHeight)))
		{
			const int c_hVal = (m_VRAM.vram8[satbase + c_index + 2] >> 2) & 0x3;

			// record how many sprites we have drawn this scanline
			spriteCount++;

			// we can only draw 8 sprites per scanline. If drawing more set overflow and stop
			if (spriteCount > c_maxspritesonscanline)
			{
				m_StatusRegister = BitSet(m_StatusRegister, 6) ;
				break;
			}

			int xPos = m_VRAM.vram8[satbase+c_index+6] & 0x3 ;
			xPos <<= 8;
			xPos |= m_VRAM.vram8[satbase + c_index + 7] ;

			if ( xPos == 0 && !IsRegBitSet(0xc,1))
			{
				m_SpriteMaskData = std::make_pair(spriteIndex, std::make_pair(ySpritePos, ySpritePos + c_ySpriteHeight));
			}
			else if (xPos == 1)
			{
				m_SpriteMask2 = ySpritePos;
			}

			xPos-= 128;
			const UBYTE c_colourpalette = ( m_VRAM.vram8[satbase+c_index+4] >> 5 ) & 0x3 ;

			UWORD tileNumber = m_VRAM.vram8[satbase+c_index+4] & 0x7 ;
			tileNumber <<= 8 ;
			tileNumber |= m_VRAM.vram8[satbase+c_index+5] ;
			
			const bool c_vFlip = TestBit(m_VRAM.vram8[satbase+c_index+4],4) ;
			const bool c_hFlip = TestBit(m_VRAM.vram8[satbase+c_index+4],3) ;

			// get the exact memory location for the current line being drawn of the tile
			// remember each line is 4 bytes
			int yIndex = vCounter-ySpritePos;
			if (c_vFlip)
			{
				yIndex = (c_ySpriteHeight-1) - yIndex ;
			}

			int startAddress = 0 ;
			int memSize = 0;

			if (IsRegBitSet(0xc,1))
			{
				memSize = 64;
				startAddress = tileNumber * memSize;
				startAddress += (8 * yIndex) ;
			}
			else
			{
				// each tile takes up 32 bytes in memory (4 bytes per line)
				memSize = 32;
				startAddress = tileNumber * memSize;
				startAddress += (4 * yIndex) ;
			}

			for (int hSize = 0; hSize <= c_hVal; ++hSize)
			{
				const int c_patternIndex = (c_hFlip ? (c_hVal - hSize) : hSize);
				const int c_patternAddress = startAddress + (c_patternIndex * ((c_vSize + 1) * memSize)) ;
				// gather the 4 bytes of data needed to draw the current line of the tile
				UBYTE data1 = m_VRAM.vram8[c_patternAddress] ;
				UBYTE data2 = m_VRAM.vram8[c_patternAddress+1] ;
				UBYTE data3 = m_VRAM.vram8[c_patternAddress+2] ;
				UBYTE data4 = m_VRAM.vram8[c_patternAddress+3] ;

				// draw all 8 pixels for the current tile line
				for (int i = 0; i < 8; ++i)
				{
					const int c_x = (xPos + i);

					if ( (c_x < 0) || (c_x >= c_screenres.first) || (IsRegBitSet(0,5) && c_x < 8))
					{
						++m_SpritePixelCount;
						if (m_SpritePixelCount >= c_maxpixels)
							return;
						continue;
					}

					int colourindex = 0 ;
					// quick hack
					switch( c_hFlip ? ( 7 - i) : i )
					{
						case 0 : colourindex = data1 >> 4 ; break ;
						case 1 : colourindex = data1 & 0xF ; break ;
						case 2 : colourindex = data2 >> 4 ; break ;
						case 3 : colourindex = data2 & 0xF ; break ;
						case 4 : colourindex = data3 >> 4 ; break ;
						case 5 : colourindex = data3 & 0xF ; break ;
						case 6 : colourindex = data4 >> 4 ; break ;
						case 7 : colourindex = data4 & 0xF ; break ;
					}
									

					const UWORD c_colour = m_CRAM[(16 * c_colourpalette) + colourindex] ;

					const UBYTE c_red =  GetColourShade( c_colour & 0xF ) ;
					const UBYTE c_green =  GetColourShade( ( c_colour >> 4 ) & 0xF ) ;
					const UBYTE c_blue = GetColourShade( (c_colour >> 8) & 0xF ) ;

					
					bool transparentSte = false;
					if (IsRegBitSet(0xc,3) && (m_SpriteLineData[c_x] > spriteIndex))
					{
						transparentSte = ApplySTE( highPriority, c_x, true, colourindex, c_colourpalette ) ;
						if (transparentSte)
						{
							++m_SpritePixelCount;
							m_SpriteLineData[c_x] = spriteIndex;
							if (m_SpritePixelCount >= c_maxpixels)
								return;
							continue;
						}
					}

					if (m_SpriteLineData[c_x] != c_maxInt)
						m_StatusRegister = BitSet(m_StatusRegister, 5);		

					if (m_SpriteLineData[c_x] > spriteIndex)
					{
						++m_SpritePixelCount;
						if ( ((c_red == 0xe0) && (c_green == 0x0) && (c_blue == 0xe0)) || (colourindex == 0) || transparentSte)
							continue ;
						// think (m_SpriteMaskData.second.first != m_SpriteMask2) is wrong 
						if ((m_VCounter >= m_SpriteMaskData.second.first) && (m_VCounter <= m_SpriteMaskData.second.second) && (spriteIndex > m_SpriteMaskData.first) && ( (m_SpriteMask2 == 0xFF) || m_SpriteMaskData.second.first == m_SpriteMask2))
							continue;

						m_SpriteLineData[c_x] = spriteIndex;
						
						m_DisplayAdapter.WriteToScreen( c_x, vCounter, c_red, c_green, c_blue ) ;
						if (m_SpritePixelCount >= c_maxpixels)
							return;
					}
				}
				xPos += 8;
			}
		}
	} while (nextsprite != 0 && m_MaxSprites > 0);
}

//////////////////////////////////////////////////////////////////////

void YM7101::RenderWindow( const UWORD baseAddress, bool priorityhigh )
{
	const int c_vCounter = m_VCounter ;

	const std::pair<int,int> c_screenres = GetScreenResolution() ;

	int hSize = 32;
	int vSize = 32;
	const int c_numTiles = c_screenres.first / 8;
	if (c_numTiles == 40)
		hSize = 64;

    const bool c_hRight = TestBit(m_Registers[0x11],7) ;
	const UBYTE c_hPos = (m_Registers[0x11] & 0x1F) * 2;
	const bool c_vDown = !TestBit(m_Registers[0x12],7) ;
	const UBYTE c_vPos = (m_Registers[0x12] & 0x1F) * 8;

	if (c_hPos == 0 && c_vPos == 0)
		return;

	// a row is 8 pixels large
	const int c_tilerow = c_vCounter / 8 ;

	const int c_startRow = c_vDown ? 0 : c_vPos ;
	const int c_endRow = c_vDown ? c_vPos : (vSize* 8) ;
	const bool c_inVertArea = (c_vCounter >= c_startRow) && (c_vCounter < c_endRow);

	int startCol = c_hRight ? c_hPos : 0 ;
	int endCol = c_hRight ? c_numTiles : c_hPos ;

    if (c_inVertArea)
    {
        startCol = 0 ;
        endCol = c_numTiles;
    }

	// draw all 32 columns for this scanline (we only actually draw a pixel of the current column which is the scanline)
	for (int tilecolumn = startCol; tilecolumn < endCol; ++tilecolumn)
	{
		// draw all 8 pixels in the column
		for (int xAdvance = 0 ; xAdvance < 8; ++xAdvance)
		{
			// get correct position in memory for this tile
			UWORD nameBaseOffset = baseAddress ;
			nameBaseOffset += c_tilerow * (hSize * 2) ; //each scanline has 32 tiles (1 tile per column) but 1 tile is 2 bytes in memory
			nameBaseOffset += tilecolumn * 2 ; // each tile is two bytes in memory

			const bool c_hFlip = TestBit(m_VRAM.vram8[nameBaseOffset], 3) ;
			const bool c_vFlip = TestBit(m_VRAM.vram8[nameBaseOffset], 4);

			const int c_tilexpixel = ( c_hFlip ? (7 - xAdvance) : xAdvance);
			const int c_tileypixel = (c_vFlip ? (7 - (c_vCounter % 8)) : (c_vCounter % 8));

			const int c_absx = (tilecolumn * 8) + xAdvance ;

			// check if correct priority
			if (TestBit(m_VRAM.vram8[nameBaseOffset], 7) != priorityhigh)
			{
				continue ;
			}

			// get tile information data from memory
			UWORD tileData = (m_VRAM.vram8[nameBaseOffset] & 0x7) << 8 ;
			tileData |= m_VRAM.vram8[nameBaseOffset+1] ;
			tileData *= 32 ;

			// each pattern line if 4 bytes. Offset contains correct pattern line
			tileData += 4 * c_tileypixel;

			// get pattern line data
			UBYTE data1 = m_VRAM.vram8[tileData] ;
			UBYTE data2 = m_VRAM.vram8[tileData+1] ;
			UBYTE data3 = m_VRAM.vram8[tileData+2] ;
			UBYTE data4 = m_VRAM.vram8[tileData+3] ;

			if (c_absx >= c_screenres.first || (IsRegBitSet(0,5 && c_absx < 8)))
			{
				continue ;
			}

			int colourindex = 0 ;
			// quick hack

			switch(c_tilexpixel)
			{
			case 0 : colourindex = data1 >> 4 ; break ;
			case 1 : colourindex = data1 & 0xF ; break ;
			case 2 : colourindex = data2 >> 4 ; break ;
			case 3 : colourindex = data2 & 0xF ; break ;
			case 4 : colourindex = data3 >> 4 ; break ;
			case 5 : colourindex = data3 & 0xF ; break ;
			case 6 : colourindex = data4 >> 4 ; break ;
			case 7 : colourindex = data4 & 0xF ; break ;
			}


			const UBYTE c_colourpalette = ( m_VRAM.vram8[nameBaseOffset] >> 5 ) & 0x3 ;

 			if (IsRegBitSet(0xc,3))
 			{
 				ApplySTE( priorityhigh, c_absx, false, colourindex, c_colourpalette ) ;
 			}

			if (colourindex == 0)
				continue;


			const int c_colourIndex = (16 * c_colourpalette) + colourindex;

			const UWORD c_colour = m_CRAM[c_colourIndex] ;

			const UBYTE c_blue = GetColourShade( (c_colour >> 8) & 0xF ) ;
			const UBYTE c_red =  GetColourShade( c_colour & 0xF ) ;
			const UBYTE c_green =  GetColourShade( ( c_colour >> 4 ) & 0xF ) ;

	//		m_LineHighlightData[c_absx] = DisplayAdapter::STE_NORMAL;

			// we now have all the data needed to colour this pixel. So write it to the screen
			m_DisplayAdapter.WriteToScreen(c_absx, c_vCounter, c_red, c_green, c_blue);
		}
	}
}

//////////////////////////////////////////////////////////////////////

void YM7101::DumpVRAM( ) const
{
	using namespace std;
	ofstream outputFile( "YM7101.ram", ofstream::binary );
	for (int i = 0 ; i < 0x10000; ++i)
	{
		if ((i % 2) == 0)
			outputFile << m_VRAM.vram8[i+1] ;
		else
			outputFile << m_VRAM.vram8[i-1] ;
	}
	outputFile.close() ;
}
