#include "Config_cs_core.h"
#include "LoggingManager.h"
#include "Logfile.h"
#include "Logging.h"
#include "../Utils/Throw.h"
#include <boost/format.hpp>
#include <boost/date_time.hpp>
#include <boost/assign/std/Set.hpp>
#include <cs_core/IniFiles/IniFiles.h>
using namespace boost::assign;

namespace codeslinger
{
	namespace logging
	{
		////////////////////////////////////////////////////////////////

		LoggingManager::LoggingManager( )
			: m_enabled(true)
		{
			m_sublayers += "info", "wrn", "err";
			AddNewGroup("Main");

			boost::shared_ptr<Logfile> logfileAll = boost::shared_ptr<Logfile>(
				new Logfile( "All_combined" ) );
			AddLogFile("All", "combined", logfileAll);
			m_enabled = codeslinger::inifile::IniFileReturnDefault<bool>("LoggingEnabled", true);
		}

		////////////////////////////////////////////////////////////////

		LoggingManager::~LoggingManager( )
		{
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::AddNewGroup(const std::string& name)
		{
			boost::shared_ptr<Logfile> logfileCombined = boost::shared_ptr<Logfile>(
				new Logfile( (boost::format("%1%_combined") % name ).str()));
			AddLogFile(  name, "combined", logfileCombined ) ;

			boost::shared_ptr<Logfile> logfileGroup = boost::shared_ptr<Logfile>(
				new Logfile( (boost::format("%1%_group") % name ).str()));
			AddLogFile(  name, "group", logfileGroup ) ;

			for (std::set<std::string>::const_iterator it = m_sublayers.begin(), 
				itEnd = m_sublayers.end(); it != itEnd; ++it)
			{
				boost::shared_ptr<Logfile> logfile = boost::shared_ptr<Logfile>(
					new Logfile( (boost::format("%1%_%2%") % name % (*it)).str()));
				AddLogFile(  name, *it, logfile ) ;
				logfile->AddDestination(logfileCombined);
			}
			m_groups.insert(name);
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::AddLogFile( const std::string& groupName, 
			const std::string& subsetName, const boost::shared_ptr<Logfile>& logfile )
		{
			const std::string c_name = (boost::format("%1%_%2%") % groupName % subsetName).str();
			if (m_LogfileMap.find( c_name ) == m_LogfileMap.end())
			{
				m_LogfileMap.insert(std::make_pair(c_name, logfile)) ;
			}
			else
			{
				THROW(( boost::format("Error: Log file already in loggingmanager"\
					"map with unique id %1%") % c_name ).str() ) ;
			}
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::LogMessage(const std::string& groupName, 
			const std::string& subsetName, int level, const std::string& message)
		{
			if (m_enabled)
			{
				const std::string c_name = (boost::format("%1%_%2%") % groupName % subsetName).str();
				LOG_FILE_MAP::const_iterator cit = m_LogfileMap.find( c_name ) ;

				if (cit != m_LogfileMap.end())
				{
					const boost::shared_ptr<Logfile>& logfile = (*cit).second ;
					if (logfile->IsEnabled() && logfile->GetLevel() >= level)
					{
						std::string msg = ( boost::format("%1%[%2% %3%]: %4%") % GetTimeStamp() 
							% logfile->GetName() % level % message ).str() ;
						(*cit).second->LogMessage(msg) ;
						GetLogfile(groupName, "group")->LogMessage(msg);
						GetLogfile("All", "combined")->LogMessage(msg);
					}
				}
				else
				{
					THROW( ( boost::format("Error: Cannot log message %1% to log file id %2% as "\
						"this is not in the map") % message % c_name  ).str() ) ;
				}
			}
		}

		////////////////////////////////////////////////////////////////

		boost::shared_ptr<Logfile> LoggingManager::GetLogfile( const std::string& groupName,
			const std::string& subsetName) const
		{
			const std::string c_name = (boost::format("%1%_%2%") % groupName % subsetName).str();
			boost::shared_ptr<Logfile> res ;

			LOG_FILE_MAP::const_iterator cit = m_LogfileMap.find( c_name ) ;
			
			if (cit != m_LogfileMap.end())
			{
				res = (*cit).second ;
			}
			else
			{
				THROW( ( boost::format("Error: Cannot find log file with id %1%") 
					% c_name  ).str() ) ;
			}
			return res ;
		}

		////////////////////////////////////////////////////////////////

		std::string LoggingManager::GetTimeStamp( ) const
		{
			const boost::posix_time::ptime now=
				boost::posix_time::microsec_clock::local_time();

			boost::posix_time::time_facet*const f=
				new boost::posix_time::time_facet("%d-%m-%Y [%T.%f]");


			std::ostringstream msg;
			msg.imbue(std::locale(msg.getloc(),f));
			msg << now;
			return msg.str();
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::MergeGroupSubsets( const std::string& source, 
			const std::string& destination )
		{
			std::set<std::string> layers = m_sublayers;
			for ( std::set<std::string>::const_iterator it = layers.begin(),
				itEnd = layers.end(); it != itEnd; ++it)
			{
				const boost::shared_ptr<Logfile>& c_source = GetLogfile(source, *it);
				const boost::shared_ptr<Logfile>& c_dest = GetLogfile(destination, *it);
				c_source->AddDestination(c_dest);
			}
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::AddNewSubset( const std::string& subset )
		{
			for ( std::set<std::string>::const_iterator it = m_groups.begin(),
				itEnd = m_groups.end(); it != itEnd; ++it)
			{
				boost::shared_ptr<Logfile> logfileCombined = GetLogfile(*it, "combined");
				boost::shared_ptr<Logfile> logfile = boost::shared_ptr<Logfile>(
					new Logfile( (boost::format("%1%_%2%") % (*it) % subset).str()));
				AddLogFile( *it, subset, logfile ) ;
				logfile->AddDestination(logfileCombined);
			}
			m_sublayers += subset;
		}

		////////////////////////////////////////////////////////////////

		void LoggingManager::SetEnabled( bool enabled )
		{
			m_enabled = enabled;
		}

		////////////////////////////////////////////////////////////////

		bool LoggingManager::GetEnabled() const
		{
			return m_enabled;
		}

		////////////////////////////////////////////////////////////////
	};
};