#include "Config_M68k.h"
#include "M68K.h"
#include <cs_core/Logging/Logging.h>
#include <cs_core/Utils/Maths.h>
#include <cs_core/Utils/Debugging.h>
#include <boost/format.hpp>

using codeslinger::logging::LogMessage ;
using codeslinger::maths::BitConditionalSet ;
using codeslinger::maths::BitSet ;
using codeslinger::maths::BitSetAll ;
using codeslinger::maths::BitReset ;
using codeslinger::maths::TestBit ;
using codeslinger::maths::BitCount ;
using codeslinger::maths::BitRotateLeft ;
using codeslinger::maths::BitRotateRight ;

//  Things to check
//  When in super user mode (an interupt) and the interupt handler intentional removes super user mode (updates SR and removes superuser bit) is it supposed to exit interrupt handler back to game code?
//  Interrupts like divide by 0, trace (STOP uses trace I think)


void M68K::OpcodeInvalid()
{
	const UWORD c_opcode = m_FuncReadMemory16( m_ProgramCounterStart ) ;

	LogMessage( "M68k","err", ( boost::format( "Processing invalid or unimplemented opcode 0x%1$x at address 0x%2$x" )
		% c_opcode % m_ProgramCounterStart ).str() ) ;

	ASSERT(false) ;
}

void M68K::OpcodeADD(UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	case 4: ds = SIZE_BYTE ; break ;
	case 5: ds = SIZE_WORD ; break ;
	case 6: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeAdd data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	const bool c_direction = opmode < 4 ;
	const bool c_extend = mode == EA_ADDRESS_REG && ds == SIZE_WORD;
	EA_DATA getdata = GetEAOperand(mode, eareg, ds, !c_direction && !c_extend) ;
	CheckAndChangeCAdd(getdata.operand,  m_RegistersData[reg], ds, false);
	CheckAndChangeVAdd(getdata.operand,  m_RegistersData[reg], ds, false);
	UDWORD result = getdata.operand + m_RegistersData[reg] ;

	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);

	if (c_direction)
	{
		SetDataRegistry(reg, result, ds);
		switch(ds)
		{
			case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break;  // long shorter than word due to word sign extending? makes sense
			default: m_CyclesThisUpdate += 4; break;
		}
	}
	else
	{
		if (c_extend)
		{
			result = SignExtend(UWORD(getdata.operand)) + SignExtend(UWORD(m_RegistersData[reg])) ;
			SetAddressRegistry(eareg, result, SIZE_WORD) ;
			switch(ds)
			{
			case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break;  // long shorter than word due to word sign extending? makes sense
			default: m_CyclesThisUpdate += 8; break;
			}
			ds = SIZE_LONG;
		}
		else
		{
			SetEAOperand(mode, eareg, result, ds) ;
			switch(ds)
			{
			case SIZE_LONG: m_CyclesThisUpdate += 12; break;
			default: m_CyclesThisUpdate += 8; break;
			}
		}
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above
}

void M68K::OpcodeAND(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 0 :
	case 4 : ds = SIZE_BYTE; break ;
	case 1 :
	case 5 : ds = SIZE_WORD;  break ;
	default : ds = SIZE_LONG; break ;
	}

	const bool c_memtodatareg = opmode < 4 ;
	EA_DATA getdata = GetEAOperand(mode, reg, ds,!c_memtodatareg) ;
	UDWORD result = 0 ;

	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);


	if (c_memtodatareg)
	{
		result = getdata.operand & m_RegistersData[datareg] ;
		SetDataRegistry(datareg,result,ds);
		switch(ds)
		{
		case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
		default: m_CyclesThisUpdate += 4; break;
		}
	}
	else
	{
		result = m_RegistersData[datareg] & getdata.operand ;
		SetEAOperand( mode, reg, result, ds ) ;
		switch(ds)
		{
		case SIZE_LONG: m_CyclesThisUpdate += 12; break;
		default: m_CyclesThisUpdate += 8; break;
		}
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
}

void M68K::OpcodeCMP( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 0: ds = SIZE_BYTE ; m_CyclesThisUpdate += 4; break ;
	case 1: ds = SIZE_WORD ; m_CyclesThisUpdate += 4; break ;
	case 2: ds = SIZE_LONG ; m_CyclesThisUpdate += 6; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeCMP data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getdata = GetEAOperand( mode, eareg, ds, false) ;
	const UDWORD c_data = m_RegistersData[reg] ;
	const UDWORD c_result = c_data - getdata.operand ;

	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	CheckAndChangeCSub(c_data, getdata.operand, ds, false);
	CheckAndChangeVSub(c_data, getdata.operand, ds, false);

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeDIVS( UBYTE reg, UBYTE mode, UBYTE eareg )
{
	SDWORD sDataReg = m_RegistersData[reg];
	EA_DATA uSource = GetEAOperand(mode, eareg, SIZE_WORD, false);
	m_ProgramCounter += uSource.pcadvance;
	m_CyclesThisUpdate += uSource.cycles;
	m_CyclesThisUpdate += 158;

	if (SWORD(uSource.operand) == 0)
	{
		LogMessage("M68k", "err", "Opcode DIVS done divide by 0. Should do a TRAP whatever that is. Returning out of instruction");
		return;
	}
	// check for overflow
	bool hasOverflown = false;
	SDWORD sDiv = sDataReg / SWORD(uSource.operand);
	if (sDiv > 0x7FFF)
	{
		hasOverflown = true;
	}

	m_CCR = BitConditionalSet( hasOverflown, m_CCR, FLAG_V );

	if (!hasOverflown)
	{
		SWORD quotient = static_cast<SWORD>(sDiv);
		SWORD remainder = static_cast<SWORD>( sDataReg - (SDWORD(quotient) * SDWORD(SWORD(uSource.operand))));
		remainder = BitConditionalSet(TestBit(sDataReg, 31), remainder, 15);
		m_RegistersData[reg] = (UDWORD(remainder) << 16) | UDWORD(quotient);	
		m_CCR = BitReset(m_CCR, FLAG_C);
		m_CCR = BitConditionalSet( TestBit(quotient,15), m_CCR, FLAG_N ) ;
		m_CCR = BitConditionalSet( quotient == 0, m_CCR, FLAG_Z );
	}
}

void M68K::OpcodeDIVU( UBYTE reg, UBYTE mode, UBYTE eareg )
{
	UDWORD sDataReg = m_RegistersData[reg];
	EA_DATA uSource = GetEAOperand(mode, eareg, SIZE_WORD, false);
	m_ProgramCounter += uSource.pcadvance;
	m_CyclesThisUpdate += uSource.cycles;
	m_CyclesThisUpdate += 140;

	if (UWORD(uSource.operand) == 0)
	{
		LogMessage("M68k", "err", "Opcode DIVU done divide by 0. Should do a TRAP whatever that is. Returning out of instruction");
		return;
	}
	// check for overflow
	bool hasOverflown = false;
	UDWORD sDiv = sDataReg /UWORD(uSource.operand);
	if (sDiv > 0xFFFF)
	{
		hasOverflown = true;
	}

	m_CCR = BitConditionalSet( hasOverflown, m_CCR, FLAG_V );

	if (!hasOverflown)
	{
		UWORD quotient = static_cast<UWORD>(sDiv);
		UWORD remainder = static_cast<UWORD>( sDataReg - (UDWORD(quotient) * UDWORD(UWORD(uSource.operand))));
		m_RegistersData[reg] = (UDWORD(remainder) << 16) | UDWORD(quotient);	
		m_CCR = BitReset(m_CCR, FLAG_C);
		m_CCR = BitConditionalSet( TestBit(quotient,15), m_CCR, FLAG_N ) ;
		m_CCR = BitConditionalSet( quotient == 0, m_CCR, FLAG_Z );
	}
}

void M68K::OpcodeEOR( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 4: ds = SIZE_BYTE ; break ;
	case 5: ds = SIZE_WORD ; break ;
	case 6: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeEOR data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	const UDWORD c_sourceData = m_RegistersData[reg] ;
	EA_DATA getdata = GetEAOperand(mode, eareg, ds, true);
	UDWORD result = getdata.operand ^ c_sourceData  ;

	SetEAOperand(mode, eareg, result, ds) ;

	
	if (mode == EA_DATA_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8 : 4;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 12 : 8;
		m_ProgramCounter += getdata.pcadvance ;
	}


	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
}

void M68K::OpcodeMULSWORD( UBYTE reg, UBYTE mode, UBYTE eareg )
{
	EA_DATA getdata = GetEAOperand(mode, eareg, SIZE_WORD, false) ;
	getdata.operand = (SWORD)getdata.operand ;
	const SWORD c_dataReg = (SWORD)m_RegistersData[reg] ;
	const SDWORD c_result = (SWORD)c_dataReg * (SWORD)getdata.operand ;
	m_RegistersData[reg] = c_result ; // DOES THIS NEED ASSIGNRESULT FUNCTION WITH SIZE WORD?
	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
	m_CyclesThisUpdate += 70 ;

	CheckAndChangeZ(c_result, SIZE_WORD);
	CheckAndChangeN(c_result, SIZE_WORD);
	m_CCR = BitReset(m_CCR, FLAG_C) ;
	m_CCR = BitReset(m_CCR, FLAG_Z) ; // overflow only occurs on LONG
}


void M68K::OpcodeMULUWORD( UBYTE reg, UBYTE mode, UBYTE eareg )
{
	EA_DATA getdata = GetEAOperand(mode, eareg, SIZE_WORD, false) ;
	getdata.operand = (UWORD)getdata.operand ;
	const UWORD c_dataReg = (UWORD)m_RegistersData[reg] ;
	const UDWORD c_result = (UWORD)c_dataReg * (UWORD)getdata.operand ;
	m_RegistersData[reg] = c_result ;// DOES THIS NEED ASSIGNRESULT FUNCTION WITH SIZE WORD?
	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
	m_CyclesThisUpdate += 70 ;
	CheckAndChangeZ(c_result, SIZE_WORD);
	CheckAndChangeN(c_result, SIZE_WORD);
	m_CCR = BitReset(m_CCR, FLAG_C) ;
	m_CCR = BitReset(m_CCR, FLAG_Z) ; // overflow only occurs on LONG
}

void M68K::OpcodeOR( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	case 4: ds = SIZE_BYTE ; break ;
	case 5: ds = SIZE_WORD ; break ;
	case 6: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeOR data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	const bool c_direction = opmode < 4 ;

	EA_DATA getdata = GetEAOperand( mode, eareg, ds, !c_direction ) ;
	const UDWORD c_regdata = m_RegistersData[reg] ;
	
	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);


	UDWORD result = 0 ;

	if (c_direction)
	{
		result = getdata.operand | c_regdata ;
		SetDataRegistry(reg, result, ds);
		switch(ds)
		{
		case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
		default: m_CyclesThisUpdate += 4; break;
		}
	}
	else
	{
		result = c_regdata | getdata.operand ;
		SetEAOperand(mode, eareg, result, ds) ;
		switch(ds)
		{
		case SIZE_LONG: m_CyclesThisUpdate += 12; break; 
		default: m_CyclesThisUpdate += 8; break;
		}
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
}

void M68K::OpcodeSUB( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	case 4: ds = SIZE_BYTE ; break ;
	case 5: ds = SIZE_WORD ; break ;
	case 6: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeSub data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	const bool c_direction = opmode < 4 ;
	const bool c_extend = mode == EA_ADDRESS_REG && ds == SIZE_WORD;
	EA_DATA getdata = GetEAOperand(mode, eareg, ds, !c_direction && !c_extend) ;
	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);


	const UDWORD c_left = c_direction ? m_RegistersData[reg] : getdata.operand ;
	const UDWORD c_right = c_direction ? getdata.operand : m_RegistersData[reg];
	CheckAndChangeCSub( c_left, c_right, ds, false );
	CheckAndChangeVSub( c_left, c_right, ds, false );

	UDWORD result = c_left - c_right ;
	if (c_direction)
	{
		SetDataRegistry(reg, result, ds);
		switch(ds)
		{
		case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
		default: m_CyclesThisUpdate += 4; break;
		}
	}
	else
	{
		if (c_extend)
		{
			result = SignExtend(UWORD(c_left)) - SignExtend(UWORD(c_right)) ;
			SetAddressRegistry(eareg, result, SIZE_WORD) ;
			switch(ds)
			{
			case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
			default: m_CyclesThisUpdate += 8; break;
			}
			ds = SIZE_LONG;
		}
		else
		{
			SetEAOperand(mode, eareg, result, ds) ;
			switch(ds)
			{
			case SIZE_LONG: m_CyclesThisUpdate += 12; break; 
			default: m_CyclesThisUpdate += 8; break;
			}
		}
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above
}

void M68K::OpcodeADDA(UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 3: ds = SIZE_WORD ; break ;
	case 7: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeAdda data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getdata = GetEAOperand(mode, eareg, ds, false) ;
	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);

	switch(ds)
	{
	case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
	default: m_CyclesThisUpdate += 8; break;
	}

	if (ds == SIZE_WORD)
		getdata.operand = SignExtend(UWORD(getdata.operand));

	const UDWORD c_result = m_RegistersAddress[reg] + getdata.operand ;

	SetAddressRegistry(reg, c_result, ds); 

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeCMPA( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 3: ds = SIZE_WORD ; break ;
	case 7: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeSubA data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getdata = GetEAOperand(mode, eareg, ds, false) ;

	
	if (ds == SIZE_WORD)
	{
		getdata.operand = SignExtend( (UWORD) getdata.operand ) ;
		ds = SIZE_LONG;
	}

	const UDWORD c_left = m_RegistersAddress[reg] ;

	const UDWORD c_result = c_left - getdata.operand  ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles + 6 ;

	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	CheckAndChangeCSub(c_left, getdata.operand, ds, false);
	CheckAndChangeVSub(c_left, getdata.operand, ds, false);
}

void M68K::OpcodeSUBA( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	DataSize ds ;
	switch(opmode)
	{
	case 3: ds = SIZE_WORD ;break ;
	case 7: ds = SIZE_LONG ;break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeSubA data size mode %1% " )
			% opmode  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getdata = GetEAOperand(mode, eareg, ds,false) ;
	const bool c_cycleExtend = (getdata.eatype == EA_DATA_REG) || 
		(getdata.eatype == EA_ADDRESS_REG) || (getdata.mode7Type == EA_MODE_7_IMMEDIATE_DATA);

	switch(ds)
	{
	case SIZE_LONG: m_CyclesThisUpdate += c_cycleExtend ? 8 : 6; break; 
	default: m_CyclesThisUpdate += 8; break;
	}

	if (ds == SIZE_WORD)
	{
		getdata.operand = SignExtend( (UWORD) getdata.operand ) ;
	}

	const UDWORD c_left = m_RegistersAddress[reg] ;
	const UDWORD c_result = c_left - getdata.operand  ;
	SetAddressRegistry(reg, c_result, ds);

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeANDICR( )
{
	UWORD data = m_FuncReadMemory16(m_ProgramCounter);
	m_ProgramCounter +=2 ;
	UBYTE lowByteCCR = UBYTE(m_CCR);
	UBYTE lowByteImmediate = UBYTE(data);
	lowByteCCR &= lowByteImmediate;
	m_CCR &= 0xFF00;
	m_CCR |= lowByteCCR;
	m_CyclesThisUpdate += 20;
}

void M68K::OpcodeANDISR()
{
	//CHECK THIS WITH PC c76 on Golden Axe. It seems to cause either a TRAP or exit super user mode back to game code
	if (m_InSuperVisorMode)
	{
		UWORD data = m_FuncReadMemory16(m_ProgramCounter);
		m_ProgramCounter +=2 ;
		m_CCR = m_CCR & data;
		m_CyclesThisUpdate += 20;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}
}

void M68K::OpcodeCHK( UBYTE reg, UBYTE size, UBYTE mode, UBYTE eareg )
{
	LogMessage("M68k", "wrn", 0, "OpcodeCHK not tested");
	SWORD data = m_RegistersData[reg];
	EA_DATA eaData = GetEAOperand(mode, eareg, SIZE_WORD, false);
	m_ProgramCounter += eaData.pcadvance;
	m_CyclesThisUpdate += eaData.cycles;
	m_CyclesThisUpdate += 10;
	if (data < 0)
	{
		OpcodeTRAP(6);
		m_CCR = BitSet( m_CCR, FLAG_N);
		m_CyclesThisUpdate += 44;
	}
	else if (data > (SWORD)eaData.operand)
	{
		OpcodeTRAP(6);
		m_CCR = BitReset( m_CCR, FLAG_N);
	}
}

void M68K::OpcodeEORICCR()
{
	UBYTE data = m_FuncReadMemory8(m_ProgramCounter+1);
	m_ProgramCounter +=2 ;
	m_CCR ^= data;
	m_CyclesThisUpdate += 20;
}


void M68K::OpcodeEORISR()
{
	if (m_InSuperVisorMode)
	{
		UWORD data = m_FuncReadMemory16(m_ProgramCounter);
		m_ProgramCounter +=2 ;
		m_CCR ^= data;
		m_CyclesThisUpdate += 20;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}
}

void M68K::OpcodeORICR()
{
	UBYTE data = m_FuncReadMemory8(m_ProgramCounter+1);
	m_ProgramCounter +=2 ;
	m_CCR |= data;
	m_CyclesThisUpdate += 20;
}

void M68K::OpcodeORISR()
{
	if (m_InSuperVisorMode)
	{
		UWORD data = m_FuncReadMemory16(m_ProgramCounter);
		m_ProgramCounter +=2 ;
		m_CCR = m_CCR | data;
		m_CyclesThisUpdate += 20;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}
}

void M68K::OpcodeMoveSR( UBYTE mode, UBYTE reg )
{
	if (m_InSuperVisorMode)
	{
		EA_DATA data = SetEAOperand(mode,reg,m_CCR, SIZE_WORD);
		m_ProgramCounter += data.pcadvance;
		m_CyclesThisUpdate += data.cycles;
		m_CyclesThisUpdate += (mode == EA_DATA_REG) ? 6 : 8 ;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}	
}

void M68K::OpcodeMoveToCCR( UBYTE mode, UBYTE reg )
{
	EA_DATA data = GetEAOperand(mode, reg, SIZE_WORD,false);
	m_CCR &= 0xff00;
	m_CCR |= (UBYTE)data.operand;
	m_ProgramCounter += data.pcadvance;
	m_CyclesThisUpdate += data.cycles;
	m_CyclesThisUpdate += 12;
}

void M68K::OpcodeMoveToSR( UBYTE mode, UBYTE reg )
{
	if (m_InSuperVisorMode)
	{
		EA_DATA getdata = GetEAOperand( mode, reg, SIZE_WORD, false) ;
		m_CCR = (UWORD)getdata.operand ;
		m_CyclesThisUpdate += 12 ;
		m_ProgramCounter += getdata.pcadvance ;
		m_CyclesThisUpdate += getdata.cycles ;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}

}

void M68K::OpcodeEXG( UBYTE regx, UBYTE opmode, UBYTE regy )
{
	UDWORD temp = 0;
	if (opmode == 8)
	{
		temp = m_RegistersData[regx];
		m_RegistersData[regx] = m_RegistersData[regy];
		m_RegistersData[regy] = temp;
	}
	else if (opmode == 9)
	{
		temp = m_RegistersAddress[regx];
		m_RegistersAddress[regx] = m_RegistersAddress[regy];
		m_RegistersAddress[regy] = temp;
	}
	else if (opmode == 17)
	{
		temp = m_RegistersData[regx];
		m_RegistersData[regx] = m_RegistersAddress[regy];
		m_RegistersAddress[regy] = temp;
	}

	m_CyclesThisUpdate += 6;
}

void M68K::OpcodeEXT(UBYTE opmode, UBYTE reg )
{
	m_CyclesThisUpdate += 4 ;
	switch( opmode )
	{
	case 2:
		SetDataRegistry(reg, SignExtend( (UBYTE) m_RegistersData[reg] ), SIZE_WORD);
		break ;
	case 3:
		m_RegistersData[reg] = SignExtend( (UWORD) m_RegistersData[reg] ) ;
		break ;
	case 7:
		m_RegistersData[reg] = SignExtend( SignExtend( (UBYTE) m_RegistersData[reg] ) ) ;
		break ;
	}
	const UDWORD data = m_RegistersData[reg] ;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	m_CCR = BitReset(m_CCR, FLAG_C) ;
	CheckAndChangeZ(data, (opmode == 2) ? SIZE_WORD : SIZE_LONG);
	CheckAndChangeN(data, (opmode == 2) ? SIZE_WORD : SIZE_LONG);
}

void M68K::OpcodeLINKWORD( UBYTE reg )
{
	//	ASSERT(false) ; // test this especially due to address registers
	m_CyclesThisUpdate += 16 ;
	m_RegistersAddress[0x7] -= 4 ;
	m_FuncWriteMemory32( m_RegistersAddress[0x7], m_RegistersAddress[reg]) ;
	m_RegistersAddress[reg] = m_RegistersAddress[0x7] ;

	const UDWORD c_displacement = SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) ;
	m_ProgramCounter += 2;
	m_RegistersAddress[0x7] += c_displacement ;
}

void M68K::OpcodeLINKLONG( UBYTE reg )
{
	m_CyclesThisUpdate += 16 ;
	ASSERT(false) ; // test this especially due to address registers
	m_RegistersAddress[0x7] -= 4 ;
	m_FuncWriteMemory32( m_RegistersAddress[0x7], m_RegistersAddress[reg]) ;
	m_RegistersAddress[reg] = m_RegistersAddress[0x7] ;

	const UDWORD c_displacement = m_FuncReadMemory32(m_ProgramCounter)  ;
	m_ProgramCounter += 4;
	m_RegistersAddress[0x7] += c_displacement ;
}

void M68K::OpcodeMOVEUSP(UBYTE direction, UBYTE reg )
{
	m_CyclesThisUpdate += 4 ;

	if (m_InSuperVisorMode)
	{
		if (direction == 0)
		{
			m_USP = m_RegistersAddress[reg] ;
		}
		else
		{
			m_RegistersAddress[reg] = m_USP ;
		}
	}
}

void M68K::OpcodeNOP()
{
	m_CyclesThisUpdate += 4 ;
}

void M68K::OpcodeRTE()
{
	if (m_InSuperVisorMode)
	{
		m_CyclesThisUpdate += 20 ;
		m_CCR = m_FuncReadMemory16(m_RegistersAddress[0x7]);
		m_RegistersAddress[0x7] += 2 ;
		m_ProgramCounter = m_FuncReadMemory32(m_RegistersAddress[0x7]);
		m_RegistersAddress[0x7] += 4 ;
		m_servicingInterrupt = false;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}
}

void M68K::OpcodeRTR()
{
	LogMessage("M68k", "wrn", 0, "OpcodeRTR not tested");

	const UWORD c_newCCR = m_FuncReadMemory16(m_RegistersAddress[0x7]);
	m_CCR &= 0xFF00;
	m_CCR |= (UBYTE)c_newCCR;
	m_RegistersAddress[0x7] += 2;
	m_ProgramCounter = m_FuncReadMemory32(m_RegistersAddress[0x7]);
	m_RegistersAddress[0x7] += 4;
	m_CyclesThisUpdate += 20;
}

void M68K::OpcodeRTS()
{
	m_CyclesThisUpdate += 16 ;
	m_ProgramCounter = m_FuncReadMemory32(m_RegistersAddress[0x7]);
	m_RegistersAddress[0x7] += 4 ;
}

void M68K::OpcodeSTOP()
{
	if (m_InSuperVisorMode)
	{
		m_CCR = m_FuncReadMemory16(m_ProgramCounter);
		m_ProgramCounter += 2;
		m_Stop = true;
		m_CyclesThisUpdate += 4;
	}
	else
	{
		RequestInterrupt(INT_PRIV_VIO, m_FuncReadMemory32( 0x20 ) );
		m_CyclesThisUpdate += 34;
	}	
}

void M68K::OpcodeSWAP( UBYTE reg )
{
	const UDWORD c_data = m_RegistersData[reg] ;
	const UWORD c_low = UWORD(c_data & 0xFFFF) ;
	const UWORD c_hi = UWORD((c_data >> 16) & 0xFFFF) ;
	const UDWORD c_newData = ((UDWORD)c_low << 16) | c_hi ;
	m_RegistersData[reg] = c_newData ;
	m_CCR = BitConditionalSet( TestBit(c_newData, 31), m_CCR, FLAG_N ) ;
	m_CCR = BitReset(m_CCR, FLAG_C) ;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	m_CCR = BitConditionalSet( c_newData == 0, m_CCR, FLAG_Z ) ;
	m_CyclesThisUpdate += 4 ;
}



void M68K::OpcodeUNLK( UBYTE reg )
{
	m_CyclesThisUpdate += 12 ;
	m_RegistersAddress[0x7] = m_RegistersAddress[reg] ;
	m_RegistersAddress[reg] = m_FuncReadMemory32(m_RegistersAddress[0x7]) ;
	m_RegistersAddress[0x7] += 4 ;
}

void M68K::OpcodeBcc(UBYTE condition, UBYTE displacement)
{
	int increaseAmount = 0 ;

	UDWORD displacementToUse = SignExtend( SignExtend( displacement) ) ; ;

	if (displacement == 0 )
	{
		UWORD word = m_FuncReadMemory16( m_ProgramCounter );
		increaseAmount = 2 ;
		displacementToUse = SignExtend(word) ;
		m_CyclesThisUpdate += 4 ;
	}
	else if (displacement == 0xFF)
	{
		displacementToUse = m_FuncReadMemory32( m_ProgramCounter );
		increaseAmount =4 ;
		m_CyclesThisUpdate += 4 ;
	}

	if (ConditionTest(condition))
	{
		m_CyclesThisUpdate += 10 ;
		m_ProgramCounter += displacementToUse;
	}
	else
	{
		m_CyclesThisUpdate += 8 ;
		m_ProgramCounter += increaseAmount ;
	}
}

void M68K::OpcodeBSR( UBYTE displacement)
{
	m_CyclesThisUpdate += 18 ;
	m_RegistersAddress[0x7] -= 4 ;

	int pcAdvance = 0;
	if (displacement == 0)
		pcAdvance = 2;
	else if (displacement == 0xFF)
		pcAdvance = 4;
	m_FuncWriteMemory32( m_RegistersAddress[0x7], m_ProgramCounter+pcAdvance) ; // make sure program counter points to next inst

	UDWORD displacementToUse = SignExtend( SignExtend( displacement) ) ; ;

	if (displacement == 0 )
	{
		UWORD word = m_FuncReadMemory16( m_ProgramCounter );
		displacementToUse = SignExtend(word) ;
	}
	else if (displacement == 0xFF) 
	{
		displacementToUse = m_FuncReadMemory32( m_ProgramCounter );
	}


	m_ProgramCounter += displacementToUse;
}

void M68K::OpcodeDBcc( UBYTE condition, UBYTE reg )
{
	if (!ConditionTest(condition))
	{
		const bool c_moveon = UWORD(m_RegistersData[reg]) == 0 ;
		UWORD lowWord = UWORD(m_RegistersData[reg]) - 1;
		UWORD highWord = UWORD(m_RegistersData[reg] >> 16) ;
		m_RegistersData[reg] = (highWord << 16) | lowWord;
		const UDWORD c_displacement = SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) ;
		if (!c_moveon)
		{
			m_CyclesThisUpdate += 10 ;
			m_ProgramCounter += c_displacement ;
		}
		else
		{
			m_RegistersData[reg] &= 0xFFFF0000;
			m_RegistersData[reg] |= 0xFFFF ;
			m_CyclesThisUpdate += 14 ;
			m_ProgramCounter += 2 ;
		}
	}
	else
	{
		m_ProgramCounter += 2 ;
		m_CyclesThisUpdate += 12 ;
	}
}

void M68K::OpcodeADDX( UBYTE regx, UBYTE size, UBYTE rm, UBYTE regy )
{
	DataSize ds;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default: break;
	}

	const UDWORD c_xflag = TestBit(m_CCR, FLAG_X) ? 1 : 0;
	if (rm==0)
	{
		// data
		const UDWORD c_x = m_RegistersData[regx];
		const UDWORD c_y = m_RegistersData[regy] ;
		const UDWORD result = c_x+c_y+c_xflag;
		SetDataRegistry(regx,result,ds);
		CheckAndChangeZ( result, ds );
		CheckAndChangeN( result, ds );
		CheckAndChangeVAdd( c_x, c_y, ds, true );
		CheckAndChangeCAdd( c_x, c_y, ds, true );
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8 : 4;		
	}
	else
	{
		ASSERT(false); // not tested this
		// ea
		const EA_DATA getea = GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regy, ds, false, 0);
		const EA_DATA geteaTemp = GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, ds, true, 0);
		const UWORD result = getea.operand + geteaTemp.operand + c_xflag;
		m_ProgramCounter += getea.pcadvance;
		const EA_DATA data = SetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, result, ds);
		m_ProgramCounter += data.pcadvance;
		
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 30 : 18;

		CheckAndChangeZ( result, ds );
		CheckAndChangeN( result, ds );
		CheckAndChangeVAdd( getea.operand, geteaTemp.operand, ds, true );
		CheckAndChangeCAdd( getea.operand, geteaTemp.operand, ds, true );
	}
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
}

void M68K::OpcodeCMPM( UBYTE regAX, UBYTE size, UBYTE regAY )
{
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; m_CyclesThisUpdate += 12; break ;
	case 1: ds = SIZE_WORD ; m_CyclesThisUpdate += 12; break ;
	case 2: ds = SIZE_LONG ; m_CyclesThisUpdate += 20; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeCMPM data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getData1 = GetEAOperand(EA_ADDRESS_REG_INDIRECT_POST_INC, regAX, ds, false);
	EA_DATA getData2 = GetEAOperand(EA_ADDRESS_REG_INDIRECT_POST_INC, regAY, ds, false);

	const UDWORD c_result = getData1.operand - getData2.operand ;

	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	CheckAndChangeCSub(getData1.operand, getData2.operand, ds, false);
	CheckAndChangeVSub(getData1.operand, getData2.operand, ds, false);

	m_ProgramCounter += getData1.pcadvance;
}

void M68K::OpcodeSUBX( UBYTE regy, UBYTE size, UBYTE rm, UBYTE regx )
{
	LogMessage("M68k", "wrn", 0, "OpcodeSUBX not tested");

	DataSize ds;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default: break;
	}

	const UDWORD c_xflag = TestBit(m_CCR, FLAG_X) ? 1 : 0;
	if (rm==0)
	{
		// data
		const UDWORD c_x = m_RegistersData[regx];
		const UDWORD c_y = m_RegistersData[regy] ;
		const UDWORD result = (c_y-c_x)-c_xflag;
		SetDataRegistry(regy,result,ds);
		CheckAndChangeZ( result, ds );
		CheckAndChangeN( result, ds );
		CheckAndChangeVSub( c_y, c_x, ds, true );
		CheckAndChangeCSub( c_y, c_x, ds, true );
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8 : 4;	
	}
	else
	{
		ASSERT(false); // not tested this
		// ea
		const EA_DATA getea = GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regy, ds, false, 0);
		const EA_DATA geteaTemp = GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, ds, true, 0);
		const UWORD result = (getea.operand - geteaTemp.operand) - c_xflag;
		m_ProgramCounter += getea.pcadvance;
		const EA_DATA data = SetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regy, result, ds);
		m_ProgramCounter += data.pcadvance;
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 30 : 18;	

		CheckAndChangeZ( result, ds );
		CheckAndChangeN( result, ds );
		CheckAndChangeVSub( getea.operand, geteaTemp.operand, ds, true );
		CheckAndChangeCSub( getea.operand, geteaTemp.operand, ds, true );
	}
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
}

void M68K::OpcodeABCD( UBYTE regx, UBYTE rm, UBYTE regy )
{
	UBYTE source = 0 ;
	UBYTE dest = 0 ;
	if (rm == 0)
	{
		dest = (UBYTE)m_RegistersData[regx];
		source = (UBYTE)m_RegistersData[regy];
	}
	else
	{
		dest = (UBYTE)GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, SIZE_BYTE, true).operand;
		source = (UBYTE)GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regy, SIZE_BYTE, false).operand;
	}

	const UBYTE c_x = TestBit( m_CCR, FLAG_X ) ? 1 : 0 ;
	UWORD result = dest + source + c_x ;

	if (((dest & 0xF) +(source & 0xF) + c_x) > 9) // bcd overflow correction
		result += 0x6;
	if ( (result >> 4) > 9) // bcd tens overflow correction
		result += 0x60;

	if (result > 0x99)
		m_CCR = BitSet(m_CCR, FLAG_C);
	else
		m_CCR = BitReset(m_CCR, FLAG_C);

	m_CCR = BitConditionalSet(TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X) ;
	if ((result & 0xFF) != 0)
		m_CCR = BitReset(m_CCR, FLAG_Z);


	if (rm == 0)
	{
		SetDataRegistry(regx, (UBYTE)result, SIZE_BYTE);
		m_CyclesThisUpdate += 6;
	}
	else
	{
		EA_DATA data = SetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, (UBYTE)result,SIZE_BYTE);
		m_CyclesThisUpdate += 18;// guess
	}
}

void M68K::OpcodeSBCD( UBYTE regy, UBYTE rm, UBYTE regx )
{
	UBYTE source = 0 ;
	UBYTE dest = 0 ;
	if (rm == 0)
	{
	 	dest = (UBYTE)m_RegistersData[regy];
	 	source = (UBYTE)m_RegistersData[regx];
	}
	else
	{
	 	dest = (UBYTE)GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regy, SIZE_BYTE, true).operand;
	 	source = (UBYTE)GetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, SIZE_BYTE, false).operand;
	}

	const UBYTE c_x = TestBit( m_CCR, FLAG_X ) ? 1 : 0 ;

	UDWORD res = (dest & 0xF) - (source & 0xF) - c_x;

	if(res > 9)
		res -= 6;
	
	res += (dest & 0xF0 ) - (source & 0xF0);
	const bool c_carry = res > 0x99;
	m_CCR = BitConditionalSet(c_carry, m_CCR, FLAG_C);
	m_CCR = BitConditionalSet(c_carry, m_CCR, FLAG_X);
	
	if(c_carry)
		res += 0xa0;
	
	if ((res & 0xFF) != 0)
		m_CCR = BitReset(m_CCR, FLAG_Z);

	if (rm == 0)
	{
		SetDataRegistry(regy, (UBYTE)res, SIZE_BYTE);
		m_CyclesThisUpdate += 6;
	}
	else
	{
		EA_DATA data = SetEAOperand(EA_ADDRESS_REG_INDIRECT_PRE_DEC, regx, (UBYTE)res,SIZE_BYTE);
		m_CyclesThisUpdate += data.cycles;
		m_CyclesThisUpdate += 18;
	}
}


void M68K::OpcodeCLR( UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeCLR data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}
	EA_DATA data = SetEAOperand(mode, reg,0,ds) ;
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 6 : 4 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 12 : 8 ;
		m_CyclesThisUpdate += data.cycles ;
	}
	m_ProgramCounter += data.pcadvance ;
	m_CCR = BitReset(m_CCR, FLAG_N) ;
	m_CCR = BitSet(m_CCR, FLAG_Z) ;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	m_CCR = BitReset(m_CCR, FLAG_C) ;
}

void M68K::OpcodeNBCD( UBYTE mode, UBYTE reg )
{
	UBYTE source = 0 ;
	UBYTE dest = (UBYTE)GetEAOperand(mode, reg, SIZE_BYTE, true).operand ;

	m_CCR = BitConditionalSet(dest != 0, m_CCR, FLAG_C);
	--dest;


	const UBYTE c_x = TestBit( m_CCR, FLAG_X ) ? 1 : 0 ;

	if (c_x > 0) //
	{
		++dest;
	}

	if (dest > 0x99) // really dont like this, it assumes developer hasnt messed up bcd and is mainly there to fix adding c_x to 0x99
		dest = 0x99;

	dest = 0x99 - dest;

	UWORD result = dest + source ;

	m_CCR = BitConditionalSet(TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X) ;
	if ((result & 0xFF) != 0)
		m_CCR = BitReset(m_CCR, FLAG_Z);


	EA_DATA data = SetEAOperand(mode, reg, (UBYTE)result,SIZE_BYTE);
	
	
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += 4;
	}
	else
	{
		m_CyclesThisUpdate += data.cycles + 8;
	}
}

void M68K::OpcodeNEG(UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	int testBit = 0 ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; testBit = 7;break ;
	case 1: ds = SIZE_WORD ; testBit = 15;break ;
	case 2: ds = SIZE_LONG ; testBit = 31; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeNEG data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getData = GetEAOperand(mode, reg, ds, true);
	SDWORD original = static_cast<SDWORD>(getData.operand);
	CheckAndChangeCSub(0,original, ds, false);
	CheckAndChangeVSub(0, original, ds, false);
	getData.operand = 0 - original;
	EA_DATA setData = SetEAOperand(mode,reg,getData.operand, ds);
	m_ProgramCounter += getData.pcadvance;
	m_CyclesThisUpdate += getData.cycles;
	
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 6 : 4;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8: 12;
	}

	SDWORD sign = static_cast<SDWORD>( getData.operand );
	m_CCR = BitConditionalSet(TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X);
	CheckAndChangeN(getData.operand,ds);

	bool oldZ = TestBit(m_CCR, FLAG_Z);
	CheckAndChangeZ(getData.operand, ds);
	bool newZ = TestBit(m_CCR, FLAG_Z);
	if (newZ && !oldZ)
	{
		m_CCR = BitReset(m_CCR, FLAG_Z);
	}
}

void M68K::OpcodeNEGX( UBYTE size, UBYTE mode, UBYTE reg )
{
	LogMessage("M68k", "wrn", 0, "OpcodeNEGX not tested");

	DataSize ds;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default: break;
	}

	const UDWORD c_xflag = TestBit(m_CCR, FLAG_X) ? 1 : 0;
	const EA_DATA getea = GetEAOperand(mode,reg, ds, true, 0);
	const UDWORD result = (0 - getea.operand) - c_xflag;
	const EA_DATA data = SetEAOperand(mode, reg, result, ds);
	m_ProgramCounter += data.pcadvance;
	m_CyclesThisUpdate += data.cycles;
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 6 : 4;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8: 12;
	}

	CheckAndChangeZ( result, ds );
	CheckAndChangeN( result, ds );
	CheckAndChangeVSub( 0, getea.operand, ds, true );
	CheckAndChangeCSub( 0, getea.operand, ds, true );


	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
}

void M68K::OpcodeNOT( UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeNOT data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}
	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;

	UDWORD dataToSet = 0 ;
	switch( size )
	{
	case 0:
		{
			UBYTE data = (UBYTE)getdata.operand ;
			data = ~data ;
			dataToSet = data ;
		}			
		break ;
	case 1: 
		{
			UWORD data = (UWORD)getdata.operand ;
			data = ~data ;
			dataToSet = data ;
		}
		break ;
	case 2: 
		{
			UDWORD data = getdata.operand ;
			data = ~data ;
			dataToSet = data ;
		}break ;
	}
	SetEAOperand(mode,reg, dataToSet, ds) ;
	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 6 : 4;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 8: 12;
	}

	CheckAndChangeZ(dataToSet, ds);
	CheckAndChangeN(dataToSet, ds);
	m_CCR = BitReset(m_CCR, FLAG_C) ;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
}

void M68K::OpcodeScc( UBYTE cc, UBYTE mode, UBYTE reg )
{
	UBYTE data = 0;
	bool ccTrue = false;
	if (ConditionTest(cc))
	{
		ccTrue = true;
		data = codeslinger::maths::BitSetAll(8);
	}
	EA_DATA eadata = SetEAOperand(mode, reg, data, SIZE_BYTE);
	m_ProgramCounter += eadata.pcadvance;
	m_CyclesThisUpdate += eadata.cycles;

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += ccTrue ? 6 : 4;
	}
	else
	{
		m_CyclesThisUpdate += 8 ;
	}
}

void M68K::OpcodeTAS( UBYTE mode, UBYTE reg )
{
	LogMessage("M68k", "wrn", 0, "OpcodeTAS not tested");
	EA_DATA data = GetEAOperand(mode, reg, SIZE_BYTE, true);
	m_CCR = BitReset(m_CCR, FLAG_C);
	m_CCR = BitReset(m_CCR, FLAG_V);
	CheckAndChangeZ( data.operand, SIZE_BYTE );
	CheckAndChangeN( data.operand, SIZE_BYTE );
	data.operand = BitSet(data.operand, 7);
	EA_DATA setdata = GetEAOperand( mode, reg, SIZE_BYTE, false);
	m_ProgramCounter += setdata.pcadvance;
	m_CyclesThisUpdate += setdata.cycles;
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += 4;
	}
	else
	{
		m_CyclesThisUpdate += 10 ;
	}
}

void M68K::OpcodeTST(UBYTE size, UBYTE mode, UBYTE reg)
{
	m_CyclesThisUpdate += 4 ;
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeTST data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	EA_DATA getdata = GetEAOperand( mode, reg, ds, false) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(getdata.operand, ds);
	CheckAndChangeN(getdata.operand, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
}

void M68K::OpcodeASReg( UBYTE countreg, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg )// ASL and ASR to reg
{
	int shiftcount = 0 ;
	if (ir == 0)
	{
		shiftcount = countreg == 0 ? 8 : countreg ;
	}
	else
	{
		shiftcount = m_RegistersData[countreg] % 64 ;
	}

	m_CyclesThisUpdate += 6 + (2*shiftcount) ;

	SDWORD toShift = (SDWORD)m_RegistersData[reg] ; // signed makes it an arithmetical shift
	int numbits = 0 ;

	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; numbits = 8 ; toShift &= 0xFF; break ;
	case 1: ds = SIZE_WORD ; numbits = 16 ; toShift &= 0xFFFF; break ;
	case 2: ds = SIZE_LONG ; numbits = 32; m_CyclesThisUpdate+= 2; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeLSREG data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	SDWORD newData = 0 ;

	if (dir == 0) // shift right
	{
		const bool c_bitset = TestBit( toShift, shiftcount - 1) ;
		if (shiftcount == 0)
		{
			m_CCR = BitReset( m_CCR, FLAG_C ) ;
		}
		else
		{
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		}
		switch (ds)
		{
		case SIZE_BYTE:
			newData = SBYTE(toShift) >> SBYTE(shiftcount) ;
			break;
		case SIZE_WORD:
			newData = SWORD(toShift) >> SWORD(shiftcount) ;
			break;
		case SIZE_LONG:
			newData = SDWORD(toShift) >> SDWORD(shiftcount) ;
			break;
		}
	}
	else // shift left
	{
		const bool c_bitset = TestBit( toShift, numbits - shiftcount) ;
		if (shiftcount == 0)
		{
			m_CCR = BitReset( m_CCR, FLAG_C ) ;
		}
		else
		{
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		}
		switch (ds)
		{
		case SIZE_BYTE:
			newData = SBYTE(toShift) << SBYTE(shiftcount) ;
			break;
		case SIZE_WORD:
			newData = SWORD(toShift) << SWORD(shiftcount) ;
			break;
		case SIZE_LONG:
			newData = SDWORD(toShift) << SDWORD(shiftcount) ;
			break;
		}
	}
	SetDataRegistry(reg, (UDWORD)newData, ds); // TEST THIS!!, SHOULD IT BE A 32BIT ASSIGN INSTEAD?

	m_CCR = BitReset(m_CCR, FLAG_V) ;
	CheckAndChangeZ(newData, ds);
	CheckAndChangeN(newData, ds);
}

void M68K::OpcodeASMem( UBYTE dir, UBYTE mode, UBYTE reg ) // ASL and ASR to memory
{
	EA_DATA getData = GetEAOperand(mode, reg, SIZE_WORD, true, 0);
	SDWORD toShift = (SDWORD)getData.operand ; // signed makes it an arithmetical shift
	SDWORD newData = 0 ;

	if (dir == 0) // shift right
	{
		const bool c_bitset = TestBit( toShift, 0) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;		
		newData = SWORD(toShift) >> SWORD(1) ;			
	}
	else // shift left
	{
		const bool c_bitset = TestBit( toShift, 15) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		newData = SWORD(toShift) << SWORD(1) ;
	}
	EA_DATA setData = SetEAOperand(mode, reg, (UWORD)newData, SIZE_WORD, false);

	m_CyclesThisUpdate += 8;
	m_CyclesThisUpdate += setData.cycles;
	m_ProgramCounter += setData.pcadvance;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	CheckAndChangeZ(newData, SIZE_WORD);
	CheckAndChangeN(newData, SIZE_WORD);
}

void M68K::OpcodeLSReg( UBYTE countreg, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg ) // LSL and LSR to reg
{
	int shiftcount = 0 ;
	if (ir == 0)
	{
		shiftcount = countreg == 0 ? 8 : countreg ;
	}
	else
	{
		shiftcount = m_RegistersData[countreg] % 64 ;
	}

	m_CyclesThisUpdate += 6 + (2*shiftcount) ;

	UDWORD toShift = m_RegistersData[reg] ;
	int numbits = 0 ;

	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; numbits = 8 ; toShift &= 0xFF; break ;
	case 1: ds = SIZE_WORD ; numbits = 16 ; toShift &= 0xFFFF; break ;
	case 2: ds = SIZE_LONG ; numbits = 32; m_CyclesThisUpdate+= 2; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeLSREG data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	UDWORD newData = 0 ;

	if (dir == 0) // shift right
	{
		const bool c_bitset = TestBit( toShift, shiftcount - 1) ;
		if (shiftcount == 0)
		{
			m_CCR = BitReset( m_CCR, FLAG_C ) ;
		}
		else
		{
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		}
		newData = toShift >> shiftcount ;
	}
	else // shift left
	{
		const bool c_bitset = TestBit( toShift, numbits - shiftcount) ;
		if (shiftcount == 0)
		{
			m_CCR = BitReset( m_CCR, FLAG_C ) ;
		}
		else
		{
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
			m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		}
		newData = toShift << shiftcount ;
	}
	SetDataRegistry(reg, newData, ds); // TEST THIS!!, SHOULD IT BE A 32BIT ASSIGN INSTEAD?
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	CheckAndChangeZ(newData, ds);
	CheckAndChangeN(newData, ds);
}

void M68K::OpcodeLSMem( UBYTE dir, UBYTE mode, UBYTE reg ) // LSL and LSR to memory
{
	EA_DATA getData = GetEAOperand(mode, reg, SIZE_WORD, true);
	UDWORD toShift = getData.operand ;

	UDWORD newData = 0 ;

	if (dir == 0) // shift right
	{
		const bool c_bitset = TestBit( toShift, 0) ;		
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;		
		newData = toShift >> 1 ;
	}
	else // shift left
	{
		const bool c_bitset = TestBit( toShift, 15) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_X ) ;
		m_CCR = BitConditionalSet( c_bitset, m_CCR, FLAG_C ) ;
		newData = toShift << 1 ;
	}
	EA_DATA setData = SetEAOperand(mode, reg, newData, SIZE_WORD);
	m_CyclesThisUpdate += setData.cycles + 8;
	m_ProgramCounter += setData.pcadvance;
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	CheckAndChangeZ(newData, SIZE_WORD);
	CheckAndChangeN(newData, SIZE_WORD);
}

void M68K::OpcodeRotateNoCarryReg( UBYTE count, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg )
{
	int numrotate = 0 ;

	if ( ir == 0 )
	{
		numrotate = count == 0 ? 8 : count ;
	}
	else
	{
		numrotate = m_RegistersData[count] % 64 ;
	}

	const bool c_isLeftShift = dir == 1 ;

	int numBits = 0 ;
	DataSize ds;
	switch (size)
	{
	case 0:
		m_CyclesThisUpdate += 6 + ( 2 * numrotate ) ;
		numBits = 8 ; 
		ds = SIZE_BYTE;
		break ;
	case 1:
		m_CyclesThisUpdate += 6 + ( 2 * numrotate ) ;
		numBits = 16 ;
		ds = SIZE_WORD;
		break ;
	case 2:
		m_CyclesThisUpdate += 8 + ( 2 * numrotate ) ;
		numBits = 32 ;
		ds = SIZE_LONG;
		break ;
	default: LogMessage("M68k","err", "Invalid size in func OpcodeRotateNoCarryReg"); break ;
	}
	const UDWORD c_result = c_isLeftShift ? (UDWORD)BitRotateLeft<UQWORD>( m_RegistersData[reg], numrotate, numBits ) 
		: (UDWORD)BitRotateRight<UQWORD>( m_RegistersData[reg], numrotate, numBits ) ;

	SetDataRegistry(reg, c_result, ds); 

	CheckAndChangeZ(c_result,ds);
	CheckAndChangeN(c_result,ds);
	m_CCR = BitReset(m_CCR, FLAG_V) ;
	if (numrotate == 0)
	{
		m_CCR = BitReset(m_CCR, FLAG_C) ;
	}
	else if (c_isLeftShift)
	{
		m_CCR = BitConditionalSet( TestBit(c_result, 0), m_CCR, FLAG_C ) ;
	}
	else
	{
		m_CCR = BitConditionalSet( TestBit(c_result, numBits -1), m_CCR, FLAG_C ) ; // not sure about this
	}
}

void M68K::OpcodeRotateNoCarryMem( UBYTE dir, UBYTE mode, UBYTE reg )
{
	ASSERT(false) ; // test this function
	const int c_numrotate = 1 ;
	const bool c_isLeftShift = dir == 1 ;
	const int c_numBits = 16 ;

	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_WORD, true ) ;

	const UDWORD c_result = c_isLeftShift ? BitRotateLeft( getdata.operand, c_numrotate, c_numBits ) 
		: BitRotateRight( getdata.operand, c_numrotate, c_numBits ) ;

	SetEAOperand( mode, reg, c_result, SIZE_WORD ) ;

	CheckAndChangeZ(c_result, SIZE_WORD);
	CheckAndChangeN(c_result, SIZE_WORD);
	m_CCR = BitReset(m_CCR, FLAG_V) ;

	if (c_isLeftShift)
	{
		m_CCR = BitConditionalSet( TestBit(c_result, 0), m_CCR, FLAG_C ) ;
	}
	else
	{
		m_CCR = BitConditionalSet( TestBit(c_result, c_numBits -1), m_CCR, FLAG_C ) ; // not sure about this
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles + 8 ;
}

void M68K::OpcodeRotateCarryReg( UBYTE count, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg )
{
	//	ASSERT(false) ; // test this!

	int numrotate = 0 ;

	if ( ir == 0 )
	{
		numrotate = count == 0 ? 8 : count ;
	}
	else
	{
		numrotate = m_RegistersData[count] % 64 ;
	}

	const bool c_isLeftShift = dir == 1 ;

	int numBits = 0 ;
	DataSize ds;
	switch (size)
	{
	case 0:
		m_CyclesThisUpdate += 6 + ( 2 * numrotate ) ;
		ds = SIZE_BYTE;
		numBits = 8 ; 
		break ;
	case 1:
		m_CyclesThisUpdate += 6 + ( 2 * numrotate ) ;
		numBits = 16 ; 
		ds = SIZE_WORD;
		break ;
	case 2:
		m_CyclesThisUpdate += 8 + ( 2 * numrotate ) ;
		numBits = 32 ; 
		ds = SIZE_LONG;
		break ;
	default: LogMessage("M68k","err", "Invalid size in func OpcodeRotateCarryReg"); break ;
	}

	UDWORD result = 0;
	UDWORD toRotate = m_RegistersData[reg] ;
	if (c_isLeftShift)
	{
		for (int i = 0; i < numrotate; ++i)
		{
			m_CCR = BitConditionalSet( TestBit(toRotate, numBits-1), m_CCR, FLAG_C ) ;
			result = toRotate << 1;
			result = BitConditionalSet(TestBit(m_CCR, FLAG_X), result, 0);
			m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
		}
	}
	else
	{
		for (int i = 0; i < numrotate; ++i)
		{
			m_CCR = BitConditionalSet( TestBit(toRotate, 0), m_CCR, FLAG_C ) ;
			result = toRotate >> 1;
			result = BitConditionalSet(TestBit(m_CCR, FLAG_X), result, numBits -1);
			m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
		}
	}


	SetDataRegistry(reg, result, ds); 

	CheckAndChangeZ(result, ds);
	CheckAndChangeN(result, ds);
	m_CCR = BitReset(m_CCR, FLAG_V) ;
}

void M68K::OpcodeRotateCarryMem( UBYTE dir, UBYTE mode, UBYTE reg )
{
	const int c_numrotate = 1 ;
	const bool c_isLeftShift = dir == 1 ;
	const int c_numBits = 16 ;

	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_WORD, true ) ;
	UDWORD toRotate = getdata.operand;
	UDWORD result = 0 ;

	if (c_isLeftShift)
	{
		m_CCR = BitConditionalSet( TestBit(toRotate, c_numBits-1), m_CCR, FLAG_C ) ;
		result = toRotate << 1;
		result = BitConditionalSet(TestBit(m_CCR, FLAG_X), result, 0);
		m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
	}
	else
	{
		m_CCR = BitConditionalSet( TestBit(toRotate, 0), m_CCR, FLAG_C ) ;
		result = toRotate >> 1;
		result = BitConditionalSet(TestBit(m_CCR, FLAG_X), result, c_numBits -1);
		m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ;
	}

	EA_DATA setData = SetEAOperand(mode, reg, result, SIZE_WORD) ;
	m_CyclesThisUpdate = getdata.cycles + 8;
	m_ProgramCounter += setData.pcadvance;
	CheckAndChangeZ(result, SIZE_WORD);
	CheckAndChangeN(result, SIZE_WORD);
	m_CCR = BitReset(m_CCR, FLAG_V) ;
}

void M68K::OpcodeBSETDynamic( UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;

	m_CyclesThisUpdate += 8;

	const int c_bittotest = m_FuncReadMemory8(m_ProgramCounter + 1) % c_modulus ; 
	m_ProgramCounter += 2;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds, true) ;

	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitSet(getdata.operand, c_bittotest) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBCLRDynamic( UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;

	m_CyclesThisUpdate += c_ds == SIZE_BYTE ? 8 : 10 ;

	const int c_bittotest = m_FuncReadMemory8(m_ProgramCounter + 1) % c_modulus ; 
	m_ProgramCounter += 2;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds,true) ;

	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitReset(getdata.operand, c_bittotest) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBCHGDynamic( UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;

	m_CyclesThisUpdate += 8;

	const int c_bittotest = m_FuncReadMemory8(m_ProgramCounter + 1) % c_modulus ; 
	m_ProgramCounter += 2;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds,true) ;

	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), getdata.operand, c_bittotest ) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBSETStatic(UBYTE reg, UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;
	const int c_bittotest = m_RegistersData[reg] % c_modulus ;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds, true) ;
	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitSet(getdata.operand, c_bittotest) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_CyclesThisUpdate += 12 ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBCLRStatic(UBYTE reg, UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;
	const int c_bittotest = m_RegistersData[reg] % c_modulus ;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds,true) ;
	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitReset(getdata.operand, c_bittotest) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_CyclesThisUpdate += c_ds == SIZE_BYTE ? 12 : 14 ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBCHGStatic(UBYTE reg, UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;
	const int c_bittotest = m_RegistersData[reg] % c_modulus ;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds,true) ;
	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;
	getdata.operand = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), getdata.operand, c_bittotest ) ;

	EA_DATA setdata = SetEAOperand(mode,eareg, getdata.operand, c_ds) ;

	m_CyclesThisUpdate += 12;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBTSTDynamic( UBYTE reg, UBYTE mode, UBYTE eareg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;
	const int c_bittotest = m_RegistersData[reg] % c_modulus ;
	EA_DATA getdata = GetEAOperand(mode,eareg, c_ds,false) ;
	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;	

	m_CyclesThisUpdate += c_ds == SIZE_BYTE ? 4 : 6 ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeBTSTStatic(UBYTE mode, UBYTE reg )
{
	const int c_modulus = mode == 0 ? 32 : 8 ;
	const DataSize c_ds = mode == 0 ? SIZE_LONG : SIZE_BYTE ;

	m_CyclesThisUpdate += c_ds == SIZE_BYTE ? 8 : 10 ;

	const int c_bittotest = m_FuncReadMemory8(m_ProgramCounter + 1) % c_modulus ; 
	m_ProgramCounter += 2;
	EA_DATA getdata = GetEAOperand(mode,reg, c_ds,false) ;

	m_CCR = BitConditionalSet( !TestBit(getdata.operand, c_bittotest), m_CCR, FLAG_Z ) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeADDI( UBYTE size, UBYTE mode, UBYTE reg )
{
	UDWORD data = 0 ;
	DataSize ds ;
	int pcadvance = 0 ;
	switch(size)
	{
	case 0 : 
		ds = SIZE_BYTE;
		data = m_FuncReadMemory8(m_ProgramCounter+1) ;
		pcadvance += 2 ;
		break ;
	case 1 : 
		ds = SIZE_WORD;
		data = m_FuncReadMemory16(m_ProgramCounter) ;
		pcadvance += 2 ;
		break ;
	case 2 : 
		ds = SIZE_LONG; 
		data = m_FuncReadMemory32(m_ProgramCounter) ;
		pcadvance += 4 ;
		break ;
	default : 
		ASSERT(false); 
		break ;
	}

	m_ProgramCounter += pcadvance ;

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;

	const UDWORD c_result = getdata.operand + data ;
	CheckAndChangeCAdd(getdata.operand, data, ds, false);
	CheckAndChangeVAdd(getdata.operand, data, ds, false);

	SetEAOperand( mode, reg, c_result, ds ) ;

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 16 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 12 : 20 ;
	}

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above
}

void M68K::OpcodeADDQ(UBYTE data, UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE ; break ;
	case 1: ds = SIZE_WORD ; break ;
	case 2: ds = SIZE_LONG ; break ;
	default:
		LogMessage( "M68k","err", ( boost::format( "Processing invalid OpcodeADDQ data size mode %1% " )
			% size  ).str() ) ;

		ASSERT(false) ;
		break ;
	}

	int c_numToAdd = data == 0 ? 8 : data ;
	const bool c_extend = mode == EA_ADDRESS_REG && ds == SIZE_WORD;
	if (c_extend)
		c_numToAdd = SignExtend(UWORD(c_numToAdd));

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	const UDWORD c_result = getdata.operand + c_numToAdd ;
	SetEAOperand(mode, reg, c_result, ds ) ;

	if (c_extend)
		ds = SIZE_LONG;

	bool alterCC = true;

	if (mode == EA_DATA_REG)
	{
		m_CyclesThisUpdate += (ds==SIZE_LONG) ? 8 : 4;
	}
	else if (mode == EA_ADDRESS_REG)
	{
		alterCC = false;
		m_CyclesThisUpdate += (ds==SIZE_BYTE) ? 4 : 8;
	}
	else
	{
		m_CyclesThisUpdate += (ds==SIZE_LONG) ? 12 : 8;
	}
	if (alterCC)
	{
		CheckAndChangeCAdd(getdata.operand, c_numToAdd, ds, false);
		CheckAndChangeVAdd(getdata.operand, c_numToAdd, ds, false);
		CheckAndChangeZ(c_result, ds);
		CheckAndChangeN(c_result, ds);
		m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above
	}
	

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;
}

void M68K::OpcodeANDI(UBYTE size, UBYTE mode, UBYTE reg)
{
	DataSize ds ;
	UDWORD data = 0 ;

	int pcadvance = 0 ;

	switch(size)
	{
	case 0 : 
		ds = SIZE_BYTE;
		data = m_FuncReadMemory8(m_ProgramCounter+1) ;
		pcadvance += 2 ;
		break ;
	case 1 : 
		ds = SIZE_WORD;
		data = m_FuncReadMemory16(m_ProgramCounter) ;
		pcadvance += 2 ;
		break ;
	case 2 : 
		ds = SIZE_LONG; 
		data = m_FuncReadMemory32(m_ProgramCounter) ;
		pcadvance += 4 ;
		break ;
	default : ASSERT(false); break ;
	}
	m_ProgramCounter += pcadvance;
	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	const UDWORD c_result = data & getdata.operand ;

	SetEAOperand( mode, reg, c_result, ds ) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;


	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 16 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 20 : 12 ;
	}
}

void M68K::OpcodeCMPI( UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;

	UDWORD data = 0 ;
	int pcadvance = 0 ;

	switch(size)
	{
	case 0 : 
		ds = SIZE_BYTE;
		data = m_FuncReadMemory8(m_ProgramCounter+1) ;
		pcadvance += 2 ;
		break ;
	case 1 : 
		ds = SIZE_WORD;
		data = m_FuncReadMemory16(m_ProgramCounter) ;
		pcadvance += 2 ;
		break ;
	case 2 : 
		ds = SIZE_LONG; 
		data = m_FuncReadMemory32(m_ProgramCounter) ;
		pcadvance += 4 ;
		break ;
	default : 
		ASSERT(false); 
		break ;
	}


	m_ProgramCounter += pcadvance ;

	EA_DATA getdata = GetEAOperand( mode, reg, ds, false) ;
	const UDWORD c_result = getdata.operand - data ;

	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	CheckAndChangeCSub(getdata.operand, data, ds, false);
	CheckAndChangeVSub(getdata.operand, data, ds, false);

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 14 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 12 : 8 ;
	}
}

void M68K::OpcodeEORI( UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	int pcIncrease = 0 ;

	UDWORD data = 0 ;

	switch(size)
	{
	case 0: 
		data = m_FuncReadMemory8( m_ProgramCounter + 1 ) ; 
		pcIncrease = 2; 
		ds = SIZE_BYTE ; 
		break ;
	case 1: 
		data = m_FuncReadMemory16( m_ProgramCounter ) ; 
		pcIncrease = 2; 
		ds = SIZE_WORD ;
		break ;
	case 2: 
		data = m_FuncReadMemory32( m_ProgramCounter ) ; 
		pcIncrease = 4; 
		ds = SIZE_LONG ;
		break ;
	}

	m_ProgramCounter += pcIncrease ;

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	getdata.operand ^= data ;

	SetEAOperand( mode, reg, getdata.operand, ds ) ;

	m_ProgramCounter += getdata.pcadvance ;

	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(getdata.operand, ds);
	CheckAndChangeN(getdata.operand, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 16 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 20 : 12 ;
	}
}

void M68K::OpcodeMOVEQ( UBYTE reg, UBYTE data )
{
	m_CyclesThisUpdate += 4 ;
	const UDWORD c_extendedData = SignExtend( SignExtend( data ) ) ;
	m_RegistersData[reg] = c_extendedData ;

	CheckAndChangeZ(c_extendedData, SIZE_LONG);
	CheckAndChangeN(c_extendedData, SIZE_LONG);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;
}

void M68K::OpcodeORI( UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	int pcIncrease = 0 ;

	UDWORD data = 0 ;

	switch(size)
	{
	case 0: 
		data = m_FuncReadMemory8( m_ProgramCounter + 1 ) ; 
		pcIncrease = 2; 
		ds = SIZE_BYTE ; 
		break ;
	case 1: 
		data = m_FuncReadMemory16( m_ProgramCounter ) ; 
		pcIncrease = 2; 
		ds = SIZE_WORD ;
		break ;
	case 2: 
		data = m_FuncReadMemory32( m_ProgramCounter ) ; 
		pcIncrease = 4; 
		ds = SIZE_LONG ;
		break ;
	}

	m_ProgramCounter += pcIncrease ;

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	getdata.operand |= data ;

	SetEAOperand( mode, reg, getdata.operand, ds ) ;

	m_ProgramCounter += getdata.pcadvance ;

	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(getdata.operand, ds);
	CheckAndChangeN(getdata.operand, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 16 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 20 : 12 ;
	}
}

void M68K::OpcodeSUBI( UBYTE size, UBYTE mode, UBYTE reg )
{
	UDWORD data = 0 ;
	DataSize ds ;
	int pcadvance = 0 ;
	switch(size)
	{
	case 0 : 
		ds = SIZE_BYTE;
		data = m_FuncReadMemory8(m_ProgramCounter+1) ;
		pcadvance += 2 ;
		break ;
	case 1 : 
		ds = SIZE_WORD;
		data = m_FuncReadMemory16(m_ProgramCounter) ;
		pcadvance += 2 ;
		break ;
	case 2 : 
		ds = SIZE_LONG; 
		data = m_FuncReadMemory32(m_ProgramCounter) ;
		pcadvance += 4 ;
		break ;
	default : 
		ASSERT(false); 
		break ;
	}

	m_ProgramCounter += pcadvance ;

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	CheckAndChangeCSub( getdata.operand, data, ds, false );
	CheckAndChangeVSub( getdata.operand, data, ds, false );

	const UDWORD c_result = getdata.operand - data ;

	SetEAOperand( mode, reg, c_result, ds ) ;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	CheckAndChangeZ(c_result, ds);
	CheckAndChangeN(c_result, ds);
	m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above

	if (mode == EA_DATA_REG || mode == EA_ADDRESS_REG)
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 16 : 8 ;
	}
	else
	{
		m_CyclesThisUpdate += (ds == SIZE_LONG) ? 20 : 12 ;
	}
}

void M68K::OpcodeSUBQ( UBYTE data, UBYTE size, UBYTE mode, UBYTE reg )
{
	DataSize ds ;
	switch(size)
	{
	case 0: ds = SIZE_BYTE; break ;
	case 1: ds = SIZE_WORD; break ;
	case 2: ds = SIZE_LONG; break ;
	default:
		LogMessage("M68k","err", "Invalid size in SUBQ") ;
		break ;
	}
	UDWORD c_numToSub = data == 0 ? 8 : data ;
	const bool c_signExtend = mode == EA_ADDRESS_REG && ds == SIZE_WORD;
	if (c_signExtend)
		c_numToSub = SignExtend(UWORD(c_numToSub));

	EA_DATA getdata = GetEAOperand( mode, reg, ds, true ) ;
	const UDWORD c_result = getdata.operand - c_numToSub ;
	SetEAOperand(mode, reg, c_result, ds ) ;

	if (c_signExtend)
		ds = SIZE_LONG;

	m_ProgramCounter += getdata.pcadvance ;
	m_CyclesThisUpdate += getdata.cycles ;

	bool alterCC = true;
	if (mode == EA_DATA_REG)
	{
		m_CyclesThisUpdate += (ds==SIZE_LONG) ? 8 : 4;
	}
	else if (mode == EA_ADDRESS_REG)
	{
		alterCC = false;
		m_CyclesThisUpdate += (ds==SIZE_BYTE) ? 4 : 8;
	}
	else
	{
		m_CyclesThisUpdate += (ds==SIZE_LONG) ? 12 : 8;
	}
	if (alterCC)
	{
		CheckAndChangeCSub(getdata.operand, c_numToSub, ds, false);
		CheckAndChangeVSub(getdata.operand, c_numToSub, ds, false);
		CheckAndChangeZ(c_result, ds);
		CheckAndChangeN(c_result, ds);
		m_CCR = BitConditionalSet( TestBit(m_CCR, FLAG_C), m_CCR, FLAG_X ) ; // flag C set a few lines above
	}	
}

void M68K::OpcodeMOVEP( UBYTE dataReg, UBYTE opmode, UBYTE addressReg )
{
	const SDWORD c_displacement = SignExtend( m_FuncReadMemory16( m_ProgramCounter ) ) ;
	const UDWORD c_address = m_RegistersAddress[addressReg] + c_displacement;
	const bool c_isWord = opmode == 4 || opmode == 6;
	const bool c_isEven = (c_address % 2) == 0;
	if (c_isWord)
		m_CyclesThisUpdate += 16;
	else
		m_CyclesThisUpdate += 24;


	if (opmode < 6)
	{
		const UWORD c_word1 = m_FuncReadMemory16(c_address);
		const UWORD c_word2 = m_FuncReadMemory16(c_address+2);
		if (c_isWord)
		{			
			m_RegistersData[dataReg] &= 0xFFFF0000;
			if (c_isEven)
			{
				m_RegistersData[dataReg] |= (c_word1 & 0xFF00) | (c_word2 >> 8);
			}
			else
			{
				m_RegistersData[dataReg] |= ((c_word1 & 0xFF) << 8) | (c_word2 & 0xFF);
			}
		}
		else
		{
			const UWORD c_word3 = m_FuncReadMemory16(c_address+4);
			const UWORD c_word4 = m_FuncReadMemory16(c_address+6);
			if (c_isEven)
			{
				m_RegistersData[dataReg] = ((c_word1 & 0xFF00) << 16)| ((c_word2 && 0xFF00) << 8) | (c_word3 & 0xFF00) | (c_word4 >> 8);
			}
			else
			{
				m_RegistersData[dataReg] = ((c_word1 & 0xFF) << 24)| ((c_word2 && 0xFF) << 16) | ((c_word3 & 0xFF) << 8) | (c_word4 & 0xFF);
			}
		}

	}
	else
	{
		if (c_isWord)
		{			
			m_FuncWriteMemory8(c_address, UBYTE(m_RegistersData[dataReg]>>8));
			m_FuncWriteMemory8(c_address+2, UBYTE(m_RegistersData[dataReg]));
		}
		else
		{
			m_FuncWriteMemory8(c_address, UBYTE(m_RegistersData[dataReg]>>24));
			m_FuncWriteMemory8(c_address+2, UBYTE(m_RegistersData[dataReg]>>16));
			m_FuncWriteMemory8(c_address+4, UBYTE(m_RegistersData[dataReg]>>8));
			m_FuncWriteMemory8(c_address+6, UBYTE(m_RegistersData[dataReg]));
		}
	}

	m_ProgramCounter += 2;
}

void M68K::OpcodeJMP( UBYTE mode, UBYTE reg )
{
	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_LONG, false) ; // not sure if SIZE_LONG correct

	m_ProgramCounter = getdata.pointer ;
	if (getdata.eatype == EA_ADDRESS_REG_INDIRECT)
		m_CyclesThisUpdate += 8;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_POST_INC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
		m_CyclesThisUpdate += 10;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_INDEX)
		m_CyclesThisUpdate += 14;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_SHORT)
		m_CyclesThisUpdate += 10;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_LONG)
		m_CyclesThisUpdate += 12;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_DISPLACEMENT)
		m_CyclesThisUpdate += 10;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_PREINDEX)
		m_CyclesThisUpdate += 14;
}

void M68K::OpcodeJSR( UBYTE mode, UBYTE reg )
{
	m_RegistersAddress[0x7] -= 4 ;
	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_LONG, false ) ;
	m_ProgramCounter += getdata.pcadvance;
	m_FuncWriteMemory32(m_RegistersAddress[0x7], m_ProgramCounter) ;
	m_ProgramCounter = getdata.pointer;

	if (getdata.eatype == EA_ADDRESS_REG_INDIRECT)
		m_CyclesThisUpdate += 16;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_POST_INC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
		m_CyclesThisUpdate += 18;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_INDEX)
		m_CyclesThisUpdate += 22;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_SHORT)
		m_CyclesThisUpdate += 18;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_LONG)
		m_CyclesThisUpdate += 20;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_DISPLACEMENT)
		m_CyclesThisUpdate += 18;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_PREINDEX)
		m_CyclesThisUpdate += 22;
}

void M68K::OpcodeLEA(UBYTE regtoload, UBYTE mode, UBYTE reg)
{
	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_LONG, false ) ;
	m_RegistersAddress[regtoload] = getdata.pointer ;
	m_ProgramCounter += getdata.pcadvance ;

	if (getdata.eatype == EA_ADDRESS_REG_INDIRECT)
		m_CyclesThisUpdate += 4;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_POST_INC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
		m_CyclesThisUpdate += 8;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_INDEX)
		m_CyclesThisUpdate += 12;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_SHORT)
		m_CyclesThisUpdate += 8;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_LONG)
		m_CyclesThisUpdate += 12;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_DISPLACEMENT)
		m_CyclesThisUpdate += 8;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_PREINDEX)
		m_CyclesThisUpdate += 12;
}

void M68K::OpcodePEA( UBYTE mode, UBYTE reg )
{
	EA_DATA getdata = GetEAOperand( mode, reg, SIZE_LONG, false ) ;
	m_ProgramCounter += getdata.pcadvance ;
	m_RegistersAddress[0x7] -= 4 ;
	m_FuncWriteMemory32(m_RegistersAddress[0x7], getdata.pointer) ;
	if (getdata.eatype == EA_ADDRESS_REG_INDIRECT)
		m_CyclesThisUpdate += 12;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_POST_INC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
		m_CyclesThisUpdate += 0;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
		m_CyclesThisUpdate += 16;
	else if (getdata.eatype == EA_ADDRESS_REG_INDIRECT_INDEX)
		m_CyclesThisUpdate += 20;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_SHORT)
		m_CyclesThisUpdate += 16;
	else if (getdata.mode7Type == EA_MODE_7_ABS_ADDRESS_LONG)
		m_CyclesThisUpdate += 20;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_DISPLACEMENT)
		m_CyclesThisUpdate += 16;
	else if (getdata.mode7Type == EA_MODE_7_PC_WITH_PREINDEX)
		m_CyclesThisUpdate += 20;
}

void M68K::OpcodeMOVEM(UBYTE direction, UBYTE size, UBYTE mode, UBYTE reg)
{
	//	DebugDump();
	// warning this function is convoluted and ugly. Over doing it to make this fucker easier to debug

	const UWORD c_regmask = m_FuncReadMemory16( m_ProgramCounter ) ;
	
	m_ProgramCounter += 2 ;

	UDWORD addressOffset = 0 ;

	// worked this out by looking at the 2 ea tables for MOVEM and seeing which modes are the same for both tables
	bool isControlMode = (mode == 2) || (mode == 5) || (mode == 6) || (mode == 7 && (reg == 0 || reg == 1 || reg == 2 || reg == 3));

	const bool c_regtomem = direction == 0 ;
	const bool c_isword =  size == 0;
	m_CyclesThisUpdate += BitCount( c_regmask, 16) * (c_isword?4:8);
 
	if (!c_regtomem)
	{
		if (mode == EA_ADDRESS_REG_INDIRECT)
			m_CyclesThisUpdate += 12 ;
		else if (mode == EA_ADDRESS_REG_INDIRECT_POST_INC)
			m_CyclesThisUpdate += 12;
		else if (mode == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
			m_CyclesThisUpdate += 0  ;
		else if (mode == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
			m_CyclesThisUpdate += 16;
		else if (mode == EA_ADDRESS_REG_INDIRECT_INDEX)
			m_CyclesThisUpdate += 18;
		else if (mode == 7 && reg ==  EA_MODE_7_ABS_ADDRESS_SHORT)
			m_CyclesThisUpdate += 16;
		else if (mode == 7 && reg ==  EA_MODE_7_ABS_ADDRESS_LONG)
			m_CyclesThisUpdate += 20;
		else if (mode == 7 && reg ==  EA_MODE_7_PC_WITH_DISPLACEMENT)
			m_CyclesThisUpdate += 16;
		else if (mode == 7 && reg ==  EA_MODE_7_PC_WITH_PREINDEX)
			m_CyclesThisUpdate += 18;

		
	}
	else
	{
		if (mode == EA_ADDRESS_REG_INDIRECT)
			m_CyclesThisUpdate += 8 ;
		else if (mode == EA_ADDRESS_REG_INDIRECT_POST_INC)
			m_CyclesThisUpdate += 0;
		else if (mode == EA_ADDRESS_REG_INDIRECT_PRE_DEC)
			m_CyclesThisUpdate += 8  ;
		else if (mode == EA_ADDRESS_REG_INDIRECT_DISPLACEMENT)
			m_CyclesThisUpdate += 12;
		else if (mode == EA_ADDRESS_REG_INDIRECT_INDEX)
			m_CyclesThisUpdate += 14;
		else if (mode == 7 && reg ==  EA_MODE_7_ABS_ADDRESS_SHORT)
			m_CyclesThisUpdate += 12;
		else if (mode == 7 && reg ==  EA_MODE_7_ABS_ADDRESS_LONG)
			m_CyclesThisUpdate += 16;
		else if (mode == 7 && reg ==  EA_MODE_7_PC_WITH_DISPLACEMENT)
			m_CyclesThisUpdate += 0;
		else if (mode == 7 && reg ==  EA_MODE_7_PC_WITH_PREINDEX)
			m_CyclesThisUpdate += 0;
	}

	bool predecMode = false;
	if (mode == 4)
		predecMode = true;
	if (mode == 7)
	{
		//	_asm int 3;
		//if (reg == 2 || reg == 3)
		//	predecMode = true;
	}

	bool postDecMode = !isControlMode && !predecMode;

	if (predecMode)
	{
		ASSERT(c_regtomem);
	}
	else if (postDecMode)
	{
		ASSERT(!c_regtomem);
	}

	bool regdirection = !predecMode ;

	int pcAdvance = 0 ;
	int directionInc = regdirection ? 1 : -1;

	for (int index = 0; index < 16; ++index)
	{
		int newIndex = index%8;
		if (!regdirection)newIndex = 7 - newIndex;
		const bool c_isaddressreg = (regdirection && (index > 7)) || (!regdirection && (index < 8)) ;

		if (TestBit( c_regmask, index) ) 
		{
			if (c_regtomem)
			{
				UDWORD data = 0 ;

				if (c_isaddressreg)
				{
					data = m_RegistersAddress[newIndex];
				}
				else
				{
					data = m_RegistersData[newIndex];
				}

				if (c_isword && mode == EA_ADDRESS_REG)
				{
					data = SignExtend( (UWORD)data ) ;
				}
				// MAY NEED TO CHECK ADDRESS REG HERE FORE MODE 1
				EA_DATA setdata = SetEAOperand( mode, reg, data, c_isword ? SIZE_WORD : SIZE_LONG, addressOffset ) ;
				if (isControlMode)
				{
					addressOffset += c_isword ? 2 : 4;
				}
				else if (predecMode)
				{
					//				addressOffset -= c_isword ? 2 : 4;
				}
				pcAdvance = setdata.pcadvance ;
				if (pcAdvance > 0)
				{
					// not sure if pcAdvance should affect m_ProgramCounter immediately or at
					// the end of the insturction like memtoreg (below) does
					//				ASSERT(false) ;
				}
			}
			else
			{
				EA_DATA getdata = GetEAOperand(mode, reg, c_isword ? SIZE_WORD : SIZE_LONG, false,addressOffset) ;
				if (isControlMode)
				{
					addressOffset += c_isword ? 2 : 4;
				}
				pcAdvance = getdata.pcadvance ;

				if (c_isword)
				{
					getdata.operand = SignExtend( (UWORD)getdata.operand ) ;
				}
				if (c_isaddressreg)
				{
					m_RegistersAddress[newIndex] = getdata.operand ;
				}
				else
				{
					m_RegistersData[newIndex] = getdata.operand ;
				}
			}
		}		
	}	
	m_ProgramCounter += pcAdvance ;
}

void M68K::OpcodeTRAP( UBYTE vec )
{
	RequestInterrupt(INT_Trap, m_FuncReadMemory32((32 + vec)*4) );
	m_CyclesThisUpdate += 38;
}

void M68K::OpcodeTRAPV()
{
	LogMessage("M68k", "wrn", 0, "OpcodeTRAPV not tested");
	if (TestBit(m_CCR, FLAG_V))
		OpcodeTRAP(7);
	m_CyclesThisUpdate += 4;
}

void M68K::OpcodeILLEGAL()
{
	ASSERT(false);
	LogMessage("M68k", "wrn", 0, "OpcodeILLEGAL not tested");
	OpcodeTRAP(4);
	m_CyclesThisUpdate += 34;
}

void M68K::OpcodeMOVE(UBYTE size, UBYTE destReg, UBYTE destMode, UBYTE sourceMode, UBYTE sourceReg)
{
	DataSize ds ;
	switch(size)
	{
	case 1 : ds = SIZE_BYTE; break ;
	case 3 : ds = SIZE_WORD;  break ; 
	case 2 : ds = SIZE_LONG;  break ;
	default: LogMessage("M68k", "err", "Invalid Move Size"); break;
	}

	EA_DATA getdata = GetEAOperand(sourceMode, sourceReg, ds, false) ;
	m_ProgramCounter += getdata.pcadvance ;
	EA_DATA setdata = SetEAOperand( destMode, destReg, getdata.operand, ds ) ;
	m_ProgramCounter += setdata.pcadvance ;

	CheckAndChangeZ(getdata.operand, ds);
	CheckAndChangeN(getdata.operand, ds);
	m_CCR = BitReset( m_CCR, FLAG_V ) ;
	m_CCR = BitReset( m_CCR, FLAG_C ) ;

	int sourceCycles = sourceMode;
	if (sourceMode == EA_MODE_7)
		sourceCycles += sourceReg;
	int destCycles = destMode;
	if (destMode == EA_MODE_7)
		destCycles += destReg;

	if (ds == SIZE_LONG)
		m_CyclesThisUpdate += m_MoveCyclesLong[sourceCycles][destCycles];
	else
		m_CyclesThisUpdate += m_MoveCyclesByteWord[sourceCycles][destCycles];
}

void M68K::OpcodeMOVEA(UBYTE size, UBYTE destReg, UBYTE sourceMode, UBYTE sourceReg)
{
	DataSize ds ;
	switch(size)
	{
	case 3 : ds = SIZE_WORD;  break ; 
	case 2 : ds = SIZE_LONG;  break ;
	default: LogMessage("M68k", "err", "Invalid Move Size"); break;
	}

	EA_DATA getdata = GetEAOperand(sourceMode, sourceReg, ds, false) ;
	if (ds == SIZE_WORD)
		getdata.operand = SignExtend(UWORD(getdata.operand));
	
	m_RegistersAddress[destReg] = getdata.operand;

	m_ProgramCounter += getdata.pcadvance ;

	int sourceCycles = sourceMode;
	if (sourceMode == EA_MODE_7)
		sourceCycles += sourceReg;

	if (ds == SIZE_LONG)
		m_CyclesThisUpdate += m_MoveCyclesLong[sourceCycles][EA_ADDRESS_REG];
	else
		m_CyclesThisUpdate += m_MoveCyclesByteWord[sourceCycles][EA_ADDRESS_REG];
}

void M68K::OpcodeANDEADEST(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg )
{
	OpcodeAND(datareg,opmode,mode,reg);
}

void M68K::OpcodeANDEASOURCE(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg )
{
	OpcodeAND(datareg,opmode,mode,reg);
}

void M68K::OpcodeOREADEST( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeOR(reg,opmode,mode,eareg);
}

void M68K::OpcodeOREASOURCE( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeOR(reg,opmode,mode,eareg);
}

void M68K::OpcodeADDEADest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeADD(reg,opmode,mode,eareg);
}

void M68K::OpcodeADDDNDest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeADD(reg,opmode,mode,eareg);
}

void M68K::OpcodeSUBDNDest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeSUB(reg, opmode, mode, eareg);
}

void M68K::OpcodeSUBEADest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg )
{
	OpcodeSUB(reg, opmode, mode, eareg);
}

void M68K::OpcodeReset()
{
	Reset();
	m_CyclesThisUpdate = 132;
}