#include "Config_M68k.h"
#include "M68K.h"
#include <cs_core/Logging/Logging.h>
#include <cs_core/Utils/Maths.h>
#include <cs_core/Utils/Debugging.h>
#include <cs_core/Utils/Throw.h>
#include <cs_core/IniFiles/IniFiles.h>
#include <fstream>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>
#include <cs_core/Logging/LoggingManager.h>

using codeslinger::maths::TestBit ;

using codeslinger::logging::LogMessage ;

////////////////////////////////////////////////////////////////////

M68K::PTR M68K::Create( )
{
	return M68K::PTR( new M68K() ) ;
}

////////////////////////////////////////////////////////////////////

M68K::M68K(void)
	: m_ProgramCounter(0x200)
	, m_CCR(0)
	, m_Stop(false)
	, m_servicingInterrupt(false)
	, m_SSP(0)
	, m_USP(0)
	, m_InSuperVisorMode(true)
{
}

////////////////////////////////////////////////////////////////////

void M68K::Init( )
{
	codeslinger::inifile::IniFileLoadSettings("M68k", "M68k.ini" ) ;
	codeslinger::logging::LoggingAddNewGroup("M68k");
	if (codeslinger::inifile::IniFileReturnDefault("M68k","CreateOpcodeTable",false))
	{
		OutputOpcodeTable() ;
	}	
	Reset( ) ;
	InitOpcodeTable() ;
	CreateOpcodeTable() ;
	CreateMoveCycleTables();
}

////////////////////////////////////////////////////////////////////

M68K::~M68K(void)
{
}

////////////////////////////////////////////////////////////////////

void M68K::CreateMoveCycleTables()
{
	for (int i = 0 ; i < c_numEa; ++i)
	{
		for (int j = 0 ; j < c_numEa; ++j)
		{
			m_MoveCyclesByteWord[i][j] = 0;
			m_MoveCyclesLong[i][j] = 0;
		}
	}
	m_MoveCyclesByteWord[EA_DATA_REG][EA_DATA_REG] = 4 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG] = 4 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT] = 8 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 8 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 8 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 12 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_INDEX] = 14 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 12 ;
	m_MoveCyclesByteWord[EA_DATA_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 16 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_DATA_REG] = 4 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG] = 4 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_INDEX] = 14 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 16 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_DATA_REG] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_INDEX] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 20 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_DATA_REG] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG] = 8 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_POST_INC] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_INDEX] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 20 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_DATA_REG] = 10 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG] = 10 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT] = 14 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_POST_INC] = 14 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 14 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_INDEX] = 20 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 22 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_DATA_REG] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 16 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 20 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_INDEX] = 22 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 20 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 24 ;

	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_DATA_REG] = 14 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG] = 48 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_POST_INC] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 18 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 22 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_INDEX] = 24 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 22 ;
	m_MoveCyclesByteWord[EA_ADDRESS_REG_INDIRECT_INDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 26 ;

	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_DATA_REG] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_INDEX] = 22 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 24 ;

	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_DATA_REG] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 24 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_INDEX] = 26 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 24 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 28 ;

	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_DATA_REG] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_INDEX] = 22 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 20 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 24 ;

	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_DATA_REG] = 14 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG] = 14 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT] = 18 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_POST_INC] = 18 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 18 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 22 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_INDEX] = 24 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 22 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 26 ;

	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_DATA_REG] = 8 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG] = 8 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_POST_INC] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 12 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_INDEX] = 18 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 16 ;
	m_MoveCyclesByteWord[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 20 ;

	m_MoveCyclesLong[EA_DATA_REG][EA_DATA_REG] = 4 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG] = 4 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT] = 12 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 12 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 12 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 16 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_ADDRESS_REG_INDIRECT_INDEX] = 18 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 16 ;
	m_MoveCyclesLong[EA_DATA_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 20 ;

	m_MoveCyclesLong[EA_ADDRESS_REG][EA_DATA_REG] = 4 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG] = 4 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 16 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_ADDRESS_REG_INDIRECT_INDEX] = 18 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 16 ;
	m_MoveCyclesLong[EA_ADDRESS_REG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 20 ;

	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_DATA_REG] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_ADDRESS_REG_INDIRECT_INDEX] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 28 ;

	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_DATA_REG] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_POST_INC] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 20 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_ADDRESS_REG_INDIRECT_INDEX] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_POST_INC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 28 ;

	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_DATA_REG] = 14 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG] = 14 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT] = 22 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_POST_INC] = 22 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 22 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_ADDRESS_REG_INDIRECT_INDEX] = 28 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_PRE_DEC][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 30 ;

	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_DATA_REG] = 16 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG] = 16 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 24 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 28 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_INDEX] = 30 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 28 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 32 ;

	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_DATA_REG] = 18 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG] = 18 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_POST_INC] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 26 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 30 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_ADDRESS_REG_INDIRECT_INDEX] = 32 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 30 ;
	m_MoveCyclesLong[EA_ADDRESS_REG_INDIRECT_INDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 34 ;

	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_DATA_REG] = 16 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG] = 16 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_ADDRESS_REG_INDIRECT_INDEX] = 30 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_SHORT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 32 ;

	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_DATA_REG] = 20 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG] = 20 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_POST_INC] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 32 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_ADDRESS_REG_INDIRECT_INDEX] = 34 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 32 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_ABS_ADDRESS_LONG][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 36 ;

	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_DATA_REG] = 16 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG] = 16 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_POST_INC] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_ADDRESS_REG_INDIRECT_INDEX] = 30 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 28 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_DISPLACEMENT][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 32 ;

	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_DATA_REG] = 18 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG] = 18 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT] = 26 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_POST_INC] = 26 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 26 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 30 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_ADDRESS_REG_INDIRECT_INDEX] = 32 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 30 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_PC_WITH_PREINDEX][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 34 ;

	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_DATA_REG] = 12 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG] = 12 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT] = 20 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_POST_INC] = 20 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_PRE_DEC] = 20 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_DISPLACEMENT] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_ADDRESS_REG_INDIRECT_INDEX] = 26 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_SHORT] = 24 ;
	m_MoveCyclesLong[EA_MODE_7+EA_MODE_7_IMMEDIATE_DATA][EA_MODE_7 + EA_MODE_7_ABS_ADDRESS_LONG] = 28 ;
}

////////////////////////////////////////////////////////////////////

void M68K::Reset( )
{
	m_CCR = 0x2700 ;
	m_Stop = false;
	for (int i = 0; i <8; ++i)
	{
		m_RegistersData[i] = 0x0;
		m_RegistersAddress[i] =0x0;
	}

	m_RegistersAddress[0x7] = m_FuncReadMemory32(0x0) ;
	m_SSP = m_RegistersAddress[0x7];
	m_USP = m_RegistersAddress[0x7];
	m_InSuperVisorMode = true;
	m_ProgramCounter = m_FuncReadMemory32(0x4) ;

//  	for (int i = 0 ; i < 0xFF; i+=4)
//  	{
//  	 UDWORD test = m_FuncReadMemory32(i);
//  	 if (test == 0x200)
//  		 _asm int 3;
//  	}

	m_ProgramCounterStart = m_ProgramCounter ;
}

////////////////////////////////////////////////////////////////////

void  M68K::SetAddressRegistry(int id, UDWORD result, DataSize ds)
{
	switch(ds)
	{
	case SIZE_BYTE:
		LogMessage("M68k", "err", "Trying to assign a byte to an address reg. Didnt think you could do this");
		break;
	case SIZE_WORD:
		m_RegistersAddress[id] = result; 
		break;
	case SIZE_LONG: 
		m_RegistersAddress[id] = result; 
		break;
	}
}

////////////////////////////////////////////////////////////////////

void  M68K::SetDataRegistry(int id, UDWORD result, DataSize ds)
{
	switch(ds)
	{
		case SIZE_BYTE:
			m_RegistersData[id] &= 0xffffff00;
			m_RegistersData[id] |= (UBYTE)result;
			break;
		case SIZE_WORD:
			m_RegistersData[id] &= 0xffff0000;
			m_RegistersData[id] |= (UWORD)result;
			break;
		case SIZE_LONG: 
			m_RegistersData[id] = result; 
			break;
	}
}

////////////////////////////////////////////////////////////////////

void M68K::RequestAutoVectorInterrupt(InterruptType level)
{
	int address = m_FuncReadMemory32((static_cast<int>(level) * 4) + 0x60);
	m_interrupts.insert(std::make_pair(level, address));
}

////////////////////////////////////////////////////////////////////

void M68K::RequestInterrupt(InterruptType level, int address)
{
	m_interrupts.insert(std::make_pair(level, address));
}

////////////////////////////////////////////////////////////////////

void M68K::CheckForPrivilegeChange()
{
	const bool c_oldSuperVisorMode = m_InSuperVisorMode;
	m_InSuperVisorMode = TestBit(m_CCR, 13);
	if (c_oldSuperVisorMode != m_InSuperVisorMode)
	{
		if (c_oldSuperVisorMode)
		{
			m_SSP = m_RegistersAddress[0x7];
			m_RegistersAddress[0x7] = m_USP;
		}
		else
		{
			m_USP = m_RegistersAddress[0x7];
			m_RegistersAddress[0x7] = m_SSP;
		}
	}
}

////////////////////////////////////////////////////////////////////

void M68K::CheckForInterrupts()
{
	if (m_interrupts.empty())
		return ;
	const std::pair<InterruptType,int> type = (*m_interrupts.rbegin());

	const int c_interruptlevel = (m_CCR >> 8) & 0x7 ; 
	if (type.first > INT_AutoVector_6 || (type.first > c_interruptlevel))
	{
		const UWORD c_oldCCR = m_CCR;
		m_interrupts.erase(--m_interrupts.rbegin().base());
		m_CCR = codeslinger::maths::BitSet(m_CCR, 13);
		CheckForPrivilegeChange();
		m_servicingInterrupt = true;
		m_Stop = false;

		m_RegistersAddress[0x7] -= 4 ;
		m_FuncWriteMemory32(m_RegistersAddress[0x7], m_ProgramCounter) ;
		m_RegistersAddress[0x7] -= 2 ;
		m_FuncWriteMemory16(m_RegistersAddress[0x7], c_oldCCR) ;
		
		// set interrupt priority to equal to this interrupt type so lower priorities cant interrupt
		if (type.first <= INT_AutoVector_7)
		{
			m_CCR &= 0xF8FF ;
			m_CCR |= type.first << 8 ;
		}
		
		m_ProgramCounter = type.second ;
	}	
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncWriteMemory16( const t_FuncWriteMemory16& func )
{
	m_FuncWriteMemory16 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncReadMemory32( const t_FuncReadMemory32& func )
{
	m_FuncReadMemory32 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncWriteMemory32( const t_FuncWriteMemory32& func )
{
	m_FuncWriteMemory32 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncReadMemory16( const t_FuncReadMemory16& func )
{
	m_FuncReadMemory16 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncWriteMemory8( const t_FuncWriteMemory8& func )
{
	m_FuncWriteMemory8 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::SetFuncReadMemory8( const t_FuncReadMemory8& func )
{
	m_FuncReadMemory8 = func ;
}

////////////////////////////////////////////////////////////////////

void M68K::DebugDump()
{
	if (codeslinger::logging::LoggingManager::GetSingletonPtr()->GetEnabled())
	{
 		LogMessage("M68k", "info",0,"=============================");
 		for (int i = 0; i < 8; i+=2)
 		{
 			LogMessage("M68k", "info",0, (boost::format("A%1%=0x%2$x    A%3%=0x%4$x") % i % m_RegistersAddress[i] % (i+1) % m_RegistersAddress[i+1]).str());
 		}
 		for (int i = 0; i < 8; i+=2)
 		{
 			LogMessage("M68k", "info",0, (boost::format("D%1%=0x%2$x    D%3%=0x%4$x") % i % m_RegistersData[i] % (i+1) % m_RegistersData[i+1]).str());
 		}
 		LogMessage("M68k", "info",0, (boost::format("PC=0x%1$x    SR=0x%2$x") % m_ProgramCounter % m_CCR).str());
 		LogMessage("M68k", "info",0, (boost::format("CYCLES = 0x%1$x OLD PC=0x%2$x") % m_CyclesThisUpdate % m_ProgramCounterStart).str());
 		LogMessage("M68k", "info",0,"=============================");
 		//_asm int 3;
	}
}

////////////////////////////////////////////////////////////////////

int M68K::Update(  )
{
	CheckForInterrupts();
	CheckForPrivilegeChange();
	
	if (m_Stop)
	{
		return 4;
	}
 	m_CyclesThisUpdate = 0 ;
 	const UDWORD c_instruction = m_FuncReadMemory16( m_ProgramCounter ) ;

	m_ProgramCounterStart = m_ProgramCounter ; 
 	m_ProgramCounter += 2;
	m_OpcodeTable[c_instruction].Invoke( this ) ;

	return m_CyclesThisUpdate ;
}

////////////////////////////////////////////////////////////////////

bool M68K::ConditionTest(int condition) const
{

	// use http://en.wikibooks.org/wiki/68000_Assembly#Conditional_tests
	// this is also very useful http://www.easy68k.com/paulrsm/doc/trick68k.htm
	// this is also excellent, probably best of all http://tict.ticalc.org/docs/68kguide.txt
	bool res = false ;
	switch( condition )
	{
	case 0: // T*
		res = true ; 
		break ;
	case 1: // F*
		res = false ; 
		break ;
	case 2: // HI
		res = !TestBit(m_CCR, FLAG_C) && !TestBit(m_CCR, FLAG_Z);
		break ;
	case 3: // LS
		res = TestBit(m_CCR, FLAG_C) || TestBit(m_CCR, FLAG_Z);
		break ;
	case 4: // CC
		res = !TestBit(m_CCR, FLAG_C) ;
		break ;
	case 5: // CS
		res = TestBit(m_CCR, FLAG_C) ;
		break ;
	case 6: // BNE
		res = !TestBit( m_CCR, FLAG_Z) ;
		break ;
	case 7: // BEQ
		res = TestBit( m_CCR, FLAG_Z ) ;
		break ;
	case 9: // VS
		res = TestBit( m_CCR, FLAG_V ) ;
		break;
	case 10: //PL
		res = !TestBit( m_CCR, FLAG_N);
		break;
	case 11: //MI
		res = TestBit( m_CCR, FLAG_N);
		break;
	case 12: //GE
		res = TestBit(m_CCR, FLAG_N) == TestBit(m_CCR, FLAG_V) ;
		break ;
	case 13: //BLT
		res = TestBit(m_CCR, FLAG_N) != TestBit(m_CCR, FLAG_V) ;
		break ;
	case 14: // GT
		res = (TestBit(m_CCR, FLAG_N) == TestBit(m_CCR, FLAG_V)) && !TestBit(m_CCR, FLAG_Z) ;
		break ;
	case 15: // LE
		res = (TestBit(m_CCR, FLAG_N) != TestBit(m_CCR, FLAG_V)) || TestBit(m_CCR, FLAG_Z)  ;
		break ;
	default:
		LogMessage("M68k","wrn", ( boost::format( "ConditionTest not yet implemented condition code %1%")
			% condition ).str() ) ;
		ASSERT(false) ;
		break ;
	}
	return res ;
}

////////////////////////////////////////////////////////////////////

static std::vector<int> GetExclusionList( const std::string& strList )
{
	std::vector<int> res ;
	size_t lastPos = 0 ;
	size_t currentPos = strList.find( ',' ) ;
	while (currentPos != std::string::npos)
	{
		std::string szNum( strList.begin() + lastPos, strList.begin() + currentPos ) ;
		int num = boost::lexical_cast<int>( szNum ) ;
		res.push_back(num) ;
		++currentPos ;
		lastPos = currentPos ;
		currentPos = strList.find(',', currentPos ) ;
	}
	std::string szNum( strList.begin() + lastPos, strList.end() ) ;
	int num = boost::lexical_cast<int>( szNum ) ;
	res.push_back(num) ;
	return res ;
}

void M68K::OutputOpcodeTable()
{
 	boost::filesystem::create_directory("Temp") ;
 	std::ofstream output;
 	std::ifstream input;
 	input.open("m68kopcodes.txt") ;
 	output.open("m68kopcodedeclarations.txt") ;
 
 	if (!input.is_open() || !output.is_open())
 	{
 		return ;
 	}
 
 	while(!input.eof())
 	{
 		char buffer[256] ;
 		input.getline( buffer, 256 ) ;
 
 		std::string line = buffer ;
 
 		const size_t c_endinstructionpos = line.find( ' ' ) ;
 		if (c_endinstructionpos == std::string::npos)
 		{
 			break ;
 		}
 
 		const size_t c_enddeclaationpos = line.find( ' ', c_endinstructionpos + 1 ) ;
 		if (c_enddeclaationpos == std::string::npos)
 		{
 			break ;
 		}
 
 		const size_t c_endsignaturepos = line.find( ')', c_enddeclaationpos + 1 ) ;
 		if (c_endsignaturepos == std::string::npos)
 		{
 			break ;
 		}
 
 
 		const std::string c_instruction = std::string( line.begin(), 
 			line.begin() + c_endinstructionpos ) ;
 		const std::string c_declaration = std::string( line.begin() + c_endinstructionpos + 1,  
 			line.begin() + c_enddeclaationpos ) ;
 		const std::string c_signature = std::string( line.begin() + c_enddeclaationpos + 1,  
 			line.begin() + c_endsignaturepos + 1 ) ;
 		const std::string c_excludes = std::string( line.begin() + c_endsignaturepos + 1, 
 			line.end() ) ;
 
 		t_exclusionMap exclusionmap ;
 
 		if ( c_excludes.find( "Excludes") != std::string::npos )
 		{
 			size_t currentPoint = c_excludes.find('=') ;
 			while ( currentPoint != std::string::npos )
 			{
 				std::string letter = std::string( c_excludes.begin() + currentPoint - 1, 
 					c_excludes.begin() + currentPoint ) ;
 				char ch = letter[0] ;
 				
 
 				size_t spacepos = c_excludes.find( ' ', currentPoint ) ;
 				if (spacepos == std::string::npos)
 				{
 					spacepos = c_excludes.size() ;
 				}
 
 				std::string condString = std::string(c_excludes.begin() + currentPoint + 1, c_excludes.begin() + spacepos);
 				size_t condPos = condString.find("cond");
 				if ( condPos != std::string::npos)
 				{
 					char newChar = condString[ condPos + 5];
 					int newVal = condString[ condPos + 7] - '0';
 					size_t endPos = condString.find(')');
 					std::string exclusionlist(condString.begin() +condPos + 11, 
 						condString.begin() + endPos);
 					std::vector<int> exclusions = GetExclusionList( exclusionlist ) ;
 
 					if (!exclusions.empty())
 					{
 						std::map<int, std::vector<int> > map3;
 						std::multimap<char, std::map<int,std::vector<int> > > map2;
 						map3.insert( std::make_pair( newVal, exclusions ) ) ;
 						map2.insert( std::make_pair( newChar, map3));
 						exclusionmap.insert(std::make_pair(ch, map2)) ;
 					}
 					currentPoint = c_excludes.find('=',  currentPoint + endPos + 1) ;
 				}
 				else
 				{
 					std::string exclusionlist( c_excludes.begin() + currentPoint + 1, 
 						c_excludes.begin() + spacepos) ;
 					std::vector<int> exclusions = GetExclusionList( exclusionlist ) ;
 
 					if (!exclusions.empty())
 					{
 						std::multimap<char, std::map<int,std::vector<int> > > map2;
 
 						for (size_t i = 0 ; i < exclusions.size(); ++i)
 						{
 							int val = exclusions[i];
 							std::map<int, std::vector<int> > map3;
 							map3.insert( std::make_pair( val, std::vector<int>(1, val)) ) ;
 							map2.insert( std::make_pair( ch, map3 ));
 						}
 						exclusionmap.insert( std::make_pair(ch, map2));
 					}
 
 					currentPoint = c_excludes.find('=', currentPoint + 1) ;
 				}				
 			}
 		}
 
 		boost::dynamic_bitset<> bitset( c_instruction.size(), 0 ) ;
 		std::map<char, boost::dynamic_bitset<> > embedded_data ;
 		OutputInstructionData( 0, bitset, c_instruction, c_declaration, 
 			embedded_data, exclusionmap ) ;
 
 		output << "void " ;
 		output << c_declaration ;
 		output << c_signature ;
 		output << " ; " ;
 		output << std::endl ;
 
 	}
 
 	input.close() ;
 	output.close() ;
 
 	MergeInstructionData( ) ;
}

////////////////////////////////////////////////////////////////////

void M68K::SaveInstruction( boost::dynamic_bitset<> bitset, const std::string& declaration, 
						   std::map<char, boost::dynamic_bitset<> > embedded_data,
						   const t_exclusionMap& c_exclusionmap)
{
 //	check to see if any of the embedded data should be excluded
 	 	for (std::map<char, boost::dynamic_bitset<> >::const_iterator it = embedded_data.begin(),
 	 		itEnd = embedded_data.end(); it != itEnd; ++it)
 	 	{
 	 		const char c_char = (*it).first ;
 	 		t_exclusionMap::const_iterator itex = c_exclusionmap.find(c_char) ;
 	 		if (itex != c_exclusionmap.end())
 	 		{
 	 			for (std::multimap<char, std::map<int, std::vector<int> > >::const_iterator it2 = itex->second.begin(), it2End = itex->second.end(); it2 != it2End; ++it2)
 	 			{
 	 				if (embedded_data.find( it2->first) != embedded_data.end())
 	 				{
 	 					const size_t c_val = embedded_data[it2->first].to_ulong();
 	 					std::map<int, std::vector<int> >::const_iterator itexlustions = it2->second.find(c_val);
 	 					if (itexlustions != it2->second.end())
 	 					{
 	 						const size_t c_mainVal = it->second.to_ulong();
 	 						if (std::find(itexlustions->second.begin(), itexlustions->second.end(),c_mainVal) != itexlustions->second.end())
 	 							return;
 	 					}
 	 				}
 	 			}
 	 		}
 	 	}
 	
 	
 		std::ofstream output;
 		output.open(
 			( boost::format("Temp/%1%.txt") % declaration ).str().c_str() ,
 			std::ios_base::app) ;
 	
 		if (!output.is_open())
 		{
 			return ;
 		}
 	
 		output << "m68k->AddOpcodeFunction(" ;
 		output << bitset.to_ulong() ;
 		output << ", M68K::Invoker( &M68K::" ;
 		output << declaration ;
 	
 		for (std::map<char, boost::dynamic_bitset<> >::const_iterator it = embedded_data.begin(),
 			itEnd = embedded_data.end(); it != itEnd; ++it)
 		{
 			output << ", " ;
 			output << (*it).second.to_ulong() ;
 		}
 	
 	
 		output << ") ) ; " << std::endl ;
 	
 		output.close() ;
}

////////////////////////////////////////////////////////////////////

void M68K::OutputInstructionData( int position, boost::dynamic_bitset<> bitset, 
								 const std::string& instruction, 
								 const std::string& declaration,
								 std::map<char, boost::dynamic_bitset<> > embedded_data,
								 const t_exclusionMap& c_exclusionmap)
{	
 	const char c_char = instruction[position] ;
 	const bool c_lastchar = (instruction.size() - 1) == position ;
 
 	if (c_char == '0' || c_char == '1') // static
 	{
 		bitset[instruction.size() - 1 - position] = c_char == '1' ;
 
 		if (c_lastchar)
 		{
 			SaveInstruction( bitset, declaration, embedded_data, c_exclusionmap ) ;
 		}
 		else
 		{
 			OutputInstructionData( position+1, bitset, 
 				instruction, declaration, embedded_data, c_exclusionmap ) ;
 		}
 	}
 	else // dynamic
 	{
 		for (int i = 0; i < 2; ++i)
 		{
 			std::map<char, boost::dynamic_bitset<> > newdata = embedded_data ;
 			boost::dynamic_bitset<> data = newdata[c_char] ;
 		
 			data.resize( data.size() + 1, 0) ;
 			data = data << 1 ;
 			data[0] = i == 1 ;
 			newdata[c_char] = data ;
 			bitset[instruction.size() - 1 - position] = i == 1 ;
 			if (c_lastchar)
 			{
 				SaveInstruction( bitset, declaration, newdata, c_exclusionmap ) ;
 			}
 			else
 			{
 				OutputInstructionData( position+1, bitset, instruction, 
 					declaration, newdata, c_exclusionmap ) ;
 			}
 		}			
 	}
}

////////////////////////////////////////////////////////////////////

void M68K::AddOpcodeFunction(UDWORD instruction, const Invoker& function)
{
	if (m_OpcodeTable[instruction] == Invoker( &M68K::OpcodeInvalid ))
	{
		m_OpcodeTable[instruction] = function ;
	}
	else
	{
		codeslinger::error::Throw( "M68K::AddOpcodeFunction", ( boost::format( "Instruction %1%"
			" already added to opcode table") % instruction).str() ) ;
	}
}

////////////////////////////////////////////////////////////////////

void M68K::InitOpcodeTable()
{
	for (UDWORD i = 0 ; i < 0x10000; ++i)
	{
		m_OpcodeTable[i] = Invoker( &M68K::OpcodeInvalid ) ;
	}
}

////////////////////////////////////////////////////////////////////

M68K::EA_DATA M68K::GetEAOperand( UBYTE mode, UBYTE reg, DataSize size, bool readOnly, UDWORD addressOffset )
{
	// remember when using address registers to look up operands in memory only use lower 16 bits 
	// of register. Also when upading the register after use, only update lower 16 bits
	EA_DATA resData = EA_DATA( ) ;
	resData.dataSize = size ;

	switch( mode )
	{
	case 0 : 
		resData.operand = m_RegistersData[reg] ; 
		resData.eatype = EA_DATA_REG ;
		break ;
	case 1 : 
		resData.operand = m_RegistersAddress[reg] ; 
		resData.eatype = EA_ADDRESS_REG ;
		break ;
	case 2:
		{
			UDWORD address = m_RegistersAddress[reg] ;
			address += addressOffset;
			resData.pointer = address ;
			resData.eatype = EA_ADDRESS_REG_INDIRECT ;

			switch(size)
			{
			case SIZE_WORD:
				resData.cycles = 4 ;
				resData.operand = m_FuncReadMemory16( address ) ; 
				break ;
			case SIZE_LONG:
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory32( address ) ; 
				break ;
			case SIZE_BYTE:
				resData.cycles = 4 ;
				resData.operand = m_FuncReadMemory8( address) ;
				break ;
			}
		}
		break ;
	case 3:
		{
			UDWORD address = m_RegistersAddress[reg] ;
			resData.pointer = address ;
			resData.eatype = EA_ADDRESS_REG_INDIRECT_POST_INC ;

			switch(size)
			{
			case SIZE_WORD:
				resData.cycles = 4 ;
				resData.operand = m_FuncReadMemory16( address ) ; 
				if (!readOnly)
					m_RegistersAddress[reg]+= 2 ; 
				break ;
			case SIZE_LONG:
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory32( address ) ; 
				if (!readOnly)
					m_RegistersAddress[reg]+= 4 ;
				break ;
			case SIZE_BYTE:
				resData.cycles = 4 ;
				resData.operand = m_FuncReadMemory8( address ) ; 
				if (!readOnly)
					m_RegistersAddress[reg] += 1 ;
				if (reg == 7) // stack pointer is incremened by 2 to keep word aligned
				{
					if (!readOnly)
						m_RegistersAddress[reg] += 1 ;
				}
				break ;
			}
		}
		break ;
	case 4:
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT_PRE_DEC ;
			switch(size)
			{
			case SIZE_WORD:
				{
					resData.cycles = 6 ;
					UDWORD address = m_RegistersAddress[reg]-2 ;
					if (!readOnly)
						m_RegistersAddress[reg]-= 2 ;
					resData.pointer = address ;
					resData.operand = m_FuncReadMemory16( address ) ;
				}
				break ;
			case SIZE_LONG: 
				{
					resData.cycles = 10 ;
					UDWORD address = m_RegistersAddress[reg]-4 ;					
					if (!readOnly)
						m_RegistersAddress[reg]-= 4 ;
					resData.pointer = address ;
					resData.operand = m_FuncReadMemory32( address ) ;
				}
				break ;
			case SIZE_BYTE: 
				{
					resData.cycles = 6 ;
					int toreduce = 1;
					
					UDWORD address = m_RegistersAddress[reg]-toreduce ;
					if (!readOnly)
						m_RegistersAddress[reg]-= toreduce ;
					
					resData.pointer = address ;
					resData.operand = m_FuncReadMemory8( address ) ;
				}
				break ;
			}
		}
		break ;
	
	case 5:
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT_DISPLACEMENT ;
			UDWORD address = m_RegistersAddress[reg] + addressOffset;
			const UDWORD c_displacement = SignExtend( m_FuncReadMemory16( m_ProgramCounter ) ) ;
			const SDWORD test = c_displacement ;
			address += test ;
			resData.pointer = address ;
			resData.pcadvance = 2 ;
			switch(size)
			{
			case SIZE_BYTE:
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory8(address);
				break ;
			case SIZE_WORD:
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory16(address);
				break ;
			case SIZE_LONG:
				resData.cycles = 12 ;
				resData.operand = m_FuncReadMemory32(address);
				break ;
			}
		}
		break ;

	case 6:
		{
			// http://www.ee.unlv.edu/~b1morris/cpe300/docs/slides08_cisc.pdf
			// according to page 543 of Reference manual only "Address Register Indirect With Index (8-Bit Displacement) Mode"
			// is supported by 68k mode 6.
			resData.eatype = EA_ADDRESS_REG_INDIRECT_INDEX ;
			const UWORD c_data = m_FuncReadMemory16(m_ProgramCounter) ;
			resData.pcadvance = 2 ;
			const bool c_inDataReg = !TestBit(c_data, 15) ;
			const int c_indexReg = (c_data >> 12) & 0x7;
			const bool c_isLong = TestBit(c_data, 11) ;
			const UDWORD c_displacement = SignExtend( SignExtend( UBYTE( c_data & 0xFF ) ) ) ;
			resData.pointer = c_inDataReg ? m_RegistersData[ c_indexReg ] : m_RegistersAddress[ c_indexReg ] ;
			if (!c_isLong)
			{
				resData.pointer = SignExtend(UWORD( resData.pointer & 0xFFFF)) ;
			}

			// couldn't find any info on how to get scale so referenced MUSASHI m68k code
			int scale = (c_data >> 9) & 0x3;
			switch( scale )
			{
			case 0: break; // scale of 1 so dont do anything
			case 1: resData.pointer *= 2; break;
			case 2: resData.pointer *= 4; break;
			case 3: resData.pointer *= 8; break;
			default:
				LogMessage("M68k", "err", "Impossible scale value");
				break;
			}

			resData.pointer += m_RegistersAddress[reg] ;			
			resData.pointer += c_displacement ;
			resData.pointer += addressOffset;
			switch( size )
			{
				case SIZE_BYTE:
					resData.cycles = 10 ;
					resData.operand = m_FuncReadMemory8(resData.pointer) ;
					break ;
				case SIZE_WORD:
					resData.cycles = 10 ;
					resData.operand = m_FuncReadMemory16(resData.pointer) ;
					break ;
				case SIZE_LONG:
					resData.cycles = 14 ;
					resData.operand = m_FuncReadMemory32(resData.pointer) ;
					break ;
			}
		}
		break ;

		// when doing 6 dont forget cycles this update
	case 7 :
		resData.eatype = EA_MODE_7 ;
		switch(reg)
		{
		case 0:
			resData.mode7Type = EA_MODE_7_ABS_ADDRESS_SHORT ;
			resData.pointer = SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) + addressOffset;
			resData.pcadvance = 2 ;
			if (size == SIZE_BYTE)
			{
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory8(resData.pointer) ;
			}
			else if (size == SIZE_WORD)
			{
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory16(resData.pointer) ;
			}
			else
			{
				resData.cycles = 12 ;
				resData.operand = m_FuncReadMemory32(resData.pointer) ;
			}
			break ;
		case 1:
			resData.mode7Type = EA_MODE_7_ABS_ADDRESS_LONG ;
			resData.pointer = ( m_FuncReadMemory16(m_ProgramCounter) << 16 ) |
					m_FuncReadMemory16(m_ProgramCounter+2) ;
			resData.pcadvance = 4 ;
			resData.pointer += addressOffset;
			if (size == SIZE_BYTE)
			{
				resData.cycles = 12 ;
				resData.operand = m_FuncReadMemory8(resData.pointer) ;
			}
			else if (size == SIZE_WORD)
			{
				resData.cycles = 12 ;
				resData.operand = m_FuncReadMemory16(resData.pointer) ;
			}
			else
			{
				resData.cycles = 16 ;
				resData.operand = m_FuncReadMemory32(resData.pointer) ;
			}
			break ;
		case 2:
			resData.mode7Type = EA_MODE_7_PC_WITH_DISPLACEMENT ;
			resData.pointer = m_ProgramCounter + SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) + addressOffset ;
			resData.pcadvance = 2 ;
			if (size == SIZE_BYTE)
			{
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory8(resData.pointer) ;
			}
			else if (size == SIZE_WORD)
			{
				resData.cycles = 8 ;
				resData.operand = m_FuncReadMemory16(resData.pointer) ;
			}
			else
			{
				resData.cycles = 12 ;
				resData.operand = m_FuncReadMemory32(resData.pointer) ;
			}
			break ;
		case 3:
			{
				// http://www.ee.unlv.edu/~b1morris/cpe300/docs/slides08_cisc.pdf
				// according to page 543 of Reference manual only "Program Counter Indirect With Index (8-Bit Displacement) Mode"
				// is supported by 68k for ea mode 7 reg 3. 
				resData.mode7Type = EA_MODE_7_PC_WITH_PREINDEX ;
				const UWORD c_data = m_FuncReadMemory16(m_ProgramCounter) ;
				const bool c_inDataReg = !TestBit(c_data, 15) ;
				const int c_indexReg = (c_data >> 12) & 0x7;
				const bool c_isLong = TestBit(c_data, 11) ;
				const UDWORD c_displacement = SignExtend( SignExtend( UBYTE( c_data & 0xFF ) ) ) ;
				resData.pointer = c_inDataReg ? m_RegistersData[ c_indexReg ] : m_RegistersAddress[ c_indexReg ] ;
				if (!c_isLong)
				{
					resData.pointer = SignExtend(UWORD( resData.pointer & 0xFFFF)) ;
				}
				// couldn't find any info on how to get scale so referenced MUSASHI m68k code
				int scale = (c_data >> 9) & 0x3;
				switch( scale )
				{
				case 0: break; // scale of 1 so dont do anything
				case 1: resData.pointer *= 2; break;
				case 2: resData.pointer *= 4; break;
				case 3: resData.pointer *= 8; break;
				default:
					LogMessage("M68k", "err", "Impossible scale value");
					break;
				}
				resData.pointer += m_ProgramCounter ;
				resData.pcadvance = 2;
				resData.pointer += c_displacement ;
				resData.pointer += addressOffset;
				switch( size )
				{
				case SIZE_BYTE:
					resData.cycles = 10 ;
					resData.operand = m_FuncReadMemory8(resData.pointer) ;
					break ;
				case SIZE_WORD:
					resData.cycles = 10 ;
					resData.operand = m_FuncReadMemory16(resData.pointer) ;
					break ;
				case SIZE_LONG:
					resData.cycles = 14 ;
					resData.operand = m_FuncReadMemory32(resData.pointer) ;
					break ;
				}
			}
			break;
			case 4:
				resData.mode7Type = EA_MODE_7_IMMEDIATE_DATA ;
				resData.pointer = m_ProgramCounter ;

				if (size == SIZE_BYTE)
				{
					resData.cycles = 4 ;
					resData.operand = m_FuncReadMemory8(m_ProgramCounter+1) ;
					resData.pcadvance = 2 ;
				}
				else if (size == SIZE_WORD)
				{
					resData.cycles = 4 ;
					resData.operand = m_FuncReadMemory16(m_ProgramCounter) ;
					resData.pcadvance = 2 ;
				}
				else
				{
					resData.cycles = 8 ;
					resData.operand = m_FuncReadMemory32(m_ProgramCounter) ;
					resData.pcadvance = 4 ;
				}
				break ;
			default:
				ASSERT(false) ;
				LogMessage("M68k","wrn", ( 
					boost::format("Yet to implement EA mode 7 with reg %1%") % reg ).str()) ;
				break ;
		}
		break ;
	
	default:
		ASSERT(false) ;
		LogMessage("M68k","wrn", ( boost::format("Yet to implement EA mode %1%") % mode ).str()) ;
		break ;
	}
	return resData ;
}

////////////////////////////////////////////////////////////////////

M68K::EA_DATA M68K::SetEAOperand( UBYTE mode, UBYTE reg, UDWORD data, DataSize size, UDWORD addressOffset )
{
	// remember when using address registers to look up operands in memory only use lower 16 bits 
	// of register. Also when upading the register after use, only update lower 16 bits

	EA_DATA resData = EA_DATA( ) ;
	resData.dataSize = size ;

	switch( mode )
	{
	case 0 : 
		SetDataRegistry(reg, data, size);
		resData.eatype = EA_DATA_REG ;
		break ;
	case 1 : 
		SetAddressRegistry(reg, data, size);
		resData.eatype = EA_ADDRESS_REG ;
		break ;

	case 2:
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT ;

			UDWORD address = m_RegistersAddress[reg] + addressOffset;
			resData.pointer = address ;

			switch(size)
			{
			case SIZE_WORD: 
				resData.cycles = 4 ;
				m_FuncWriteMemory16( address, (UWORD)data ) ; 
				break ;
			case SIZE_LONG:
				resData.cycles = 8 ;
				m_FuncWriteMemory32( address, data ) ; 
				break ;
			case SIZE_BYTE:
				resData.cycles = 4 ;
				m_FuncWriteMemory8( address, (UBYTE)data ) ; 
				break ;
			}
		}
		break ;
	case 3:
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT_POST_INC ;

			UDWORD address = m_RegistersAddress[reg] ;
			resData.pointer = address ;

			switch(size)
			{
			case SIZE_WORD:
				resData.cycles = 4 ;
				m_FuncWriteMemory16( address, (UWORD)data ) ; 
				m_RegistersAddress[reg]+= 2 ; 
				break ;
			case SIZE_LONG:
				resData.cycles = 8 ;
				m_FuncWriteMemory32( address, data ) ; 
				m_RegistersAddress[reg]+= 4 ;
				break ;
			case SIZE_BYTE: 
				resData.cycles = 4 ;
				m_FuncWriteMemory8( address, (UBYTE)data ) ; 
				m_RegistersAddress[reg]+= 1 ;
				if (reg == 7) // keep stack aligned with word
				{
					m_RegistersAddress[reg] += 1 ;
				}
				break ;
			}
		}
		break ;
	case 4:
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT_PRE_DEC ;
			switch(size)
			{
			case SIZE_WORD:
				{
					resData.cycles = 6 ;
					m_RegistersAddress[reg]-= 2 ;
					UDWORD address = m_RegistersAddress[reg] ;
					resData.pointer = address ;
					m_FuncWriteMemory16( address, (UWORD)data ) ;
				}
				break ;
			case SIZE_LONG: 
				{
					resData.cycles = 10 ;
					m_RegistersAddress[reg]-= 4 ;
					UDWORD address = m_RegistersAddress[reg] ;
					resData.pointer = address ;
					m_FuncWriteMemory32( address, data ) ;
				}
				break ;
			case SIZE_BYTE: 
				{	
					resData.cycles = 6 ;
					m_RegistersAddress[reg]-= 1 ;
					if (reg == 7) // keep stack word aligned
					{
						m_RegistersAddress[reg] -= 1 ;
					}
					UDWORD address = m_RegistersAddress[reg] ;
					resData.pointer = address ;
					m_FuncWriteMemory8( address, (UBYTE)data ) ;
				}
				break ;
			}
		}
		break ;
	case 5: 
		{
			resData.eatype = EA_ADDRESS_REG_INDIRECT_DISPLACEMENT ;
			const UDWORD c_immediate = SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) ;
			const UDWORD address = c_immediate + m_RegistersAddress[reg] + addressOffset ;
			resData.pointer = address ;

			switch(size)
			{
			case SIZE_WORD:
				resData.cycles = 8 ;
				m_FuncWriteMemory16( address, (UWORD)data ) ; 
				break ;
			case SIZE_LONG:
				resData.cycles = 12 ;
				m_FuncWriteMemory32( address, data ) ; 
				break ;
			case SIZE_BYTE:
				resData.cycles = 8 ;
				m_FuncWriteMemory8(address, (UBYTE)data) ;
				break ;
			}

			resData.pcadvance = 2 ;
		}
		break ;
	case 6:
		{
			// http://www.ee.unlv.edu/~b1morris/cpe300/docs/slides08_cisc.pdf
			// according to page 543 of Reference manual only "Address Register Indirect With Index (8-Bit Displacement) Mode"
			// is supported by 68k mode 6.
			resData.eatype = EA_ADDRESS_REG_INDIRECT_INDEX ;
			const UWORD c_data = m_FuncReadMemory16(m_ProgramCounter) ;
			resData.pcadvance = 2 ;
			const bool c_inDataReg = !TestBit(c_data, 16) ;
			const int c_indexReg = (c_data >> 12) & 0x7;
			const bool c_isLong = TestBit(c_data, 11) ;
			const UDWORD c_displacement = SignExtend( SignExtend( UBYTE( c_data & 0xFF ) ) ) ;
			resData.pointer = c_inDataReg ? m_RegistersData[ c_indexReg ] : m_RegistersAddress[ c_indexReg ] ;
			
			if (!c_isLong)
			{
				resData.pointer = SignExtend(UWORD( resData.pointer & 0xFFFF)) ;
			}
			// couldn't find any info on how to get scale so referenced MUSASHI m68k code
			int scale = (c_data >> 9) & 0x3;
			switch( scale )
			{
			case 0: break; // scale of 1 so dont do anything
			case 1: resData.pointer *= 2; break;
			case 2: resData.pointer *= 4; break;
			case 3: resData.pointer *= 8; break;
			default:
				LogMessage("M68k", "err", "Impossible scale value");
				break;
			}

			resData.pointer += m_RegistersAddress[reg] ;
			resData.pointer += c_displacement ;
			resData.pointer += addressOffset;
			switch( size )
			{
			case SIZE_BYTE:
				resData.cycles = 10 ;
				m_FuncWriteMemory8(resData.pointer, (UBYTE)data) ;
				break ;
			case SIZE_WORD:
				resData.cycles = 10 ;
				m_FuncWriteMemory16(resData.pointer, (UWORD)data) ;
				break ;
			case SIZE_LONG:
				resData.cycles = 14 ;
				m_FuncWriteMemory32(resData.pointer, data) ;
				break ;
			}
		}
		break ;
	case 7 :
		resData.eatype = EA_MODE_7 ;
		switch(reg)
		{
		case 0:
			resData.mode7Type = EA_MODE_7_ABS_ADDRESS_SHORT ;
			resData.pointer = SignExtend( m_FuncReadMemory16(m_ProgramCounter) ) + addressOffset;
			resData.pcadvance = 2 ;
			if (size == SIZE_BYTE)
			{
				resData.cycles = 8 ;
				m_FuncWriteMemory8(resData.pointer, (UBYTE)data) ;
			}
			else if (size == SIZE_WORD)
			{
				resData.cycles = 8 ;
				m_FuncWriteMemory16(resData.pointer, (UWORD)data) ;
			}
			else
			{
				resData.cycles = 12 ;
				m_FuncWriteMemory32(resData.pointer, data) ;
			}
			break ;
		case 1:
			resData.mode7Type = EA_MODE_7_ABS_ADDRESS_LONG ;
			resData.pointer = ( m_FuncReadMemory16(m_ProgramCounter) << 16 ) |
				m_FuncReadMemory16(m_ProgramCounter+2) ;
			resData.pointer += addressOffset;
			resData.pcadvance = 4 ;

			if (size == SIZE_BYTE)
			{
				resData.cycles = 12 ;
				m_FuncWriteMemory8(resData.pointer, (UBYTE)data) ;
			}
			else if (size == SIZE_WORD)
			{
				resData.cycles = 12 ;
				m_FuncWriteMemory16(resData.pointer, (UWORD)data) ;
			}
			else
			{
				resData.cycles = 16 ;
				m_FuncWriteMemory32(resData.pointer, data) ;
			}
			break ;
		case 3:
			{
				// http://www.ee.unlv.edu/~b1morris/cpe300/docs/slides08_cisc.pdf
				// according to page 543 of Reference manual only "Program Counter Indirect With Index (8-Bit Displacement) Mode"
				// is supported by 68k for ea mode 7 reg 3.
				resData.mode7Type = EA_MODE_7_PC_WITH_PREINDEX ;
				const UWORD c_data = m_FuncReadMemory16(m_ProgramCounter) ;
				const bool c_inDataReg = !TestBit(c_data, 16) ;
				const int c_indexReg = (c_data >> 12) & 0x7;
				const bool c_isLong = TestBit(c_data, 11) ;
				const UDWORD c_displacement = SignExtend( SignExtend( UBYTE( c_data & 0xFF ) ) ) ;
				resData.pointer = c_inDataReg ? m_RegistersData[ c_indexReg ] : m_RegistersAddress[ c_indexReg ] ;
				
				if (!c_isLong)
				{
					resData.pointer = SignExtend(UWORD( resData.pointer & 0xFFFF)) ;
				}
			
				// couldn't find any info on how to get scale so referenced MUSASHI m68k code
				int scale = (c_data >> 9) & 0x3;
				switch( scale )
				{
					case 0: break; // scale of 1 so dont do anything
					case 1: resData.pointer *= 2; break;
					case 2: resData.pointer *= 4; break;
					case 3: resData.pointer *= 8; break;
					default:
						LogMessage("M68k", "err", "Impossible scale value");
						break;
				}
				
				resData.pointer +=  m_FuncReadMemory16(m_ProgramCounter+2) ;
				resData.pcadvance = 4 ;
				resData.pointer += c_displacement ;
				resData.pointer += addressOffset;
				switch( size )
				{
				case SIZE_BYTE:
					resData.cycles = 10 ;
					m_FuncWriteMemory8(resData.pointer, (UBYTE)data) ;
					break ;
				case SIZE_WORD:
					resData.cycles = 10 ;
					m_FuncWriteMemory16(resData.pointer, (UWORD)data) ;
					break ;
				case SIZE_LONG:
					resData.cycles = 14 ;
					m_FuncWriteMemory32(resData.pointer, data) ;
					break ;
				}
			}
		break ;
		
		default:
			ASSERT(false) ;
			LogMessage("M68k","wrn", ( 
				boost::format("Yet to implement EA mode 7 with reg %1%") % reg ).str()) ;
			break ;
		}
		break ;
	default:
		ASSERT(false) ;
		LogMessage("M68k","wrn", ( boost::format("Yet to implement EA mode %1%") % mode ).str()) ;
		break ;
	}
	return resData ;
}

////////////////////////////////////////////////////////////////////

UWORD M68K::SignExtend( UBYTE byte )
{
	UWORD res = byte ;
	if (codeslinger::maths::TestBit(byte, 7))
	{
		res |= 0xFF00 ;
	}
	return res ;
}

////////////////////////////////////////////////////////////////////

UDWORD M68K::SignExtend(UWORD word)
{
	UDWORD res = word ;
	if (codeslinger::maths::TestBit(word, 15))
	{
		res |= 0xFFFF0000 ;
	}
	return res ;
}

////////////////////////////////////////////////////////////////////

UDWORD M68K::GetMaxSize(DataSize ds) const
{
	int res = 0 ;
	switch(ds)
	{
	case SIZE_BYTE: res = 0xFF; break;
	case SIZE_WORD: res = 0xFFFF; break;
	case SIZE_LONG: res = 0xFFFFFFFF; break;
	}
	return res;
}

////////////////////////////////////////////////////////////////////

void M68K::CheckAndChangeCAdd(UDWORD dataLeft, UDWORD dataRight, DataSize ds, bool includeX)
{
	const UQWORD c_maxSize = GetMaxSize(ds);
	dataLeft &= c_maxSize;
	dataRight &= c_maxSize;
	UQWORD bigAssNumber = (UQWORD)dataLeft + (UQWORD)dataRight;
	if (includeX && TestBit(m_CCR, FLAG_X))
		++bigAssNumber;
	m_CCR = codeslinger::maths::BitConditionalSet( bigAssNumber > (UQWORD)c_maxSize, m_CCR, FLAG_C );
}

////////////////////////////////////////////////////////////////////

void M68K::CheckAndChangeCSub(UDWORD dataLeft, UDWORD dataRight, DataSize ds, bool includeX)
{
	const int c_maxSize = GetMaxSize(ds);
	dataLeft &= c_maxSize;
	dataRight &= c_maxSize;
	UQWORD bigAssNumber1 = dataRight;
	if (includeX && TestBit(m_CCR, FLAG_X))
		++bigAssNumber1;
	UQWORD bigAssNumber2 = dataLeft;

	m_CCR = codeslinger::maths::BitConditionalSet( bigAssNumber1 > bigAssNumber2, m_CCR, FLAG_C );
}

////////////////////////////////////////////////////////////////////
// http://teaching.idallen.com/dat2343/10f/notes/040_overflow.txt 
void M68K::CheckAndChangeVAdd(SDWORD dataLeft, SDWORD dataRight, DataSize ds, bool includeX)
{
	switch(ds)
	{
	case SIZE_BYTE:
		{
			SBYTE result = (SBYTE)dataLeft + (SBYTE)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				++result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,7) == TestBit(dataRight, 7)) && (TestBit(dataLeft,7) != TestBit(result,7)), m_CCR, FLAG_V );
		}
		break;
	case SIZE_WORD: 
		{
			SWORD result = (SWORD)dataLeft + (SWORD)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				++result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,15) == TestBit(dataRight, 15)) && (TestBit(dataLeft,15) != TestBit(result,15)), m_CCR, FLAG_V );
		}
		break;
	case SIZE_LONG: 
		{
			SDWORD result = (SDWORD)dataLeft + (SDWORD)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				++result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,31) == TestBit(dataRight, 31)) && (TestBit(dataLeft,31) != TestBit(result,31)), m_CCR, FLAG_V );
		}
		break;
	}
}

////////////////////////////////////////////////////////////////////
// http://teaching.idallen.com/dat2343/10f/notes/040_overflow.txt 
void M68K::CheckAndChangeVSub(SDWORD dataLeft, SDWORD dataRight, DataSize ds, bool includeX)
{
	switch(ds)
	{
	case SIZE_BYTE:
		{
			SBYTE result = (SBYTE)dataLeft - (SBYTE)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				--result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,7) != TestBit(dataRight, 7)) && (TestBit(dataRight,7) == TestBit(result,7)), m_CCR, FLAG_V );
		}
		break;
	case SIZE_WORD: 
		{
			SWORD result = (SWORD)dataLeft - (SWORD)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				--result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,15) != TestBit(dataRight, 15)) && (TestBit(dataRight,15) == TestBit(result,15)), m_CCR, FLAG_V );
		}
		break;
	case SIZE_LONG: 
		{
			SDWORD result = (SDWORD)dataLeft - (SDWORD)dataRight;
			if (includeX && TestBit(m_CCR,FLAG_X))
				--result;
			m_CCR = codeslinger::maths::BitConditionalSet( 
				(TestBit(dataLeft,31) != TestBit(dataRight, 31)) && (TestBit(dataRight,31) == TestBit(result,31)), m_CCR, FLAG_V );
		}
		break;
	}
}


////////////////////////////////////////////////////////////////////

void  M68K::CheckAndChangeZ(UDWORD data, DataSize ds)
{
	const int c_andMask = GetMaxSize(ds);
	m_CCR = codeslinger::maths::BitConditionalSet( (data & c_andMask) == 0, m_CCR, FLAG_Z );
}

////////////////////////////////////////////////////////////////////

void M68K::CheckAndChangeN(UDWORD data, DataSize ds)
{
	int msb = 0 ;
	switch(ds)
	{
	case SIZE_BYTE: msb = 7; break;
	case SIZE_WORD: msb = 15; break;
	case SIZE_LONG: msb = 31; break;
	}

	m_CCR = codeslinger::maths::BitConditionalSet( codeslinger::maths::TestBit(data, msb), m_CCR, FLAG_N );
}

////////////////////////////////////////////////////////////////////

void M68K::MergeInstructionData()
{
 	std::ofstream output;
 	output.open("m68kopcodetable.txt") ;
 
 	std::vector<std::string> opcodes ;
 
 	boost::filesystem::directory_iterator end_itr; // default construction yields past-the-end
 
 	// loop through all files in this directory
 	for ( boost::filesystem::directory_iterator itr( "Temp" );itr != end_itr;++itr )
 	{
 		// make sure we are not looking at a symbolic link
 		if ( !boost::filesystem::symbolic_link_exists(itr->path()) && 
 			!boost::filesystem::is_directory(itr->status()))
 		{
 			const std::string& c_filename = itr->path().file_string() ;
 			const std::string& c_leaf = itr->path().leaf() ;
 			const std::string& c_ext = itr->path().extension() ;
 			const size_t c_extpos = c_leaf.find( c_ext ) ;
 			const std::string& c_instname = std::string( c_leaf.begin(), 
 				c_leaf.begin() + c_extpos ) ;
 
 			std::ifstream opcodefile( c_filename.c_str() ) ;
 
 
 			if (opcodefile.is_open())
 			{
 				int part = 1 ;
 				int count = 0 ;
 
 				const std::string c_funcName = ( boost::format("Add%1%_Part%2%") 
 					% c_instname % part ).str() ;
 				output << ( boost::format( "void %1%( const M68K::PTR& m68k )" )  
 					% c_funcName ).str().c_str() << std::endl ;
 				output << "{" << std::endl ;
 
 				opcodes.push_back(c_funcName) ;
 								
 				while (!opcodefile.eof())
 				{
 					char buffer[256] ;
 					opcodefile.getline( buffer, 256) ;
 					output << buffer << std::endl ;
 
 					count++ ;
 
 					if (count >= 1000)
 					{
 						count = 0 ;
 						++part ;
 						output << "}" << std::endl << std::endl ;
 						const std::string c_funcName = ( boost::format("Add%1%_Part%2%") 
 							% c_instname % part ).str() ;
 						output << ( boost::format( "void %1%( const M68K::PTR& m68k )" )  
 							% c_funcName ).str().c_str() << std::endl ;
 						output << "{" << std::endl ;
 						opcodes.push_back(c_funcName) ;
 					}
 				}
 				output << "}" << std::endl << std::endl ;
 				opcodefile.close() ;
 			}			
 
 			boost::filesystem::remove(itr->path()) ;
 		}		
 	}
 
 	output << "void M68K::CreateOpcodeTable( ) " << std::endl ;
 	output << "{" << std::endl ;
 	for (size_t i = 0 ; i < opcodes.size() ; ++i)
 	{
 		output << "\t" ;
 		output << opcodes[i] ;
 		output << "( shared_from_this() ) ;" << std::endl ;
 	}
 	output << "}" << std::endl ;
 
 	output.close() ;
}

////////////////////////////////////////////////////////////////////