#pragma once
#ifndef _M68K_H
#define _M68K_H

#include <boost/shared_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/function.hpp>
#include <boost/dynamic_bitset.hpp>
#include <map>
#include <set>

typedef boost::uint8_t UBYTE ;
typedef boost::int8_t SBYTE ;
typedef boost::uint16_t UWORD ;
typedef boost::int16_t SWORD ;
typedef boost::uint32_t UDWORD ;
typedef boost::int32_t SDWORD ;
typedef boost::uint64_t UQWORD ;
typedef boost::int64_t SQWORD ;



class M68K : public boost::enable_shared_from_this<M68K>
{
	typedef void (*t_FuncWriteMemory32)(UDWORD, UDWORD);
	typedef void (*t_FuncWriteMemory16)(UDWORD, UWORD);
	typedef void (*t_FuncWriteMemory8)(UDWORD, UBYTE);
	typedef UWORD (*t_FuncReadMemory16)(UDWORD) ;
	typedef UDWORD (*t_FuncReadMemory32)(UDWORD) ;
	typedef UBYTE (*t_FuncReadMemory8)(UDWORD) ;

	enum DataSize
	{
		SIZE_BYTE,
		SIZE_WORD,
		SIZE_LONG
	} ;

	enum EA_TYPES
	{
		EA_DATA_REG,
		EA_ADDRESS_REG,
		EA_ADDRESS_REG_INDIRECT,
		EA_ADDRESS_REG_INDIRECT_POST_INC,
		EA_ADDRESS_REG_INDIRECT_PRE_DEC,
		EA_ADDRESS_REG_INDIRECT_DISPLACEMENT,
		EA_ADDRESS_REG_INDIRECT_INDEX,
		EA_MODE_7
	};

	enum EA_MODE_7_TYPE
	{
		EA_MODE_7_INVALID = -1,
		EA_MODE_7_ABS_ADDRESS_SHORT = 0,
		EA_MODE_7_ABS_ADDRESS_LONG = 1,
		EA_MODE_7_PC_WITH_DISPLACEMENT = 2,
		EA_MODE_7_PC_WITH_PREINDEX = 3,

		EA_MODE_7_IMMEDIATE_DATA = 4,
		EA_MODE_7_NUM = 5
	};
	static const int c_numEa = EA_MODE_7 + EA_MODE_7_NUM;

	int m_MoveCyclesByteWord[c_numEa][c_numEa];
	int m_MoveCyclesLong[c_numEa][c_numEa];

	struct EA_DATA
	{
		int cycles ;
		UDWORD operand ;
		UDWORD pointer ;
		EA_TYPES eatype ;
		int pcadvance ;
		EA_MODE_7_TYPE mode7Type ;
		DataSize dataSize ;

		EA_DATA( ) : cycles(0), operand(0), pointer(0), eatype(EA_DATA_REG), pcadvance(0), mode7Type(EA_MODE_7_INVALID), dataSize(SIZE_WORD) {}
	};

public:
	class Invoker;

	enum InterruptType
	{
		INT_AutoVector_0 = 0, INT_AutoVector_1, INT_AutoVector_2, INT_AutoVector_3, INT_AutoVector_4, INT_AutoVector_5, INT_AutoVector_6, INT_AutoVector_7, INT_Trap, INT_PRIV_VIO 
	};

	typedef boost::shared_ptr<M68K> PTR ;

	static PTR Create( ) ;
	~M68K(void);

	void Init( ) ;
	void SetFuncWriteMemory32( const t_FuncWriteMemory32& func ) ;
	void SetFuncReadMemory32( const t_FuncReadMemory32& func ) ;
	void SetFuncWriteMemory16( const t_FuncWriteMemory16& func ) ;
	void SetFuncReadMemory16( const t_FuncReadMemory16& func ) ;
	void SetFuncWriteMemory8( const t_FuncWriteMemory8& func ) ;
	void SetFuncReadMemory8( const t_FuncReadMemory8& func ) ;
	int Update( ) ;

	void AddOpcodeFunction( UDWORD instruction, const Invoker& function ) ;

	void RequestAutoVectorInterrupt( InterruptType type ) ;

	static const UDWORD sc_clockcyclespersecond = 7670000 ;
private:
	t_FuncWriteMemory32 m_FuncWriteMemory32;
	t_FuncWriteMemory16 m_FuncWriteMemory16;
	t_FuncWriteMemory8 m_FuncWriteMemory8;
	t_FuncReadMemory16 m_FuncReadMemory16;
	t_FuncReadMemory32 m_FuncReadMemory32;
	t_FuncReadMemory8 m_FuncReadMemory8;

	M68K(void);


	void OutputOpcodeTable( ) ;

	typedef std::map< char, std::multimap<char, std::map<int, std::vector<int> > > > t_exclusionMap ;
	void OutputInstructionData( int position, boost::dynamic_bitset<> bitset, 
		const std::string& instruction,	const std::string& declaration, 
		std::map<char, boost::dynamic_bitset<> > embedded_data,
		const t_exclusionMap& c_exclusionmap) ;

	void SaveInstruction( boost::dynamic_bitset<> bitset, const std::string& declaration, 
		std::map<char, boost::dynamic_bitset<> > embedded_data,
		const t_exclusionMap& c_exclusionmap) ;


	void Reset( ) ;
	void InitOpcodeTable( ) ;

	void CreateOpcodeTable( ) ;
	void MergeInstructionData( ) ;

	EA_DATA GetEAOperand( UBYTE mode, UBYTE reg, DataSize size, bool readOnly, UDWORD addressOffest = 0 ) ;
	EA_DATA SetEAOperand( UBYTE mode, UBYTE reg, UDWORD data, DataSize size, UDWORD addressOffset = 0 ) ;

	UWORD SignExtend( UBYTE byte ) ;
	UDWORD SignExtend( UWORD word ) ;

	bool ConditionTest( int condition ) const ;
	void DebugDump();

	UDWORD GetMaxSize(DataSize ds) const;
	void CheckAndChangeZ(UDWORD data, DataSize ds);
	void CheckAndChangeN(UDWORD data, DataSize ds);
	void CheckAndChangeCAdd(UDWORD dataLeft, UDWORD dataRight, DataSize ds, bool includeX);
	void CheckAndChangeVAdd(SDWORD dataLeft, SDWORD dataRight, DataSize ds, bool includeX);
	void CheckAndChangeCSub(UDWORD dataLeft, UDWORD dataRight, DataSize ds, bool includeX);
	void CheckAndChangeVSub(SDWORD dataLeft, SDWORD dataRight, DataSize ds, bool includeX);

	void SetDataRegistry(int id, UDWORD result, DataSize ds);
	void SetAddressRegistry(int id, UDWORD result, DataSize ds);
	void CheckForInterrupts( );
	void CheckForPrivilegeChange( ) ;
	void RequestInterrupt( InterruptType level, int address ) ;
	void CreateMoveCycleTables( ) ;

	static const int FLAG_X = 4 ;
	static const int FLAG_N = 3 ;
	static const int FLAG_Z = 2 ;
	static const int FLAG_V = 1 ;
	static const int FLAG_C = 0 ;

	UDWORD m_ProgramCounter ;
	UDWORD m_ProgramCounterStart ;
	UDWORD m_CyclesThisUpdate ;
	UWORD m_CCR ;
	bool m_Stop ;
	bool m_servicingInterrupt ;

	std::multimap<InterruptType, int> m_interrupts;

	static const int c_numdataregisters = 8 ;
	static const int c_numaddressregisters = 8 ;
	UDWORD m_RegistersData[c_numdataregisters] ;
	UDWORD m_RegistersAddress[c_numaddressregisters] ;
	UDWORD m_SSP;
	UDWORD m_USP;
	bool m_InSuperVisorMode;

public:
	// opcodes
	void OpcodeInvalid( ) ;
	void OpcodeANDEASOURCE(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg ) ; 
	void OpcodeANDEADEST(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg ) ; 
	void OpcodeANDI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeANDICR( ) ; 
	void OpcodeANDISR( ) ; 
	void OpcodeOREASOURCE( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeOREADEST( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeORI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeORICR( ) ; 
	void OpcodeORISR( ) ; 
	void OpcodeEOR( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeEORI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeEORICCR( ) ; 
	void OpcodeEORISR( ) ; 
	void OpcodeNEG( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeNOT( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeADDEADest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeADDDNDest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeADDX( UBYTE regx, UBYTE size, UBYTE rm, UBYTE regy ) ; 
	void OpcodeADDI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeADDQ(UBYTE data, UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeADDA( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeSUBEADest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeSUBDNDest( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeSUBX( UBYTE regy, UBYTE size, UBYTE rm, UBYTE regx ) ; 
	void OpcodeSUBA( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeSUBQ( UBYTE data, UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeSUBI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeMOVE(UBYTE size, UBYTE destReg, UBYTE destMode, UBYTE sourceMode, UBYTE sourceReg ) ; 
	void OpcodeMOVEQ( UBYTE reg, UBYTE data ) ; 
	void OpcodeMOVEA(UBYTE size, UBYTE destReg, UBYTE sourceMode, UBYTE sourceReg ) ; 
	void OpcodeMOVEUSP(UBYTE direction, UBYTE reg ) ; 
	void OpcodeMoveToSR( UBYTE mode, UBYTE reg ) ; 
	void OpcodeMoveSR( UBYTE mode, UBYTE reg ) ; 
	void OpcodeMoveToCCR( UBYTE mode, UBYTE reg ) ; 
	void OpcodeMOVEP( UBYTE dataReg, UBYTE opmode, UBYTE addressReg ) ; 
	void OpcodeCMP( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeCMPA( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeCMPI( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeCMPM( UBYTE regAX, UBYTE size, UBYTE regAY ) ; 
	void OpcodeMULSWORD( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeMULUWORD( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeDIVS( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeDIVU( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeTST( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeBTSTDynamic( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeBTSTStatic(UBYTE mode, UBYTE reg ) ; 
	void OpcodeBSETStatic( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeBSETDynamic( UBYTE mode, UBYTE reg ) ; 
	void OpcodeBCLRStatic( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeBCLRDynamic( UBYTE mode, UBYTE reg ) ; 
	void OpcodeBCHGStatic( UBYTE reg, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeBCHGDynamic( UBYTE mode, UBYTE reg ) ; 
	void OpcodeLSReg( UBYTE countreg, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg ) ; 
	void OpcodeASReg( UBYTE countreg, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg ) ; 
	void OpcodeLSMem( UBYTE dir, UBYTE mode, UBYTE reg ) ; 
	void OpcodeASMem( UBYTE dir, UBYTE mode, UBYTE reg ) ; 
	void OpcodeRotateNoCarryReg( UBYTE count, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg ) ; 
	void OpcodeRotateCarryReg( UBYTE count, UBYTE dir, UBYTE size, UBYTE ir, UBYTE reg ) ; 
	void OpcodeRotateNoCarryMem( UBYTE dir, UBYTE mode, UBYTE reg ) ; 
	void OpcodeRotateCarryMem( UBYTE dir, UBYTE mode, UBYTE reg ) ; 
	void OpcodeJMP(UBYTE mode, UBYTE reg ) ; 
	void OpcodeBSR( UBYTE displacement ) ; 
	void OpcodeBcc( UBYTE condition, UBYTE displacement ) ; 
	void OpcodeDBcc( UBYTE condition, UBYTE reg ) ; 
	void OpcodeLEA( UBYTE regtoload, UBYTE mode, UBYTE reg ) ; 
	void OpcodeCLR( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeEXG( UBYTE regx, UBYTE opmode, UBYTE regy ) ; 
	void OpcodeEXT( UBYTE opmode, UBYTE reg ) ; 
	void OpcodeJSR( UBYTE mode, UBYTE reg ) ; 
	void OpcodePEA( UBYTE mode, UBYTE reg ) ; 
	void OpcodeLINKWORD( UBYTE reg ) ; 
	void OpcodeLINKLONG( UBYTE reg ) ; 
	void OpcodeUNLK( UBYTE reg ) ; 
	void OpcodeSWAP( UBYTE reg ) ; 
	void OpcodeTRAP( UBYTE vec ) ; 
	void OpcodeSTOP( ) ; 
	void OpcodeRTS( ) ; 
	void OpcodeNOP( ) ; 
	void OpcodeRTE( ) ; 
	void OpcodeMOVEM( UBYTE direction, UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeScc( UBYTE cc, UBYTE mode, UBYTE reg ) ; 
	void OpcodeABCD( UBYTE regx, UBYTE rm, UBYTE regy ) ; 
	void OpcodeNBCD( UBYTE mode, UBYTE reg ) ; 
	void OpcodeSBCD( UBYTE regy, UBYTE rm, UBYTE regx ) ; 
	void OpcodeCHK( UBYTE reg, UBYTE size, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeILLEGAL( ) ; 
	void OpcodeRTR( ) ; 
	void OpcodeTAS( UBYTE mode, UBYTE reg ) ; 
	void OpcodeTRAPV( ) ; 
	void OpcodeNEGX( UBYTE size, UBYTE mode, UBYTE reg ) ; 
	void OpcodeReset( ) ; 

	




	void OpcodeOR( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ; 
	void OpcodeAND(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg ) ; 
	void OpcodeADD(UBYTE datareg, UBYTE opmode, UBYTE mode, UBYTE reg ) ;
	void OpcodeSUB( UBYTE reg, UBYTE opmode, UBYTE mode, UBYTE eareg ) ;

	class Invoker
	{
		typedef void (M68K::*t_func0)(void) ;
		typedef void (M68K::*t_func1)(UBYTE data1) ;
		typedef void (M68K::*t_func2)(UBYTE data1, UBYTE data2) ;
		typedef void (M68K::*t_func3)(UBYTE data1, UBYTE data2, UBYTE data3) ;
		typedef void (M68K::*t_func4)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4) ;
		typedef void (M68K::*t_func5)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5) ;
		typedef void (M68K::*t_func6)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6) ;
		typedef void (M68K::*t_func7)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7) ;
		typedef void (M68K::*t_func8)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8) ;
		typedef void (M68K::*t_func9)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8, UBYTE data9) ;
		typedef void (M68K::*t_func10)(UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8, UBYTE data9, UBYTE data10) ;

	public:
		Invoker() : m_NumOperands(0), m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL)  {memset(&m_Data,0,sizeof(UBYTE) * 10);}
		Invoker(const t_func0& func) : m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func0(func), m_NumOperands(0) {memset(&m_Data,0,sizeof(UBYTE) * 10);}
		Invoker(const t_func1& func, UBYTE data1) : m_Func0(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func1(func), m_NumOperands(1){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1;	}
		Invoker(const t_func2& func, UBYTE data1, UBYTE data2) : m_Func0(NULL), m_Func1(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func2(func), m_NumOperands(2){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; }
		Invoker(const t_func3& func, UBYTE data1, UBYTE data2, UBYTE data3) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func3(func), m_NumOperands(3){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; }
		Invoker(const t_func4& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func4(func), m_NumOperands(4){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4;	}
		Invoker(const t_func5& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func5(func), m_NumOperands(5){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5;	}
		Invoker(const t_func6& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func6(func), m_NumOperands(6){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5; m_Data[5] = data6;	}
		Invoker(const t_func7& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(NULL), m_Func7(func), m_NumOperands(7){ memset(&m_Data,0,sizeof(UBYTE) * 10); m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5; m_Data[5] = data6; m_Data[6] = data7;	}
		Invoker(const t_func8& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func9(NULL), m_Func10(NULL), m_Func8(func), m_NumOperands(8){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5; m_Data[5] = data6; m_Data[6] = data7; m_Data[7] = data8;	}
		Invoker(const t_func9& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8, UBYTE data9) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func10(NULL), m_Func9(func), m_NumOperands(9){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5; m_Data[5] = data6; m_Data[6] = data7; m_Data[7] = data8; m_Data[8] = data9;	}
		Invoker(const t_func10& func, UBYTE data1, UBYTE data2, UBYTE data3, UBYTE data4, UBYTE data5, UBYTE data6, UBYTE data7, UBYTE data8, UBYTE data9, UBYTE data10) : m_Func0(NULL), m_Func1(NULL), m_Func2(NULL), m_Func3(NULL), m_Func4(NULL), m_Func5(NULL), m_Func6(NULL), m_Func7(NULL), m_Func8(NULL), m_Func9(NULL), m_Func10(func), m_NumOperands(10){ memset(&m_Data,0,sizeof(UBYTE) * 10);m_Data[0] = data1; m_Data[1] = data2; m_Data[2] = data3; m_Data[3] = data4; m_Data[4] = data5; m_Data[5] = data6; m_Data[6] = data7; m_Data[7] = data8; m_Data[8] = data9; m_Data[9] = data10;	}
	
		bool operator==( const Invoker& rhs )const
		{
			if (m_NumOperands != rhs.m_NumOperands) return false;
			for (int i = 0 ; i < 10; ++i)
			{
				if (m_Data[i] != rhs.m_Data[i])
					return false;
			}
			return (m_Func0 == rhs.m_Func0) && (m_Func1 == rhs.m_Func1) && (m_Func2 == rhs.m_Func2) && (m_Func3 == rhs.m_Func3) && (m_Func4 == rhs.m_Func4) && (m_Func5 == rhs.m_Func5) && (m_Func6 == rhs.m_Func6) && (m_Func7 == rhs.m_Func7) && (m_Func8 == rhs.m_Func8) && (m_Func9 == rhs.m_Func9) && (m_Func10 == rhs.m_Func10) ;
		}
	void Invoke(M68K* m68k)
	{
		switch (m_NumOperands)
		{
		case 0: (m68k->*m_Func0)(); break;
		case 1: (m68k->*m_Func1)(m_Data[0]); break;
		case 2: (m68k->*m_Func2)(m_Data[0], m_Data[1]); break;
		case 3: (m68k->*m_Func3)(m_Data[0], m_Data[1], m_Data[2]); break;
		case 4: (m68k->*m_Func4)(m_Data[0], m_Data[1], m_Data[2], m_Data[3]); break;
		case 5: (m68k->*m_Func5)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4]); break;
		case 6: (m68k->*m_Func6)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4], m_Data[5]); break;
		case 7: (m68k->*m_Func7)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4], m_Data[5], m_Data[6]); break;
		case 8: (m68k->*m_Func8)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4], m_Data[5], m_Data[6], m_Data[7]); break;
		case 9: (m68k->*m_Func9)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4], m_Data[5], m_Data[6], m_Data[7], m_Data[8]); break;
		case 10: (m68k->*m_Func10)(m_Data[0], m_Data[1], m_Data[2], m_Data[3], m_Data[4], m_Data[5], m_Data[6], m_Data[7], m_Data[8], m_Data[9]); break;
		default:
			throw( "Not enough operands ") ; break;
		}
	}
	private:
		UBYTE m_Data[10];		
		int m_NumOperands;
		t_func0 m_Func0;
		t_func1 m_Func1;
		t_func2 m_Func2;
		t_func3 m_Func3;
		t_func4 m_Func4;
		t_func5 m_Func5;
		t_func6 m_Func6;
		t_func7 m_Func7;
		t_func8 m_Func8;
		t_func9 m_Func9;
		t_func10 m_Func10;
	};
	private:
	Invoker m_OpcodeTable[0x10000] ;
};

#endif