#include "Config_MegaDrive.h"
#include "MegaDriveCartridge.h"
#include <cs_core/Strings/StringHelpers.h>
#include <cs_core/Strings/ci_string.h>
#include <cs_core/Utils/Maths.h>
#include <cs_windows/ZipFiles/unzip.h>

//////////////////////////////////////////////////////////

MegaDriveCartridge::PTR MegaDriveCartridge::Create( )
{
	return PTR( new MegaDriveCartridge() ) ;
}

//////////////////////////////////////////////////////////

MegaDriveCartridge::MegaDriveCartridge(void)
{
}

//////////////////////////////////////////////////////////

MegaDriveCartridge::~MegaDriveCartridge(void)
{
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadROMFile(const boost::filesystem::path& romFile )
{
	bool res = false ;
	if (boost::filesystem::exists(romFile) && boost::filesystem::is_regular_file(romFile) )
	{
		using codeslinger::string::ci_string ;

		ci_string ciExt = romFile.extension().c_str() ;

		if ( ciExt == ci_string(".bin") )
		{
			res = LoadBinFormat(romFile) ;
		}
		else if ( ciExt == ci_string(".md") )
		{
			res = LoadMDFormat(romFile) ;
		}
		else if ( ciExt == ci_string(".smd") )
		{
			res = LoadSMDFormat(romFile) ;
		}
		else if ( ciExt == ci_string(".1") )
		{
			res = LoadSplitSMDFormat(romFile) ;
		}
		else
		{
			LogMessage("err", ( boost::format("ERROR: Unrecognised ROM file extension %1%") 
				% romFile.extension() ).str() ) ;
		}
	}
	else
	{
		LogMessage("err", ( boost::format("ERROR: ROM file %1% does not exist or it is not a"
			" regular file")% romFile.extension() ).str() ) ;
	}

	return res ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::Init( const boost::filesystem::path& romFile )
{
	m_SaveRamAvailable = false;
	m_SaveRamStart = 0x0;
	m_SaveRamEnd = 0x0;
	LogMessage("info", ( boost::format("Attempting to load ROM file %1%") 
		% romFile.file_string() ).str() ) ;

	using codeslinger::string::ci_string ;

	ci_string ciExt = romFile.extension().c_str() ;

	bool res = false ;

	if (ciExt == ci_string(".zip") )
	{
		res = LoadZipFormat(romFile) ;
	}
	else
	{
		res = LoadROMFile( romFile ) ;
	}		

	if (res)
	{
		LoadROMHeader( ) ;
		res = ValidateROMHeader( ) ;
		if (res)
		{
			CreateSaveRam( ) ;
		}
	}

	return res ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadZipFormat(const boost::filesystem::path& romFile)
{
	LogMessage("info", "Examining ZIP file") ;

	codeslinger::windows::zip::HZIP hz = 
		codeslinger::windows::zip::OpenZip(romFile.file_string().c_str(), "");
	codeslinger::windows::zip::ZIPENTRY ze; 	
	codeslinger::windows::zip::GetZipItem(hz,-1,&ze);

	boost::filesystem::path toLoad ;

	int numitems=ze.index;
	for (int i=0; i<numitems; i++)
	{ 
		codeslinger::windows::zip::GetZipItem(hz,i,&ze);
		std::string filename = ( boost::format("c:/windows/temp/%1%") % ze.name ).str() ;
		LogMessage("info", ( boost::format("Extracting file to %1%") % filename.c_str() ).str() ) ;
		codeslinger::windows::zip::UnzipItem(hz,i, filename.c_str());
		
		boost::filesystem::path file = filename ;

		using codeslinger::string::ci_string ;

		ci_string ciExt = file.extension().c_str() ;
		if (ciExt == ci_string(".bin") || ciExt == ci_string(".md") || ciExt == ci_string(".smd")
			|| ciExt == ci_string(".1"))
		{
			toLoad = filename ;
		}
	}

	codeslinger::windows::zip::CloseZipU(hz);

	return LoadROMFile( toLoad ) ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadBinFormat( const boost::filesystem::path& romFile )
{
	LogMessage("info", "Loading ROM file from binary format") ;

	m_CartridgeSize = static_cast<long>( boost::filesystem::file_size(romFile) ) ;
	m_CartridgeMemory = boost::shared_array<UBYTE>( new UBYTE[m_CartridgeSize] ) ;

	FILE *in = NULL;
	in = fopen( romFile.file_string().c_str(), "rb" );
	fread(m_CartridgeMemory.get(), 1, m_CartridgeSize, in);
	fclose(in);

	LogMessage("info", "Successfully loaded ROM file from binary format") ;

	return true ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadMDFormat( const boost::filesystem::path& romFile )
{
	LogMessage("info", "Loading ROM file from MD format") ;

	m_CartridgeSize = static_cast<long>( boost::filesystem::file_size(romFile) );

	boost::shared_array<UBYTE> mdRom = boost::shared_array<UBYTE>( new UBYTE[m_CartridgeSize] ) ;

	FILE* in = fopen( romFile.file_string().c_str(), "rb" );
	fread(mdRom.get(), 1, m_CartridgeSize, in);
	fclose(in);

	m_CartridgeMemory = boost::shared_array<UBYTE>( new UBYTE[m_CartridgeSize] ) ;

	const long c_halfblocksize = m_CartridgeSize / 2 ;

	for (long i = 0 ; i < m_CartridgeSize; ++i)
	{		
		if (i < c_halfblocksize)
		{
			m_CartridgeMemory[(i*2)+1] = mdRom[i] ;
		}
		else
		{
			int position = (i * 2) - m_CartridgeSize  ;
			m_CartridgeMemory[position] = mdRom[i] ;
		}
	}

	LogMessage("info", "Successfully loaded ROM file from MD format") ;

	return true ;
}

//////////////////////////////////////////////////////////

boost::tribool MegaDriveCartridge::LoadSMDFile( const boost::filesystem::path& romFile, 
		boost::shared_array<UBYTE>& destination, long& memorySize )
{
	LogMessage("info", (boost::format( "Converting SMD file %1%") % romFile.file_string()).str() ) ;

	if ( !boost::filesystem::exists( romFile) || !boost::filesystem::is_regular_file(romFile) )
	{
		LogMessage("err", (boost::format( "ERROR: Could not open SMD file %1%. Verify it exists and"
			" it is a regular file") % romFile.file_string()).str() ) ;
		return false ;
	}
	
	const int c_smdheadersize = 512 ;
	
	long fileSize = static_cast<long>( boost::filesystem::file_size(romFile) ) ;

	memorySize = fileSize - c_smdheadersize ;

	boost::shared_array<UBYTE> smdRom = boost::shared_array<UBYTE>( new UBYTE[fileSize] ) ;
	destination = boost::shared_array<UBYTE>( new UBYTE[memorySize] ) ;

	FILE* in = fopen( romFile.file_string().c_str(), "rb" );
	fread(smdRom.get(), 1, fileSize, in);
	fclose(in);

	const int c_smdblocksize = 16384 ;
	int numBlocks = (fileSize - 512) / c_smdblocksize  ;

	for (int i = 0 ; i < numBlocks; ++i)
	{
		int even = 0 ;
		int odd = 1 ;
		UBYTE smdBlock[c_smdblocksize] ;
		UBYTE binBlock[c_smdblocksize] ;

		memcpy( smdBlock, smdRom.get() + c_smdheadersize + ( i * c_smdblocksize), c_smdblocksize ) ;

		const int c_halfblocksize = c_smdblocksize / 2 ;

		for (int j = 0; j < c_halfblocksize; ++j) 
		{ 
			binBlock[odd] = smdBlock[j];                 
			binBlock[even] = smdBlock[j+c_halfblocksize]; 
			odd += 2; 
			even += 2; 
		} 

		memcpy( destination.get() + (i * c_smdblocksize), binBlock, c_smdblocksize ) ;
	}
	LogMessage("info", (boost::format( "Finished Converting SMD file %1%") 
		% romFile.file_string()).str() ) ;

	boost::tribool res ;

	if (smdRom[0x2] == 0x0) // is this a singular smd or the last smd in a split
	{
		res = true ;
	}
	else
	{
		res = boost::indeterminate ; // still need to read more split
	}

	return res ; 
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadSMDFormat( const boost::filesystem::path& romFile )
{
	LogMessage("info", "Loading ROM file from SMD format") ;

	LoadSMDFile(romFile, m_CartridgeMemory, m_CartridgeSize) ;	

	LogMessage("info", "Successfully loaded ROM file from SMD format") ;

	return true ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::LoadSplitSMDFormat(const boost::filesystem::path& romFile)
{
	LogMessage("info", "Loading ROM file from split SMD format") ;

	boost::filesystem::path splitPath = romFile ;

	std::vector< std::pair< boost::shared_array<UBYTE>, long > > memoryBlocks ;

	boost::shared_array<UBYTE> memoryBlock ;

	long memorySize = 0 ;
	m_CartridgeSize = 0 ;

	boost::tribool loadRes = boost::indeterminate ;

	while( boost::indeterminate( loadRes ) )
	{
		 loadRes = LoadSMDFile(splitPath, memoryBlock, memorySize) ;

		// if we loaded this file
		if (loadRes == true || boost::indeterminate(loadRes))
		{
			m_CartridgeSize += memorySize ;

			memoryBlocks.push_back( std::make_pair( memoryBlock, memorySize ) ) ;
			memoryBlock.reset( ) ;
		}

		// if there is still more to process
		if (boost::indeterminate(loadRes))
		{
			std::string str = splitPath.file_string() ;
			size_t pos = str.rfind('.') ;
			if (pos != std::string::npos)
			{
				std::string filename = std::string(str.begin(), str.begin() + pos ) ;
				std::string szVal = std::string( str.begin() + pos + 1, str.end() ) ;
				int value = boost::lexical_cast<int>( szVal ) ;
				++value ;
				splitPath = ( boost::format( "%1%.%2%" ) % filename % value ).str() ;
			}
			else
			{
				LogMessage("err", (boost::format( "ERROR:Could not find fullstop in SMD file"
					" path %1%") % romFile.file_string()).str() ) ;
				return false ;
			}
		}
		
		// something went wrong loading
		if (loadRes == false)
		{
			return false ;
		}
	}

	m_CartridgeMemory = boost::shared_array<UBYTE>( new UBYTE[m_CartridgeSize] ) ;

	long currentPos = 0 ;
	for (size_t i = 0 ; i < memoryBlocks.size(); ++i)
	{
		long numBytes = memoryBlocks[i].second ;
		memcpy( m_CartridgeMemory.get() + currentPos , memoryBlocks[i].first.get(), numBytes ) ;
		currentPos += numBytes ;
	}

	LogMessage("info", "Successfully loaded ROM file from split SMD format") ;

	return true ;
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::LoadROMHeader( )
{
	m_RomHeader = *((RomHeader*)&m_CartridgeMemory[0x100]) ;

	codeslinger::maths::SwapEndian( m_RomHeader.checksum ) ;
	codeslinger::maths::SwapEndian( m_RomHeader.rom_start ) ;
	codeslinger::maths::SwapEndian( m_RomHeader.rom_end ) ;
	codeslinger::maths::SwapEndian( m_RomHeader.ram_start ) ;
	codeslinger::maths::SwapEndian( m_RomHeader.ram_end ) ;
}

void MegaDriveCartridge::CreateSaveRam()
{
	if(m_RomHeader.sram[0x0] == 'R' && m_RomHeader.sram[0x1] == 'A')
	{
		m_SaveRamStart = m_RomHeader.sram[0x4] << 24 | m_RomHeader.sram[0x5] << 16 | 
			m_RomHeader.sram[0x6] << 8  | m_RomHeader.sram[0x7];
		m_SaveRamEnd = m_RomHeader.sram[0x8] << 24 | m_RomHeader.sram[0x9] << 16 |
			m_RomHeader.sram[0xa] << 8  | m_RomHeader.sram[0xb];
	
 		if(m_SaveRamStart != m_SaveRamEnd) 
		{
 			if(m_SaveRamStart & 1)
			{	
				--m_SaveRamStart;
			}
 			if(!(m_SaveRamEnd & 1))
			{
				++m_SaveRamEnd;
			}
			m_SaveRamLength = m_SaveRamEnd - (m_SaveRamStart - 1);
			UBYTE* raw = new UBYTE[m_SaveRamLength];
			memset(raw,0xff,m_SaveRamLength);
 			m_SaveRam = boost::shared_array<UBYTE>( raw ) ;
			m_SaveRamAvailable = true;
			LoadSaveRam( );
 		}
	}

	if (m_RomHeader.rom_start != 0)
	{
		LogMessage("wrn", "Rom Start value does not equal 0. If having problems emulating this game"
			 " start here.") ;
	}
	DumpROMHeader( ) ;
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::DumpROMHeader()
{
	std::string consoleName( &m_RomHeader.segatitle[0], &m_RomHeader.segatitle[16] ) ;
	std::string copyRight( &m_RomHeader.copyright[0], &m_RomHeader.copyright[16] ) ;
	std::string domestic( &m_RomHeader.domestic[0], &m_RomHeader.domestic[48] ) ;
	std::string overseas( &m_RomHeader.overseas[0], &m_RomHeader.overseas[48] ) ;
	std::string production( &m_RomHeader.production[0], &m_RomHeader.production[14] ) ;
	std::string control( &m_RomHeader.control[0], &m_RomHeader.control[16] ) ;
	std::string modem( &m_RomHeader.modem[0], &m_RomHeader.modem[12] ) ;
	std::string memo( &m_RomHeader.memo[0], &m_RomHeader.memo[40] ) ;
	std::string country( &m_RomHeader.country[0], &m_RomHeader.country[16] ) ;
	
	consoleName = codeslinger::string::NukeSpaces( consoleName ) ;
	copyRight = codeslinger::string::NukeSpaces( copyRight ) ;
	domestic = codeslinger::string::NukeSpaces( domestic ) ;
	overseas = codeslinger::string::NukeSpaces( overseas ) ;
	production = codeslinger::string::NukeSpaces( production ) ;
	control = codeslinger::string::NukeSpaces( control ) ;
	modem = codeslinger::string::NukeSpaces( modem ) ;
	memo = codeslinger::string::NukeSpaces( memo ) ;
	country = codeslinger::string::NukeSpaces( country ) ;

	LogMessage( "info", "Rom Header Information:" ) ;
	LogMessage( "info", ( boost::format("Console Name: %1%") % consoleName ).str() ) ;
	LogMessage( "info", ( boost::format("CopyRight: %1%") % copyRight ).str() ) ;
	LogMessage( "info", ( boost::format("Domestic Game Name: %1%") % domestic ).str() ) ;
	LogMessage( "info", ( boost::format("Overseas Game Name: %1%") % overseas ).str() ) ;
	LogMessage( "info", ( boost::format("Production: %1%") % production ).str() ) ;
	LogMessage( "info", ( boost::format("Checksum: 0x%1$x") % m_RomHeader.checksum ).str() ) ;
	LogMessage( "info", ( boost::format("Control: %1%") % control ).str() ) ;
	LogMessage( "info", ( boost::format("Rom Start: 0x%1$x") % m_RomHeader.rom_start ).str() ) ;
	LogMessage( "info", ( boost::format("Rom End: 0x%1$x") % m_RomHeader.rom_end  ).str() ) ;
	LogMessage( "info", ( boost::format("Ram Start: 0x%1$x") % m_RomHeader.ram_start ).str() ) ;
	LogMessage( "info", ( boost::format("Ram End: 0x%1$x") % m_RomHeader.ram_end ).str() ) ;
	LogMessage( "info", ( boost::format("Modem: %1%") % modem ).str() ) ;
	LogMessage( "info", ( boost::format("Memo: %1%") % memo ).str() ) ;
	LogMessage( "info", ( boost::format("Country: %1%") % country ).str() ) ;
	LogMessage( "info", "Rom Header Information Complete" ) ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::ValidateROMHeader( ) const
{	
	std::string consoleName( &m_RomHeader.segatitle[0], &m_RomHeader.segatitle[16] ) ;
	consoleName = codeslinger::string::NukeSpaces( consoleName ) ;

// 	if (consoleName.find( "SEGA MEGA DRIVE" ) == std::string::npos && 
// 		consoleName.find( "SEGA GENESIS" ) == std::string::npos )
//  	{
//  		LogMessage("err", ( boost::format("ERROR: ROM file header does not contains the console"
//  			" name SEGA MEGA DRIVE or SEGA GENESIS. Contains %1%") % consoleName ).str() ) ;
//  		return false ;
//  	}

	if (!ValidateChecksum( ))
	{
		// this is not fatal
		LogMessage("wrn", "Embedded checksum does not match calculated checksum.") ;
	}

	return true ;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::ValidateChecksum( ) const
{	
	return m_RomHeader.checksum == CalculateChecksum() ;
}

//////////////////////////////////////////////////////////

UWORD MegaDriveCartridge::CalculateChecksum() const
{
	UWORD checkSum = 0 ;
	for (long i = 0x200; i < m_CartridgeSize; i += 2)
	{
		checkSum += m_CartridgeMemory[i] * 256 ;
		checkSum += m_CartridgeMemory[i+1] ;
	}
	return checkSum ;
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::FixChecksum()
{
	UWORD checkSum = CalculateChecksum() ;

	LogMessage( "info", ( boost::format("Fixing Checksum with value: 0x%1$x") % checkSum ).str() ) ;

	m_CartridgeMemory[0x18E] = (checkSum >> 8) & 0xFF ;
	m_CartridgeMemory[0x18F] = checkSum & 0xFF ;
	m_RomHeader.checksum = checkSum ;
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::PageMemory( int pagenum, unsigned long pagesize, 
	unsigned long destStartPos, boost::shared_array<UBYTE>& destination )
{
	unsigned long numtocopy = std::min<unsigned long>( pagesize, m_CartridgeSize) ;

	if ( (pagenum * pagesize) > (unsigned long)m_CartridgeSize)
	{
		LogMessage("err", ( boost::format("ERROR: Cannot page in page number %1% as"
			" cartridge size is too small") % pagenum ).str() ) ;
	}
	else
	{
		memcpy( destination.get() + destStartPos, 
			m_CartridgeMemory.get() + (pagesize * pagenum), numtocopy ) ;
	}
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::WriteSaveRam( UDWORD address, UBYTE data )
{
	if (address & 1)
		address -= 1;
	else
		address += 1;
	m_SaveRam[address-m_SaveRamStart] = data;
}

//////////////////////////////////////////////////////////

bool MegaDriveCartridge::IsSaveRamAvailable() const
{
	return m_SaveRamAvailable;
}

//////////////////////////////////////////////////////////

std::pair<UDWORD,UDWORD> MegaDriveCartridge::GetSaveRamRange() const
{
	return std::make_pair( m_SaveRamStart, m_SaveRamEnd) ;
}

//////////////////////////////////////////////////////////

UBYTE MegaDriveCartridge::ReadSaveRam( UDWORD address ) const
{
	return m_SaveRam[address-m_SaveRamStart];
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::BackupSaveRam()
{
	using namespace std;
	ofstream outputFile( (boost::format("SRAM/%1%.sram") % m_RomHeader.domestic).str(), ofstream::binary );
	for (UDWORD i = 0 ; i < m_SaveRamLength; ++i)
	{		
		outputFile << m_SaveRam[i] ;
	}
	outputFile.close() ;
}

//////////////////////////////////////////////////////////

void MegaDriveCartridge::LoadSaveRam()
{
	using namespace std;
	ifstream outputFile( (boost::format("SRAM/%1%.sram") % m_RomHeader.domestic).str(), ofstream::binary );
	if (outputFile.is_open())
	{
		for (UDWORD i = 0 ; i < m_SaveRamLength; ++i)
		{		
			outputFile >> m_SaveRam[i] ;
		}
		outputFile.close();
	}
}

//////////////////////////////////////////////////////////