#include "Config_MegaDrive.h"
#include "MegaDrive.h"
#include "InputDevice/iInputDevice.h"
#include "OutputDevice/iOutputDevice.h"
#include "AudioDevice/iAudioDevice.h"
#include "Cartridge/MegaDriveCartridge.h"
#include "Debugger/MDDebugger.h"
#include <M68k/M68K.h>
#include <YM7101/YM7101.h>
#include <cs_core/IniFiles/IniFiles.h>
#include <cs_core/Utils/Maths.h>
#include <Z80/Z80.h>
#include <YM2616/YM2616.h>
#include <SN79489/SN79489.h>

#include <boost/date_time/posix_time/posix_time.hpp>

static MegaDrive::PTR g_instance ;

////////////////////////////////////////////////////

static UBYTE ReadZ80( UWORD address )
{
	return g_instance->Z80Read(address);
}

////////////////////////////////////////////////////

static void WriteZ80( UWORD address, UBYTE data )
{
	return g_instance->Z80Write(address, data);
}

////////////////////////////////////////////////////

static UBYTE ReadZ80IO( UWORD address )
{
	return g_instance->Z80ReadIO(address);
}

////////////////////////////////////////////////////

static void WriteZ80IO( UWORD address, UBYTE data )
{
	return g_instance->Z80WriteIO(address, data);
}

////////////////////////////////////////////////////

static UDWORD ReadMemory32( UDWORD address )
{
	return g_instance->ReadMemory32(address);
}

////////////////////////////////////////////////////

static UWORD ReadMemory16( UDWORD address )
{
	return g_instance->ReadMemory16(address);
}

////////////////////////////////////////////////////

static UBYTE ReadMemory8( UDWORD address )
{
	return g_instance->ReadMemory8(address);
}

////////////////////////////////////////////////////

static void WriteMemory32( UDWORD address, UDWORD data )
{
	return g_instance->WriteMemory32(address, data);
}

////////////////////////////////////////////////////

static void WriteMemory16( UDWORD address, UWORD data )
{
	return g_instance->WriteMemory16(address, data);
}

////////////////////////////////////////////////////

static void WriteMemory8( UDWORD address, UBYTE data )
{
	return g_instance->WriteMemory8(address, data);
}

////////////////////////////////////////////////////

MegaDrive::PTR MegaDrive::Create( ) 
{
	g_instance =  PTR( new MegaDrive() ) ;
	return g_instance;
}

////////////////////////////////////////////////////

MegaDrive::MegaDrive(void)	
	: c_m68kmemorysize(0x1000000)
	, c_maxrommemory(0x400000)
	, c_z80memorysize(0x10000)
	, m_m68kHasZ80Bus(true)
	, m_z80IsReset(true)
	, m_z80BankAddress(0x0)
	, m_Z80BankIndex(0x0)
	, m_Paused(false)
	, m_Version(0xa0)
{
	

	m_ControlPads.resize( sc_numcontrolpads ) ;

	m_FPS = codeslinger::inifile::IniFileReturnDefault<int>( "DesiredFPS", 60) ;

	Reset();
}

////////////////////////////////////////////////////

MegaDrive::~MegaDrive(void)
{
}

////////////////////////////////////////////////////

void MegaDrive::Reset()
{
	CreateMemory( m_M68KMemory, c_m68kmemorysize ) ;
	CreateMemory( m_Z80Memory, c_z80memorysize ) ;
	CreateM68K( ) ;
	CreateZ80( ) ;
	CreateYM7101( ) ;
	m_YM2616 = YM2616::Create();
	m_SN79489 = SN79489::Create();
	m_Paused = false;

	// default io
	m_M68KMemory[0xA10009] = 0x00;
	m_M68KMemory[0xA1000B] = 0x00;
	m_M68KMemory[0xA1000D] = 0x00;
	m_M68KMemory[0xA1000F] = 0xFF;
	m_M68KMemory[0xA10011] = 0x00;
	m_M68KMemory[0xA10013] = 0x00;
	m_M68KMemory[0xA10015] = 0xFF;
	m_M68KMemory[0xA10017] = 0x00;
	m_M68KMemory[0xA10019] = 0x00;
	m_M68KMemory[0xA1001B] = 0xFB;
	m_M68KMemory[0xA1001D] = 0x00;
	m_M68KMemory[0xA1001F] = 0x00;

	m_Version = (UBYTE)codeslinger::inifile::IniFileReturnDefault<int>("Version", 0xa0);
	m_HasSaveRam = false;
	m_SaveRamStart = 0x0;
	m_SaveRamEnd = 0x0;
}

////////////////////////////////////////////////////

void MegaDrive::CreateM68K( )
{
	m_M68K = M68K::Create() ;
	m_M68K->SetFuncWriteMemory32( ::WriteMemory32 ) ;
	m_M68K->SetFuncReadMemory32( ::ReadMemory32 ) ;
	m_M68K->SetFuncWriteMemory16( ::WriteMemory16 ) ;
	m_M68K->SetFuncReadMemory16(  ::ReadMemory16 ) ;
	m_M68K->SetFuncWriteMemory8(  ::WriteMemory8 ) ;
	m_M68K->SetFuncReadMemory8( ::ReadMemory8 ) ;
}

////////////////////////////////////////////////////

void MegaDrive::CreateZ80()
{
	m_Z80 = Z80::Create() ;
	CONTEXTZ80* context = m_Z80->GetContext( ) ;
	memset(&context->m_CartridgeMemory,0,sizeof(context->m_CartridgeMemory)) ;
	memset(&context->m_InternalMemory,0,sizeof(context->m_InternalMemory)) ;
	
	// these may have to be set to none 0 values so for now i'll set them to all
	// 0 instead of doing a memset so it becomes easier to edit later if necessary

	context->m_RegisterAF.reg = 0x0 ; // unofficial z80
	context->m_RegisterBC.reg = 0x0  ;// unofficial z80
	context->m_RegisterDE.reg = 0x0  ;// unofficial z80
	context->m_RegisterHL.reg = 0x0  ;// unofficial z80
	context->m_RegisterAFPrime.reg = 0x0  ;// unofficial z80
	context->m_RegisterBCPrime.reg = 0x0  ;// unofficial z80
	context->m_RegisterDEPrime.reg = 0x0  ;// unofficial z80
	context->m_RegisterHLPrime.reg = 0x0  ;// unofficial z80
	context->m_RegisterIX.reg = 0xFFFF  ;
	context->m_RegisterIY.reg = 0xFFFF  ;
	context->m_RegisterI = 0 ;
	context->m_RegisterR = 0 ;
	context->m_ProgramCounter = 0 ;
	context->m_OpcodeCycle = 0;
	context->m_StackPointer.reg = 0xDFF0 ;
	context->m_InternalMemory[0xFFFF] = 2 ; // official sega doc
	context->m_InternalMemory[0xFFFE] = 1 ; // official sega doc
	context->m_FuncPtrRead = ::ReadZ80;
	context->m_FuncPtrWrite = ::WriteZ80;
	context->m_FuncPtrIORead = ::ReadZ80IO ;
	context->m_FuncPtrIOWrite = ::WriteZ80IO ;
	context->m_IFF1 = false ;
	context->m_IFF2 = false ;
	context->m_Halted = false ;
	context->m_InteruptMode = 1 ;
	context->m_NMI = false ;
	context->m_NMIServicing = false ;
	context->m_EIPending = false;
}

////////////////////////////////////////////////////

void MegaDrive::CreateMemory(boost::shared_array<UBYTE>& memory, unsigned long int memSize)
{
	UBYTE* array = new UBYTE[memSize] ;
	memset(array, 0, memSize) ;
	memory = boost::shared_array<UBYTE>(array) ;
}

////////////////////////////////////////////////////

void MegaDrive::CreateYM7101( )
{
	m_YM7101 = YM7101::Create() ;
	m_YM7101->SetFuncReadMemory16( ::ReadMemory16 ) ;
}

////////////////////////////////////////////////////

void MegaDrive::SetPause( bool pause )
{
	m_Paused = pause;
}

////////////////////////////////////////////////////

bool MegaDrive::IsPaused() const
{
	return m_Paused;
}

////////////////////////////////////////////////////

void MegaDrive::Run( )
{
	m_M68K->Init( ) ;
	m_AudioDevice->Initialise( m_SN79489->GetPlayBuffer(), m_SN79489->GetFrequency(), m_SN79489->GetBufferSize() ) ;

	UDWORD lastframetime = m_FuncTickCount() ;
	bool keepRunning = true;
 	while(keepRunning)
 	{
		UDWORD c_currenttime = m_FuncTickCount() ;
		const bool c_newframe = (lastframetime + (1000/m_FPS) ) < c_currenttime ;
		if (c_newframe)
		{
			// must happen before output device update
			std::for_each( m_ControlPads.begin(), m_ControlPads.end(), boost::bind(&iInputDevice::Update, _1 ) ) ; 

			// output window may need to check for window messages etc
			keepRunning = m_OutputDevice->Update( ) ; 

			lastframetime = c_currenttime ;
			int numcycles = 0 ;
			const int c_numcyclesframe = M68K::sc_clockcyclespersecond / m_FPS ;
			bool allowZ80 = true;
			while (numcycles < c_numcyclesframe )
			{
				if (!m_Paused)
				{
					const int c_numclicks = m_M68K->Update( ) ;
					numcycles += c_numclicks ;
					m_YM7101->Update(c_numclicks) ;
					if (m_YM7101->IsRequestingInterrupt())
					{
						const int c_type = m_YM7101->GetInterruptType() ;
						m_M68K->RequestAutoVectorInterrupt( static_cast<M68K::InterruptType>( c_type ) ) ;

						// UNTESTED - Golden axe seems to do this, might be worth investigating with that
	// 					m_Z80->IncreaseRReg( ) ;
	// 					CONTEXTZ80* context = m_Z80->GetContext() ;
	// 					if (context->m_IFF1 && context->m_InteruptMode == 1)
	// 					{
	// 						ASSERT(false);
	// 						context->m_Halted = false ;
	// 
	// 						m_Z80->PushWordOntoStack(context->m_ProgramCounter) ;
	// 						context->m_ProgramCounter = 0x38 ;
	// 						context->m_IFF1 = false ;
	// 						context->m_IFF2 = false ;
	// 					}
					}
					if (allowZ80 && !m_z80IsReset && !m_m68kHasZ80Bus)
					{
						int cycles = m_Z80->ExecuteNextOpcode();
						m_SN79489->Update((float)cycles);
					}
					allowZ80 = !allowZ80;
					std::pair<int, int> resolution = m_YM7101->GetScreenResolution() ;
					m_OutputDevice->SetVideoBufferSize( resolution.first, resolution.second ) ;
				}
				m_Debugger->Update();
			}
			
	//		if (m_YM7101->IsTimeToRender())
				m_OutputDevice->Render( m_YM7101->GetScreen() ) ;
			if (m_AudioDevice->IsPlaybackComplete())
				m_SN79489->PlaybackCompleted();
			m_AudioDevice->Play( m_SN79489->GetPlayBuffer() ) ;

		}
 	}
	if (m_HasSaveRam)
	{
		m_Cartridge->BackupSaveRam();
	}
}

////////////////////////////////////////////////////

bool MegaDrive::AttachControlPad( int controlPort, const boost::shared_ptr<iInputDevice>& control )
{
	--controlPort ;
	if ((controlPort >= sc_numcontrolpads) || (controlPort < 0) )
	{
		LogMessage("err", ( boost::format( "ERROR: Invalid Control Port %1%. Must have a value"
			" less than %2%") % controlPort % sc_numcontrolpads ).str() ) ;
		return false ;
	}
	
	control->SetFunctionKeyPress(boost::bind(
		&MegaDrive::HandleControlKeyPressed, this, controlPort, _1, _2 ) ) ;

	m_ControlPads[controlPort] = control ;
	m_controlState[controlPort] = ControllerState();
	return true ;
}

////////////////////////////////////////////////////

bool MegaDrive::AttachOutputDevice( const boost::shared_ptr<iOutputDevice>& device )
{
	m_OutputDevice = device ;
	return true ;
}

////////////////////////////////////////////////////

bool MegaDrive::AttachAudioDevice( const boost::shared_ptr<iAudioDevice>& device )
{
	m_AudioDevice = device ;
	return true;
}

////////////////////////////////////////////////////

bool MegaDrive::AttachDebugger( const boost::shared_ptr<MDDebugger>& debugger )
{
	m_Debugger = debugger;
	return true;
}

////////////////////////////////////////////////////

bool MegaDrive::InsertCartridge( const boost::shared_ptr<MegaDriveCartridge>& cartridge )
{
	m_Cartridge = cartridge ;
	PageCartridgeMemory( 0, 0 ) ;
	m_HasSaveRam = m_Cartridge->IsSaveRamAvailable();
	if (m_HasSaveRam)
	{
		std::pair<int,int> saveRamRange = m_Cartridge->GetSaveRamRange();
		m_SaveRamStart = saveRamRange.first;
		m_SaveRamEnd = saveRamRange.second;
	}
	
	return true ;
}

////////////////////////////////////////////////////

void MegaDrive::HandleControlKeyPressed( int controlId, bool keyDown, int key )
{
	using codeslinger::maths::BitConditionalSet;
	if (key < KEY_NUM)
	{
		const MegaDriveKey c_mdkey = static_cast<MegaDriveKey>( key ) ;
		ControllerState& state = m_controlState[controlId];
		switch (c_mdkey)
		{
		case KEY_UP:
			state.notSet = BitConditionalSet(!keyDown, state.notSet, 0);
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 0);
			break;
		case KEY_DOWN:
			state.notSet = BitConditionalSet(!keyDown, state.notSet, 1);
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 1);
			break;
		case KEY_LEFT:
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 2);
			break;
		case KEY_RIGHT:
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 3);
			break;
		case KEY_A:
			state.notSet = BitConditionalSet(!keyDown, state.notSet, 4);
			break;
		case KEY_B:
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 4);
			break;
		case KEY_C:
			state.isSet = BitConditionalSet(!keyDown, state.isSet, 5);
			break;
		case KEY_START:
			state.notSet = BitConditionalSet(!keyDown, state.notSet, 5);
			break;
		}
	}
	else
	{
		LogMessage("err", ( boost::format( "ERROR: Invalid Control Key %1% pressed.") 
			% key ).str() ) ;
	}
}

////////////////////////////////////////////////////

void MegaDrive::PageCartridgeMemory( int pageindex, unsigned long startpos )
{
	m_Cartridge->PageMemory( pageindex, c_maxrommemory, startpos, m_M68KMemory ) ;
}

////////////////////////////////////////////////////

UDWORD MegaDrive::ReadMemory32(const UDWORD& address)
{
	LogMessage("wrn", 2,"Performing Read 32 bit");
	return ReadMemory16( address ) << 16 | ReadMemory16(address + 2) ;
}

////////////////////////////////////////////////////

void MegaDrive::WriteMemory32(const UDWORD& address, const UDWORD& data)
{
	LogMessage("wrn", 2,"Performing Write 32 bit");
	WriteMemory16( address, UWORD(data >> 16) );
	WriteMemory16( address + 2, UWORD( data & 0xFFFF) ) ;
}

////////////////////////////////////////////////////

void MegaDrive::WriteMemory16(  UDWORD address, const UWORD& data )
{

	address = address % c_m68kmemorysize ;
	
	if (m_HasSaveRam && address >= m_SaveRamStart && address <= m_SaveRamEnd)
	{
		m_Cartridge->WriteSaveRam(address, data >> 8);
		m_Cartridge->WriteSaveRam(address+1, data & 0xFF);
		return ;
	}	

	if (address < c_maxrommemory ) // cant write to rom
	{
		//_asm int 3;
	}
	else if (address >= 0xa00000 && address < 0xC00000)
	{
		if (address >= 0xA04000 && address <= 0xA04003)
		{
			LogMessage("wrn", "Doing a 16bit write of YM2616. Not sure what to do as never thought you could");
		}
		// z80
		else if (address <= 0xa07fff)
		{
			UDWORD z80Address = address - 0xa00000;
			Z80Write(z80Address, (UBYTE)( data >> 8 ));
		//	Z80Write(z80Address+1, (UBYTE)( data & 0xFF )); // charles mcdonald hardware info, dont write LSB
		}
		else if (address <= 0xA0FFFF)
		{
			// z80
			UDWORD z80Address = address - 0xA08000;
			Z80Write(z80Address, (UBYTE)( data >> 8 ));
		//	Z80Write(z80Address+1, (UBYTE)( data & 0xFF )); // charles mcdonald hardware info, dont write LSB
		}
		else if (address >= 0xa10000 && address <= 0xa1001f) // I/O
		{
			address = codeslinger::maths::BitSet(address, 0);  // even and odd contain same data
			if (address == 0xa10003) // control 1
			{
				ControllerState& state = m_controlState[0];
				state.isTopBitSet = codeslinger::maths::TestBit(data,6);
				state.isSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.isSet, 7);
				state.notSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.notSet, 7);
			}
			else if (address == 0xa10005) // control 2
			{
				ControllerState& state = m_controlState[1];
				state.isTopBitSet = codeslinger::maths::TestBit(data,6);
				state.isSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.isSet, 7);
				state.notSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.notSet, 7);
			}
			else if (address == 0xa10007) // dunno, light gun?
			{
				LogMessage("wrn", "Trying to write data port 0xa100006, dunno what it is... light gun?");
			}
			else if (address < 0xa1000e) // control ports
			{
				m_M68KMemory[address] = (UBYTE)data;
			}
			else
			{
				m_M68KMemory[address] = (UBYTE)data;
			}
		}	
		else if (address == 0xa11100)
		{
			// see section 2.2 of hardware information.txt
			m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
			m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
			if (codeslinger::maths::TestBit(data,8))
			{
				LogMessage("wrn", 1, "writing word to 0xa11100 giving bus to M68k") ;
				m_m68kHasZ80Bus = true;
			}
			else
			{
				LogMessage("wrn", 1, "writing word to 0xa11100 giving bus to Z80") ;
				m_m68kHasZ80Bus = false;
			}
		}	
		else if (address == 0xa11200)
		{
			// see section 2.2 of hardware information.txt
			m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
			m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
			if (codeslinger::maths::TestBit(data,8))
			{
				LogMessage("wrn", 1, "writing word to 0xa11200 z80 has stopped reset") ;
				m_z80IsReset = false;
	//			CallCreateZ80
			}
			else
			{
				LogMessage("wrn", 1, "writing word to 0xa11200 z80 is reset") ;
///				m_Z80->GetContext()->m_ProgramCounter = 0;
				m_Z80->ResetLine();
				m_z80IsReset = true;
			}
		}
		else if (address == 0xa130f1)
		{
			LogMessage("wrn", (boost::format("Writing 16 bit to 0xa130f1 with data 0x%1$x (SRAM access?)") % data ).str()) ;
			m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
			m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
		}
		else
		{
			LogMessage("wrn", (boost::format("Writing 16 bit to 0x%1$x with data 0x%2$x yet to implement whatever it is") % address % data ).str()) ;
			m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
			m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
		}
	}
	else if (address >= 0xC00000 && address <= 0xC0000F) // vdp
	{
		if (address < 0xC00004) m_YM7101->WriteDataPort16( data ) ;
		else if (address < 0xC00008) m_YM7101->WriteControlPort16( data ) ;
		// rest is read only
	}
	else if (address >= 0xC00010 && address < 0xE00000)
	{
		if (address == 0xC00011 || address == 0xC00013 || address == 0xC00015 || address == 0xC00017)
		{
			LogMessage("wrn", "Trying to write 16 bit PSG. Not sure about cycles") ;
			// charles mcdonald hardware information "If you want to write to the PSG via word-wide writes, the data must be in the LSB"
			m_SN79489->WriteData(m_Z80->GetContext()->m_OpcodeCycle,(UBYTE)(data & 0xFF));
		}
		else
		{
			LogMessage("wrn", (boost::format("Writing 16 bit to 0x%1$x with data 0x%2$x yet to implement whatever it is") % address % data ).str()) ;
			m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
			m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
		}
	}
	else if (address >= 0xE00000 && address <= 0xffffff) // mirrored ram
	{
		address = address % 0x100000 ;
		address += 0xe00000;
		m_M68KMemory[address] = (UBYTE)( data >> 8 ) ;
		m_M68KMemory[address+1] =(UBYTE)(data & 0xFF) ;
	}
}

////////////////////////////////////////////////////

UWORD MegaDrive::ReadMemory16(  UDWORD address )
{
	address = address % c_m68kmemorysize ;

	UWORD res = 0 ;
	if (m_HasSaveRam && address >= m_SaveRamStart && address <= m_SaveRamEnd)
	{
		res = m_Cartridge->ReadSaveRam(address) << 8 ;
		res |= m_Cartridge->ReadSaveRam(address+1);
		return res;
	}	

	if (address < 0xa00000)
	{
		res = m_M68KMemory[address] << 8;
		res |= m_M68KMemory[address+1] ;
	}
	else if (address >= 0xa00000 && address < 0xC00000)
	{
		if (address >= 0xA04000 && address <= 0xA04003)
		{
			LogMessage("wrn", "Doing a 16bit read of YM2616. Not sure what to do as never thought you could");
			return m_YM2616->ReadData();
		}
		else if (address < 0xa07fff)
		{
			// z80
			UDWORD z80Address = address - 0xa00000;
			res = Z80Read(z80Address) << 8;
			res |= Z80Read(z80Address) ; // charles mcdonald, hardware info, LSB is duplicated into MSB
		}
		else if (address <= 0xA0FFFF)
		{
			// z80
			UDWORD z80Address = address - 0xA08000;
			res = Z80Read(z80Address) << 8;
			res |= Z80Read(z80Address) ; // charles mcdonald, hardware info, LSB is duplicated into MSB
		}
		else if (address >= 0xa10000 && address <= 0xa1001f) // I/O
		{
			address = codeslinger::maths::BitSet(address, 0);  // even and odd contain same data
			if (address == 0xa10001) // version
			{
				return m_Version;
			}
			else if (address == 0xa10003) // control 1
			{
				ControllerState& state = m_controlState[0];
				return state.isTopBitSet ? state.isSet : state.notSet;
			}
			else if (address == 0xa10005) // control 2
			{
				ControllerState& state = m_controlState[1];
				return state.isTopBitSet ? state.isSet : state.notSet;
			}
			else if (address == 0xa10007) // dunno, light gun?
			{
				LogMessage("wrn", "Trying to read data port 0xa100006, dunno what it is... light gun?");
				return 0x7F;
			}
			else if (address < 0xa1000e) // control ports
			{
				return m_M68KMemory[address];
			}
			else
			{
				return m_M68KMemory[address];
			}
		}		
		else if (address == 0xa11100)
		{
			if (m_z80IsReset || !m_m68kHasZ80Bus)
			{
				LogMessage("wrn", 1, "reading state of bus word. Z80 has it, returning 1") ;
				return 1;
			}
			else
			{
				LogMessage("wrn", 1, "reading state of bus word. M68k has it, returning 0") ;
				return 0;				
			}
		}
		else if (address == 0xa11200)
		{
			// see section 2.2 of hardware information.txt
			LogMessage("wrn", "reading 0xa11200 to query z80 - no idea what to do") ;
			return 0 ;
		}
		else if (address == 0xa130f1)
		{
			LogMessage("wrn", "Reading 16 bit to 0xa130f1 (SRAM access?)") ;
			res = m_M68KMemory[address] << 8;
			res |= m_M68KMemory[address+1] ;
		}
		else
		{
			LogMessage("wrn", (boost::format("Reading 16 bit to 0x%1$x yet to implement whatever it is") % address ).str()) ;
			res = m_M68KMemory[address] << 8;
			res |= m_M68KMemory[address+1] ;
		}
	}
	else if (address >= 0xC00000 && address <= 0xC0000F) // vdp
	{
		if (address < 0xC00004) res = m_YM7101->ReadDataPort16( ) ;
		else if (address < 0xC00008) res = m_YM7101->ReadControlPort16( ) ;
		else if (address < 0xC0000f) res = m_YM7101->GetHVCounter( ) ;
	}
	else if (address >= 0xC00010 && address < 0xE00000)
	{
		if (address == 0xC00011 || address == 0xC00013 || address == 0xC00015 || address == 0xC00017)
		{
			// charles mcdonald hardware information "Reading the PSG addresses will cause the machine to lock up."
			return 0 ;
		}
		else
		{
			LogMessage("wrn", (boost::format("Reading 16 bit to 0x%1$x yet to implement whatever it is") % address ).str()) ;
			res = m_M68KMemory[address] << 8;
			res |= m_M68KMemory[address+1] ;
		}
		
	}
	else if (address >= 0xE00000 && address <= 0xffffff) // mirrored ram
	{
		address = address % 0x100000 ;
		address += 0xe00000;
		res = m_M68KMemory[address] << 8;
		res |= m_M68KMemory[address+1] ;
	}

	return res ;
}

////////////////////////////////////////////////////

void MegaDrive::WriteMemory8(  UDWORD address, const UBYTE& data )
{	
	address = address % c_m68kmemorysize ;

	if (m_HasSaveRam && address >= m_SaveRamStart && address <= m_SaveRamEnd)
	{
		m_Cartridge->WriteSaveRam(address, data);
		return ;
	}


	if (address < c_maxrommemory ) // cant write to rom
	{
		//_asm int 3; // movep seems to be doing this on vdpdoc.bin
	}
	else if (address >= 0xa00000 && address < 0xC00000)
	{
		if (address >= 0xA04000 && address <= 0xA04003)
		{
			if (address == 0xa04000)	m_YM2616->WriteAddressPart1(data);
			else if (address == 0xa04001) m_YM2616->WriteDataPart1(data);
			else if (address == 0xa04002)	m_YM2616->WriteAddressPart2(data);
			else if (address == 0xa04003) m_YM2616->WriteDataPart2(data);
			else
			{
				LogMessage("wrn", "Writing to invalid M68k YM2616 address");
			}
		}		
		else if (address < 0xa07fff)
		{
			// z80
			UDWORD z80Address = address - 0xa00000;
			Z80Write(z80Address, data);
		}	
		else if ( address <= 0xA0FFFF)
		{
			// z80 mirror
			LogMessage("wrn", "writing 8 bit z80 mirror - probably implmeneted incorrectly") ;
			UDWORD z80Address = address - 0xA08000;
			Z80Write(z80Address, data);
		}
		else if (address >= 0xa10000 && address <= 0xa1001f) // I/O
		{
			address = codeslinger::maths::BitSet(address, 0);  // even and odd contain same data
			if (address == 0xa10003) // control 1
			{
				ControllerState& state = m_controlState[0];
				state.isTopBitSet = codeslinger::maths::TestBit(data,6);
				state.isSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.isSet, 7);
				state.notSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.notSet, 7);
			}
			else if (address == 0xa10005) // control 2
			{
				ControllerState& state = m_controlState[1];
				state.isTopBitSet = codeslinger::maths::TestBit(data,6);
				state.isSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.isSet, 7);
				state.notSet = codeslinger::maths::BitConditionalSet(codeslinger::maths::TestBit(data,7), state.notSet, 7);
			}
			else if (address == 0xa10007) // dunno, light gun?
			{
				LogMessage("wrn", "Trying to write data port 0xa100006, dunno what it is... light gun?");
			}
			else if (address < 0xa1000e) // control ports
			{
				m_M68KMemory[address] = data;
			}
			else
			{
				m_M68KMemory[address] = data;
			}
		}	
		else if (address == 0xa11100)// || address == 0xa11101) // not sure if 0xa11101 is correct
		{
			// see section 2.2 of hardware information.txt
			m_M68KMemory[address] = data ;
			if (codeslinger::maths::TestBit(data,0))
			{
				LogMessage("wrn", 1, "writing byte to 0xa11100 giving bus to M68k") ;
				m_m68kHasZ80Bus = true;
			}
			else
			{
				LogMessage("wrn", 1, "writing byte to 0xa11100 giving bus to Z80") ;
				m_m68kHasZ80Bus = false;
			}
		}
		else if (address == 0xa11200)
		{
			// see section 2.2 of hardware information.txt
			m_M68KMemory[address] = data ;
			if (codeslinger::maths::TestBit(data,0))
			{
				LogMessage("wrn", 1, "writing byte to 0xa11200 z80 has stopped the reset") ;
				m_z80IsReset = false;
			}
			else
			{
				LogMessage("wrn", 1, "writing byte to 0xa11200 z80 is reset") ;
				m_Z80->ResetLine();
				m_z80IsReset = true;
			}
		}
		else if (address == 0xa130f1)
		{
			LogMessage("wrn", (boost::format("Writing 8 bit to 0xa130f1 with data 0x%1$x (SRAM access?)") % data ).str()) ;
			m_M68KMemory[address] = data ;
		}
		else
		{
			LogMessage("wrn", (boost::format("Writing 8 bit to 0x%1$x with data 0x%2$x yet to implement whatever it is") % address % data ).str()) ;
			m_M68KMemory[address] = data ;
		}
	}
	else if (address >= 0xC00000 && address <= 0xC0000F) // vdp
	{
		if (address < 0xC00004) m_YM7101->WriteDataPort8( data ) ;
		else if (address < 0xC00008) m_YM7101->WriteControlPort8( data ) ;
		// rest is read only
	}
	else if (address >= 0xC00010 && address < 0xE00000)
	{
		if (address == 0xC00011)
		{
			LogMessage("wrn", "Trying to write 8 bit PSG. Not sure about cycles...") ;
			m_SN79489->WriteData(m_Z80->GetContext()->m_OpcodeCycle,data);
		}
		else
		{
			LogMessage("wrn", (boost::format("Writing 8 bit to %1$x with data %2$x yet to implement whatever it is") % address % data ).str()) ;
			m_M68KMemory[address] = data ;
		}
	}
	else if (address >= 0xE00000 && address <= 0xffffff) // mirrored ram
	{
		address = address % 0x100000 ;
		address += 0xe00000;
		m_M68KMemory[address] = data ;
	}
}

////////////////////////////////////////////////////

UBYTE MegaDrive::ReadMemory8(  UDWORD address )
{
	address = address % c_m68kmemorysize ;

	UBYTE res = 0 ;
	if (m_HasSaveRam && address >= m_SaveRamStart && address <= m_SaveRamEnd)
	{
		res = m_Cartridge->ReadSaveRam(address);
		return res;
	}	
	
	if (address < 0xa00000)
	{
		res = m_M68KMemory[address] ;
	}
	else if (address >= 0xa00000 && address < 0xC00000)
	{
		if (address >= 0xA04000 && address <= 0xA04003)
		{
			return m_YM2616->ReadData();
		}
		else if (address < 0xa07fff)
		{
			// z80
			UDWORD z80Address = address - 0xa00000;
			return Z80Read(z80Address) ;
		}
		else if (address <= 0xA0FFFF)
		{
			// z80
			LogMessage("wrn", "reading 8 bit z80 mirror - more than likely implemeneted incorrectly") ;
			UDWORD z80Address = address - 0xA08000;
			return Z80Read(z80Address) ;
		}
		else if (address >= 0xa10000 && address <= 0xa1001f) // I/O
		{
			address = codeslinger::maths::BitSet(address, 0);  // even and odd contain same data
			if (address == 0xa10001) // version
			{
				return m_Version;
			}
			else if (address == 0xa10003) // control 1
			{
				ControllerState& state = m_controlState[0];
				return state.isTopBitSet ? state.isSet : state.notSet;
			}
			else if (address == 0xa10005) // control 2
			{
				ControllerState& state = m_controlState[1];
				return state.isTopBitSet ? state.isSet : state.notSet;
			}
			else if (address == 0xa10007) // dunno, light gun?
			{
				LogMessage("wrn", "Trying to read data port 0xa100006, dunno what it is... light gun?");
				return 0x7F;
			}
			else if (address < 0xa1000e) // control ports
			{
				return m_M68KMemory[address];
			}
			else
			{
				return m_M68KMemory[address];
			}
		}	
		else if (address == 0xa11100 )//|| address == 0xa11101) // not sure if 0xa11101 is correct
		{			
			if (m_z80IsReset || !m_m68kHasZ80Bus)
			{
				LogMessage("wrn", 1, "reading state of bus byte. Z80 has it, returning 1") ;
				return 1;
			}
			else
			{
				LogMessage("wrn", 1, "reading state of bus byte. M68k has it, returning 0") ;
				return 0;				
			}
		}
		else if (address == 0xa11200)
		{
			// see section 2.2 of hardware information.txt
			LogMessage("wrn", "reading 0xa11200 to query z80 - no idea what to do") ;
			return 0;
		}
		else if (address == 0xa130f1)
		{
			LogMessage("wrn", "Reading 8 bit to 0xa130f1 (SRAM access?)") ;
			res = m_M68KMemory[address];
		}
		else
		{
			LogMessage("wrn", (boost::format("Reading 8 bit to 0x%1$x yet to implement whatever it is") % address ).str()) ;
			res = m_M68KMemory[address] ;
		}
	}
	else if (address >= 0xC00000 && address <= 0xC0000F) // vdp
	{
		if (address < 0xC00004)
		{
			UWORD word = m_YM7101->ReadDataPort16( ) ;
			if ((address % 2) == 0)
			{
				res = UBYTE(word >>8) ;
			}
			else
			{
				res = UBYTE(word);
			}
		}
		else if (address < 0xC00008)
		{
			UWORD word = m_YM7101->ReadControlPort16( ) ;
			if ((address % 2) == 0)
			{
				res = UBYTE(word >>8) ;
			}
			else
			{
				res = UBYTE(word);
			}
		}
		else if (address < 0xC0000f)
		{
			if ((address % 2) == 0)	res = m_YM7101->GetVCounter( ) ;
			else res = m_YM7101->GetHCounter( ) ;
		}
	}
	else if (address >= 0xC00010 && address < 0xE00000)
	{
		if (address == 0xC00011)
		{
			LogMessage("wrn", "Trying to read from 8bit PSG. Not sure what to do") ;
			return 0 ;
		}
		else
		{

			LogMessage("wrn", (boost::format("Reading 8 bit to 0x%1$x yet to implement whatever it is") % address ).str()) ;
			res = m_M68KMemory[address] ;
		}
	}
	else if (address >= 0xE00000 && address <= 0xffffff) // mirrored ram
	{
		address = address % 0x100000 ;
		address += 0xe00000;
		res = m_M68KMemory[address] ;
	}

	return res ;
}

////////////////////////////////////////////////////

void MegaDrive::SetFuncTickCount( const boost::function<UDWORD (void)>& tickcount )
{
	m_FuncTickCount = tickcount ;
}

////////////////////////////////////////////////////

UBYTE MegaDrive::Z80Read( UDWORD address )
{
	UBYTE res = 0;
	if (address < 0x2000)
	{
		LogMessage("wrn", 1, "Z80 - Reading from z80 RAM");
		res = m_Z80Memory[address];
	}
	else if (address < 0x4000)
	{
		LogMessage("wrn", "Z80 - Reading from z80 RAM mirror");
		res = m_Z80Memory[address-0x2000];
	}
	else if (address < 0x6000)
	{
		if (address >= 0x4000 && address <= 0x4003)
		{
			return m_YM2616->ReadData();
		}
		else
		{
			LogMessage("wrn", "Reading  from invalid z80 YM2616 address");
		}
	}
	else if (address < 0x6100)
	{
		LogMessage("wrn", "Z80 - Reading from bank address register");
	}	
	else if (address < 0x7F00)
	{
		LogMessage("wrn", "Z80 - Reading from unused");
		res = 0xFF;
	}
	else if (address < 0x8000)
	{
		if (address == 0x7f11)
		{
			LogMessage("wrn", "Z80 - Reading from PSG not sure what to return");
			return 0 ;
		}
		else
		{
			LogMessage("wrn", "Z80 - Reading from VDP");
		}
	}
	else
	{
		//LogMessage("wrn", "Z80 - Reading out of bounds");
		//ASSERT(false);

		UWORD addr68k = m_z80BankAddress<<15;
		addr68k+=address & 0x7fff;
		res = ReadMemory8(addr68k);
	}
	return res;
}

////////////////////////////////////////////////////

// void MegaDrive::PerformZ80Bank()
// {
// 	memcpy( m_Z80Memory.get() + 0x8000, m_M68KMemory.get() + m_z80BankAddress, 32 * 1024);
// }

////////////////////////////////////////////////////

// void MegaDrive::Z80BankWrite(UBYTE data)
// {
// 	m_z80BankAddress = codeslinger::maths::BitConditionalSet(data != 0x0, m_z80BankAddress, m_Z80BankIndex + 0xF);
// 	++m_Z80BankIndex;
// 	if (m_Z80BankIndex == 0x9)
// 	{
// 		PerformZ80Bank( );
// 		m_Z80BankIndex = 0x0 ;
// 		m_z80BankAddress = 0x0;
// 	}
// }

////////////////////////////////////////////////////

void MegaDrive::Z80Write(UDWORD address, UBYTE data)
{
	if (address < 0x2000)
	{
		LogMessage("wrn", 1, "Z80 - Writing to z80 RAM");
		m_Z80Memory[address] = data;
	}
	// UNSURE IF THIS IS ACTUALLY MIRRORED
	else if (address < 0x4000)
	{
		LogMessage("wrn", "Z80 - Writing to z80 RAM mirror");
		m_Z80Memory[address - 0x2000] = data;
	}
	else if (address < 0x6000)
	{
		if (address == 0x4000)	m_YM2616->WriteAddressPart1(data);
		else if (address == 0x4001) m_YM2616->WriteDataPart1(data);
		else if (address == 0x4002)	m_YM2616->WriteAddressPart2(data);
		else if (address == 0x4003) m_YM2616->WriteDataPart2(data);
		else
		{
			LogMessage("wrn", "Writing to invalid z80 YM2616 address");
		}
	}
	else if (address < 0x6100)
	{
		//Z80BankWrite(data);
		m_z80BankAddress>>=1;
		m_z80BankAddress+=(data&1)<<8;
		m_z80BankAddress&=0x1ff;
	}
	else if (address < 0x7F00)
	{
		LogMessage("wrn", "Z80 - Writing to unused");
	}
	else if (address < 0x8000)
	{
		if (address == 0x7f11)
		{
			m_SN79489->WriteData(m_Z80->GetContext()->m_OpcodeCycle, data);
		}
		else
		{
			LogMessage("wrn", "Z80 - Writing to VDP");
		}
	}
 	else
 	{
// 		LogMessage("wrn", "Z80 - Writing out of bounds");
//  	ASSERT(false);
		UWORD addr68k = m_z80BankAddress<<15;
		addr68k+=address & 0x7fff;
		WriteMemory8(addr68k, data);
 	}
}

////////////////////////////////////////////////////

UBYTE MegaDrive::Z80ReadIO( UDWORD address )
{
	// i think if address is 0x7f then it is the io port for the PSG. Not sure about YM2616 or VDP...
	ASSERT(false);
	return 0;
}

////////////////////////////////////////////////////

void MegaDrive::Z80WriteIO(UDWORD address, UBYTE data)
{
	// i think if address is 0x7f then it is the io port for the PSG. Not sure about YM2616 or VDP...
	ASSERT(false);
}

////////////////////////////////////////////////////