#pragma once
#ifndef _MEGA_DRIVE_H
#define _MEGA_DRIVE_H

#include <string>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>

class iInputDevice ;
class iOutputDevice ;
class iAudioDevice ;
class MegaDriveCartridge ;
class M68K ;
class Z80 ;
class YM7101 ;
class YM2616 ;
class SN79489;
class MDDebugger;

class MegaDrive
{
public:

	enum MegaDriveKey
	{
		KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, KEY_A, KEY_B, KEY_C, KEY_START, KEY_NUM
	};

public:
	typedef boost::shared_ptr<MegaDrive> PTR ;

	static PTR Create( ) ;
	
	bool AttachControlPad( int controlPort, const boost::shared_ptr<iInputDevice>& control ) ;
	bool AttachOutputDevice( const boost::shared_ptr<iOutputDevice>& device ) ;
	bool AttachAudioDevice( const boost::shared_ptr<iAudioDevice>& device );
	bool AttachDebugger( const boost::shared_ptr<MDDebugger>& debugger );
	void Run( ) ;

	bool InsertCartridge( const boost::shared_ptr<MegaDriveCartridge>& cartridge ) ;
	void SetFuncTickCount( const boost::function<UDWORD (void)>& tickcount ) ;

	void HandleControlKeyPressed( int controlId, bool keyDown, int key ) ;
	void SetPause( bool pause );
	bool IsPaused( ) const ;

	void WriteMemory32( const UDWORD& address, const UDWORD& data ) ;
	UDWORD ReadMemory32( const UDWORD& address ) ;
	void WriteMemory16( UDWORD address, const UWORD& data ) ;
	UWORD ReadMemory16( UDWORD address ) ;
	void WriteMemory8( UDWORD address, const UBYTE& data ) ;
	UBYTE ReadMemory8( UDWORD address ) ;
	UBYTE Z80Read( UDWORD address );
	UBYTE Z80ReadIO( UDWORD address );
	void Z80Write(UDWORD address, UBYTE data);
	void Z80WriteIO(UDWORD address, UBYTE data);

	virtual ~MegaDrive(void);
private:

	static const int sc_numcontrolpads = 2 ;
	const unsigned long c_m68kmemorysize ;
	const unsigned long c_maxrommemory ;
	const unsigned long c_z80memorysize;

	MegaDrive(void);

	void CreateM68K( ) ;
	void CreateZ80( ) ;
	void CreateYM7101( ) ;
	void CreateMemory(boost::shared_array<UBYTE>& memory, unsigned long int memSize) ;
	void PageCartridgeMemory( int pageindex, unsigned long startpos ) ;
	


	void Z80BankWrite(UBYTE data);
	void PerformZ80Bank( );
	void Reset();

	boost::shared_ptr<iOutputDevice> m_OutputDevice ;
	boost::shared_ptr<iAudioDevice> m_AudioDevice ;
	boost::shared_ptr<MegaDriveCartridge> m_Cartridge ;
	std::vector< boost::shared_ptr<iInputDevice> > m_ControlPads ;

	boost::shared_ptr<M68K> m_M68K ;
	boost::shared_ptr<Z80> m_Z80 ;
	boost::shared_ptr<YM7101> m_YM7101 ;
	boost::shared_ptr<YM2616> m_YM2616 ;
	boost::shared_ptr<SN79489> m_SN79489 ;
	boost::shared_ptr<MDDebugger> m_Debugger;
	boost::shared_array<UBYTE> m_M68KMemory ;
	boost::shared_array<UBYTE> m_Z80Memory;
	int m_FPS ;
	bool m_m68kHasZ80Bus;
	bool m_z80IsReset;
	UWORD m_z80BankAddress;
	UBYTE m_Z80BankIndex;
	bool m_Paused;
	UBYTE m_Version ;
	bool m_HasSaveRam ;
	UDWORD m_SaveRamStart;
	UDWORD m_SaveRamEnd;


	struct ControllerState
	{
		bool isTopBitSet;
		UBYTE notSet;
		UBYTE isSet;
		ControllerState() : isTopBitSet(false), notSet(0x33), isSet(0xff) {}
	};

	std::map<int, ControllerState> m_controlState;

	boost::function<UDWORD (void)> m_FuncTickCount ;
	friend class MDDebugger;
};

#endif