#include "Config_MegaDrive.h"
#include "MegaDrive.h"
#include "InputDevice/InputSDLWindow.h"
#include "OutputDevice/SDLGLWindow.h"
#include "AudioDevice/SDLAudioDevice.h"
#include "Cartridge/MegaDriveCartridge.h"
#include <cs_core/IniFiles/IniFiles.h>
#include <cs_core/Logging/Logging.h>
#include <SDL/SDL.h>
#include <Windows.h>
#include <cs_windows/Logging/LogMsgDestWindowsDebugStr.h>
#include "Debugger/MDDebugger.h"

////////////////////////////////////////////////////////////////////

static void IniFileInit()
{
	codeslinger::inifile::IniFileCreateManager( ) ;
	codeslinger::inifile::IniFileLoadSettings("Main", "default.ini" ) ;
}

////////////////////////////////////////////////////////////////////

static void LoggingInit( )
{
	codeslinger::logging::LoggingCreateManager( ) ;
 	codeslinger::logging::LoggingAddDestination("All", "combined",
 		boost::shared_ptr<codeslinger::logging::LogMsgDestination>
 		( new codeslinger::windows::logging::LogMsgDestWindowsDebugStr() ) ) ;
}

////////////////////////////////////////////////////////////////////

static iInputDevice::PTR CreateControlPad1( )
{
	iInputDevice::PTR res = InputSDLWindow::Create( ) ;
	res->MapKey( MegaDrive::KEY_UP, SDLK_UP ) ;
	res->MapKey( MegaDrive::KEY_DOWN, SDLK_DOWN ) ;
	res->MapKey( MegaDrive::KEY_LEFT, SDLK_LEFT ) ;
	res->MapKey( MegaDrive::KEY_RIGHT, SDLK_RIGHT ) ;
	res->MapKey( MegaDrive::KEY_A, SDLK_j ) ;
	res->MapKey( MegaDrive::KEY_B, SDLK_k ) ;
	res->MapKey( MegaDrive::KEY_C, SDLK_l ) ;
	res->MapKey( MegaDrive::KEY_START, SDLK_RETURN ) ;
	
	return res ;
}

////////////////////////////////////////////////////////////////////

static iInputDevice::PTR CreateControlPad2( )
{
	iInputDevice::PTR res = InputSDLWindow::Create( ) ;	
	res->MapKey( MegaDrive::KEY_UP, SDLK_w ) ;
	res->MapKey( MegaDrive::KEY_DOWN, SDLK_s ) ;
	res->MapKey( MegaDrive::KEY_LEFT, SDLK_a ) ;
	res->MapKey( MegaDrive::KEY_RIGHT, SDLK_d ) ;
	res->MapKey( MegaDrive::KEY_A, SDLK_f ) ;
	res->MapKey( MegaDrive::KEY_B, SDLK_g ) ;
	res->MapKey( MegaDrive::KEY_C, SDLK_h ) ;
	res->MapKey( MegaDrive::KEY_START, SDLK_SPACE ) ;
	
	return res ;
}

////////////////////////////////////////////////////////////////////

static boost::shared_ptr<MDDebugger> CreateDebugger( const boost::shared_ptr<MegaDrive>& md )
{
#ifdef ENABLE_DEBUGGER
	iInputDevice::PTR input = InputSDLWindow::Create( ) ;	
	input->MapKey( MDDebugger::DEBUG_PAUSE, SDLK_p ) ;
	input->MapKey( MDDebugger::DEBUG_DUMP_VRAM, SDLK_y ) ;
	input->MapKey( MDDebugger::DEBUG_TOGGLE_LOGGING_ENABLED, SDLK_e ) ;
	return boost::shared_ptr<MDDebugger>(new MDDebugger(md, input));
#else

	return boost::shared_ptr<MDDebugger>(new MDDebugger());
#endif
}


////////////////////////////////////////////////////////////////////

static boost::filesystem::path GetRomFilename( std::string commandLine )
{
	boost::filesystem::path res ;
	if (commandLine.empty())
	{
	//	res = codeslinger::inifile::IniFileReturnDefault<std::string>("Main","RomFile","Hard Drivin' (W) [!].bin") ;
	//	res = "snake.bin";
	//	res = "miketest.bin";
	//	res = "STETEST2.zip";
	//	res = "Mortal Kombat (1992)(Midway).zip";
	//	res = "Mortal Kombat II (1994)(Acclaim).zip";
	//	res = "Street Fighter 2 Special Champion Edition (U) [!].zip"; 
	//	res = "Sonic.bin";
	//	res = "VDPDoc.bin";
	//	res = "Golden Axe Rev 1 (1989)(Sega).bin";
	//	res = "CrazyBus v0.7 (2004)(Maneiro, Tom)(es)(PD).zip";
	//	res = "Story of Thor, The - A Successor of the Light (1995)(Sega)(EU)(en).zip";
	//	res = "Mortal Kombat 3 (1995)(Midway)(US)[h2].zip";
	//	res = "Chuck Rock (U) [b1].zip";
	//	res = "Mega Bomberman (JUE) [!].zip";
	//	res = "Virtua Fighter 2 (F) [!].zip" ;
	//	res = "Arrow.zip";
	//	res = "Toy Story (4) [b1].zip";
	//	res = "Virtua Fighter 2 (F) [!].zip";
	//	res = "Sonic the Hedgehog 2 (JUE) [t1].zip";
	//	res = "Galaxy Force II (REV 01) (JU) [!].zip";
	}
	else
	{
		if (commandLine[0] == '\"')
		{
			commandLine.erase( commandLine.begin(), commandLine.begin() + 1) ;
		}
		if (commandLine[ commandLine.size() -1 ] == '\"')
		{
			commandLine.erase( commandLine.end() - 1, commandLine.end() ) ;
		}
		res = commandLine ;		
	}
	return res ;
}

////////////////////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	PSTR lpCmdLine, INT nCmdShow)
{
	IniFileInit( ) ;
	LoggingInit( ) ;

	boost::filesystem::path romFilename( GetRomFilename( lpCmdLine ) ) ;

	MegaDrive::PTR megaDrive = MegaDrive::Create() ;

	MegaDriveCartridge::PTR cartridge = MegaDriveCartridge::Create() ;
	if (!cartridge->Init(romFilename))
	{
		LogMessage("err", ( boost::format("ERROR: Failed to load ROM file %1%") 
			% romFilename.file_string() ).str() ) ;
		return 0 ;
	}
	if (!cartridge->ValidateChecksum( ))
	{
		int res = MessageBox(NULL, "The checksum embedded in the ROM Header" 
			" appears to be incorrect. Would you like to apply a temporary fix? Note this"
			" will not modify your game ROM and there is no guarantee the game will play" 
			" correctly.", "Invalid Checksum" ,	MB_YESNO | MB_ICONWARNING) ;

		if (res == IDYES)
		{
			cartridge->FixChecksum( ) ;
		}
	}

	iInputDevice::PTR control1 = CreateControlPad1( ) ;
	iInputDevice::PTR control2 = CreateControlPad2( ) ;
	iOutputDevice::PTR output = SDLGLWindow::Create( ) ;
	iAudioDevice::PTR audio = SDLAudioDevice::Create( );

	megaDrive->SetFuncTickCount( SDL_GetTicks ) ;

	boost::shared_ptr<MDDebugger> debugger = CreateDebugger(megaDrive);

	if ( megaDrive->InsertCartridge( cartridge ) &&
		megaDrive->AttachControlPad(1, control1) &&
		megaDrive->AttachControlPad(2, control2) &&
		megaDrive->AttachOutputDevice( output ) &&
		megaDrive->AttachAudioDevice( audio ) &&
		megaDrive->AttachDebugger(debugger) )
	{
		megaDrive->Run( ) ;
	}	

	codeslinger::inifile::IniFileDestroyManager( ) ;
	codeslinger::logging::LoggingDestroyManager( ) ;

	return 0;
}

